/* Copyright 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Search all trees for domnums different.

   Five domnum, indomnum, perfect domnum, semitotdomnum, totdomnum
   n=20
   (picture in vpar/devel/try.gp)
   https://hog.grinvin.org/ViewGraphInfo.action?id=31101

   ---------
   Seven domnum, indomnum, perfect domnum, semitotdomnum, totdomnum,
   matchnum, indnum

   n=24 several
   minimum diameter 10 unique
   num 7 n=24 found [0,1,2,3,4,5,2,7,8,9,2,11,12,13,2,15,15,15,1,19,20,21,22,19]
   >>sparse6<<:W_`abc`efg`ijk`mmm_qrstq
   values 7 9 8 10 12 13 11

   maximum diameter 15 unique
   num 7 n=24 found [0,1,2,3,4,5,6,7,8,5,4,4,3,13,13,1,16,17,18,19,20,21,18,1]
   >>sparse6<<:W_`abcdefcbbakk_nopqrsp_
   values 7 8 9 10 12 13 11
   diameter 15

   ---------
   Two domnum,indomnum attained 1 way each.
   n=7 bistar 3,4.

   Two domnum,totdomnum attained 1 way each.
   n=7 complete binary.

   Three domnum,indomnum,totdomnum attained 1 way each.
   n=16 four of.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_gentreeg_main(int argc, char**argv);

const int option_verbose = 0;

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))


unsigned long long  tree_count = 0;
int found = 0;
SG_DECL(sg);
SG_DECL(canon_sg);

void
my_gentreeg_outproc (FILE *outfile, int *vpar, int n)
{
  int i;
  tree_count++;
  if (NAUTYEXTRA_PROGRESS)
    printf("n=%d %Lu\r", n, tree_count);

  if (option_verbose) {
    printf("tree %p  n=%d\n", vpar,n);
    nautyextra_print_vpar(vpar,n);
  }

  int values[8];
  int num = 0;

#define NEW_VALUE(func)         \
  values[num] = func(vpar,n);   \
  for (i=0; i<num; i++)         \
    if (values[i]==values[num]) \
      return;                   \
  num++

#define NEW_VALUE_AND_COUNT(structname,func,fieldname)         \
  {                                                  \
    struct structname info;                          \
    func(vpar,n, &info);                             \
    if (info.count != 1) return;                     \
    values[num] = info.fieldname;                    \
    for (i=0; i<num; i++)                            \
      if (values[i]==values[num])                    \
        return;                                      \
    num++;                                           \
  }

  NEW_VALUE_AND_COUNT(nautyextra_domnum_and_count,nautyextra_domnum_and_count_vpar,domnum);
  NEW_VALUE_AND_COUNT(nautyextra_indomnum_and_count,nautyextra_indomnum_and_count_vpar,indomnum);
  NEW_VALUE_AND_COUNT(nautyextra_totdomnum_and_count,nautyextra_totdomnum_and_count_vpar,totdomnum);

  /* NEW_VALUE(nautyextra_domnum_vpar); */
  /* NEW_VALUE(nautyextra_indomnum_vpar); */
  /* NEW_VALUE(nautyextra_perfect_domnum_vpar); */
  /* NEW_VALUE(nautyextra_semitotdomnum_vpar); */
  /* NEW_VALUE(nautyextra_totdomnum_vpar); */

  /* NEW_VALUE(nautyextra_indnum_vpar); */
  /* NEW_VALUE(nautyextra_matchnum_vpar); */

  printf("num %d n=%d found ",num,n);
  nautyextra_print_vpar_gp(vpar,n);
  printf("\n%s%s",SPARSE6_HEADER, nautyextra_vpar_to_s6(vpar,n));
  printf("  values");
  for (i = 0; i < num; i++)
    printf(" %d",values[i]);
  printf("\n");

  printf("  diameter %d\n", nautyextra_diameter_vpar(vpar,n));
  found++;
}

int
main (void)
{
  setbuf(stdout,NULL);
  nautyextra_view_synchronous = 1;
  nautyextra_want_progress = 1;

  SG_INIT(sg);
  SG_INIT(canon_sg);
  int num_vertices;
  for (num_vertices = 1;
       num_vertices <= 40;
       num_vertices += 1) {
    printf("gentreeg  num_vertices=%d\n", num_vertices);
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    tree_count = 0;

    int gentreeg_argc = 0;
    char *gentreeg_argv[10];
    gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
    gentreeg_argv[gentreeg_argc++] = "-q";       /* quiet */
    gentreeg_argv[gentreeg_argc++] = num_vertices_str;
    gentreeg_argv[gentreeg_argc] = NULL;
    my_gentreeg_main(gentreeg_argc,gentreeg_argv);

    if (found) break;
    printf("found %d many         \n", found);
  }

  return 0;
}
