/* Copyright 2018, 2022 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/*
  reps                   3
  n=    1 2 3 4 5  6  7  8  9
  conn=[1,2,3,6,9,15,22,36,51,87,127,0]
  disc=[0,1,2,4,6,12,18,36,54,90,135,225]
  \\                       ^^

  most minimal dominating sets in a connected graph
  conn
  not in OEIS: 1,2,3,6,9,15,22,36,51,87,127

  most minimal dominating sets in a disconnected graph
  disc
  not in OEIS:   1,2,4,6,12,18,36,54,90,135,225

  most minimal dominating sets in any graph
  vector(#conn,i,max(conn[i],disc[i]))
  1,2,3,6,9,15,22,36,54,90,135
  probably: A332406 Num graphs n nodes with totdomnum 2

  vector_product(v) = prod(i=1,#v,v[i]);
  disc=vector(#conn+1);  \\ disconnected, by partitions of connected
  for(n=1,#disc, my(m=0); \
  forpart(p=n, if(p!=Vecsmall([n]),\
  m=max(m,vector_product(vecextract(conn,p))))); \
  disc[n]=m); disc

  my(l=List([])); \
  forpart(p=12, listput(l,[vector_product(vecextract(conn,p)),Vec(p)])); vecsort(Vec(l))

  n=2 2 sets, unique
  path-2
  :An

  n=3 3 sets, unique
  3-cycle
  :BcN

  n=4 6 sets, unique
  4-cycle
  :Cci

  n=5 9 sets
  https://hog.grinvin.org/ViewGraphInfo.action?id=438
  :DaHIQN
  9^(1/5)
  1.551845

  n=6 15 sets octahedral graph
  https://hog.grinvin.org/ViewGraphInfo.action?id=226
  :Ea@_WGxGs
  15^(1/6)
  1.57041780247501973531052
  A011350

  n=7 connected or disconnected, 22 sets
  https://hog.grinvin.org/ViewGraphInfo.action?id=868
  :FaqiBQRP^
  22^(1/7)
  1.555158

  n=8, 36 sets, 2 graphs
  disconnected 4-cycle and 4-cycle
  https://hog.grinvin.org/ViewGraphInfo.action?id=1415
  :GaKM{UQ
  connected
  :GaHgyMCUSTn
  https://hog.grinvin.org/ViewGraphInfo.action?id=30664

  --
  n=9 connected 51 sets, 3 of
  261080 graphs, or 197772 graphs degree>=2
  :H`A_wa_K`WaEQG`cJ
  :H`AK?qFG[@cIYGaDL
  :H`AK@qEQOqPGSsPDK^
  https://hog.grinvin.org/ViewGraphInfo.action?id=30667
  https://hog.grinvin.org/ViewGraphInfo.action?id=30669
  https://hog.grinvin.org/ViewGraphInfo.action?id=30671

  n=9 54 by pairing 4-cycle and n=5 graph
  :H`AGichStRF
  https://hog.grinvin.org/ViewGraphInfo.action?id=30673

  --
  n=10 connected max 87 reps 1
  count 11716571 graphs, 9808209 degree>=2
  87^(1/10) = 1.562974618
  9808209.0/197772 = 49
  :IcAKWQJGhhcLK`s^
  https://hog.grinvin.org/ViewGraphInfo.action?id=30675

  n=10 any 90 sets 4-cycle and octahedral n=6
  count 12005168 graphs
  :I`AGicgU`YENSps^
  https://hog.grinvin.org/ViewGraphInfo.action?id=30677

  --
  n=11 connected, 902884343 degree>=2, max 127 sets, reps 2
  .              1006700565 all connected (A001349)
  about 20 hours
  maximum 1.553298360 = 127^(1/11)
  this n max 127 reps 2
  :J`AgqRGUXHcNCPSPCWy^
  :J`A_zDeOhhcIYGapCXN
  https://hog.grinvin.org/ViewGraphInfo.action?id=30689
  https://hog.grinvin.org/ViewGraphInfo.action?id=30691

  n=11 connected or disconnected
  1018997864 graphs (A000088)
  vpar_INTERNAL_Euler_transform([1, 1, 2, 6, 21, 112, 853, 11117, 261080, 11716571, 1006700565])[11] == 1 018 997 864
  five and six (octahedral)
  https://hog.grinvin.org/ViewGraphInfo.action?id=30693

  --
  n=12 connected   164b graphs, cf 1b of n=11
  164059830476.0/1006700565 * 16 /24  \\ estimate 108 days
  ...

  n=12 disconnected, two octahedral
  :K`?K?oCMpYeiYhsUM`KPBH

*/


#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

int option_verbose = 0;


/* --------------------------------------------------------------------------- */

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))
#define numberof(array)  (sizeof(array)/sizeof((array)[0]))


unsigned long long graphs_count;
double b_max;
int b_max_n;
int b_max_count;

int  max_count;
int  max_reps;
char *max_s6[10];

void
my_geng_outproc (FILE *outfile, graph *g, int n)
{
  const int m = 1;
  graphs_count++;

  if (NAUTYEXTRA_PROGRESS)
    printf("n=%d graphs %Lu  (reps %d)\r",
           n, graphs_count, nautyextra_progress_reps);

  NAUTYEXTRA_TYPE_ULL minimal_domsets_count
    = nautyextra_minimal_domsets_count1 (g,n);
  double b = pow(minimal_domsets_count, 1.0/n);
  if (b>b_max) {
    b_max = b;
    printf ("b new high at %Lu n=%d pow %.9f\n", graphs_count, n, b);
    printf("%s%s", SPARSE6_HEADER, nautyextra_canonical_str(g,m,n,1));
    b_max_n = n;
    b_max_count = minimal_domsets_count;
  }

  if (minimal_domsets_count > max_count) {
    max_count = minimal_domsets_count;
    max_reps = 0;
  }
  if (max_count == minimal_domsets_count) {
    if (max_reps < numberof(max_s6)) {
      if (max_s6[max_reps]) free(max_s6[max_reps]);
      max_s6[max_reps] = strdup(nautyextra_canonical_str(g,m,n,1));
    }
    max_reps++;
  }
}

void
try_octahedral_connect (void)
{
  const int n = 12;
  const int m = 1;
  graph g[n];
  EMPTYGRAPH(g,m,n);

  /*
    #       4 deg=4
    #    /  |   \
    #   5---|---3
    #   | \/ \/ |
    #   | /\ /\ |
    #   0---|---2
    #    \  |  /
    #       1 deg=4
  */
  int x;
  for (x=0; x<=6; x+=6) {
    ADDONEEDGE(g, 0+x,1+x, m);     ADDONEEDGE(g, 0+x,4+x, m);
    ADDONEEDGE(g, 1+x,2+x, m);     ADDONEEDGE(g, 2+x,4+x, m);
    ADDONEEDGE(g, 2+x,3+x, m);
    ADDONEEDGE(g, 3+x,4+x, m);     ADDONEEDGE(g, 5+x,1+x, m);
    ADDONEEDGE(g, 4+x,5+x, m);     ADDONEEDGE(g, 3+x,1+x, m);
    ADDONEEDGE(g, 5+x,0+x, m);

    /* ADDONEEDGE(g, 0+x,2+x, m); */
    /* ADDONEEDGE(g, 5+x,3+x, m); */
  }
  printf ("%d edges\n", nautyextra_num_edges(g,m,n));
  printf ("%s", ntos6(g,m,6));
  printf ("%s", nautyextra_canonical_str(g,m,6, 1));

  /* 6 one side, to 6 the other is 6*6 = 36 pairs */

  {
    NAUTYEXTRA_TYPE_ULL minimal_domsets_count
      = nautyextra_minimal_domsets_count1 (g,n);
    printf ("disjoint %lu minimal domsets\n", minimal_domsets_count);
    printf ("%s", nautyextra_canonical_str(g,m,n, 1));  }

  int edges[6];
  int ne;
  for (ne = 1; ne <= numberof(edges); ne++) {
    NAUTYEXTRA_TYPE_ULL b = 0;
    printf ("try %d edges\n", ne);
    memset(edges, '\0', sizeof(edges));
    for (;;) {
      graph ge[numberof(g)];
      memcpy(ge,g,sizeof(g));
      int e;
      for (e = 0;  e < ne; e++) {
        int edge = edges[e];
        int from = edge%6;
        int to   = (edge/6) + 6;
        assert(from < 6);
        assert(6 <= to && to < 12);
        if (ISELEMENT(GRAPHROW(ge,m,from),to)) goto NEXT_EDGES;
        ADDONEEDGE(ge, from,to, m);
        /* printf ("edge %d from %d to %d = %lu\n", */
        /*         edge, from, to, nautyextra_minimal_domsets_count1(ge,n)); */
      }
      NAUTYEXTRA_TYPE_ULL minimal_domsets_count
        = nautyextra_minimal_domsets_count1 (ge,n);
      if (minimal_domsets_count > b) {
        printf ("new high sets %lu in %d edges\n",
                minimal_domsets_count, nautyextra_num_edges(ge,m,n));
        printf ("%s", nautyextra_canonical_str(ge,m,n, 1));
        b = minimal_domsets_count;
      }

    NEXT_EDGES: ;
      int pos = 0;
      for (;;) {
        if (++edges[pos] < 6*6) break;
        edges[pos] = 0;
        if (++pos >= ne) goto NEXT_NE;
      }
    }
  NEXT_NE:
    ;
  }
}

int
main (void)
{
  nautyextra_want_progress = 1;
  setbuf(stdout,NULL);

  /* option_verbose = 1; */

  if (0) {
    try_octahedral_connect ();
    return 0;
  }

  int num_vertices;
  for (num_vertices = 11;
       num_vertices <= 11;
       num_vertices++) {
    if (option_verbose >= 1) {
      printf ("n=%d geng\n", num_vertices);
    }
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    graphs_count = 0;
    max_count = 0;
    max_reps = 0;
    /* b_max = 0.0; */
    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = "geng";
    geng_argv[geng_argc++] = "-q";     /* quiet */
    geng_argv[geng_argc++] = "-c";     /* connected graphs */
    /* geng_argv[geng_argc++] = "-d2";    /\* minimum degree *\/ */
    geng_argv[geng_argc++] = num_vertices_str;
    geng_argv[geng_argc] = NULL;
    my_geng_main(geng_argc,geng_argv);

    printf ("n=%d count %Lu graphs\n", num_vertices, graphs_count);
    printf ("  maximum %.9f = %d^(1/%d)\n",
            b_max, b_max_count, b_max_n);
    printf ("  this n max %d reps %d\n",
            max_count, max_reps);
    int i;
    for(i = 0; i < max_reps; i++)
      printf("%s", max_s6[i]);
    printf ("\n");
  }
}
