/* Copyright 2017, 2018, 2022 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Print some sequences of values.
 */


#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

int option_verbose = 0;

const int tree_min_vertices = 2;
const int tree_max_vertices = 10;   /* 23 = 14,828,074 */

const int graph_min_vertices = 2;
const int graph_max_vertices = 8;   /* 10 a bit slow (11,716,571 graphs) */

/*
  GP-DEFINE  default(strictargs,1);
  GP-DEFINE  default(recover,0);
*/

/* --------------------------------------------------------------------------- */

/* ss is a setword variable "setword ss".
   n is an int number of bits to have in ss.
   Loop over all sets of n bits in ss.
   Evaluate "body" with successive ss.
   body can goto or return etc to abort the loop.
*/
#define NAUTYEXTRA_SET1_LOOP(ss,n,body)                 \
  do {                                                  \
    assert (n >= 1);                                    \
    if (! (SETWORDSNEEDED(n) == 1)) abort();            \
    setword __NAUTYEXTRA_SET_LOOP__inc = BITT[n-1];     \
    (ss) = 0;                                           \
    do {                                                \
      body;                                             \
    } while ((ss) += __NAUTYEXTRA_SET_LOOP__inc);       \
  } while (0)

/* Return TRUE if g has an independent set with its complement also
   independent.  Return FALSE if not.
 */
int
has_indcomp (const graph *g, int m, int n)
{
  /* printf ("has_indcomp() n=%d\n", n); */
  assert (n >= 0);
  if (n == 0) return 1;   /* n=0 empty set is only indset */
  if (n > WORDSIZE) {
    fprintf(stderr, "nautyextra_indomsets_count_sg() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  setword ss;
  NAUTYEXTRA_SET1_LOOP(ss,n,
                       if (nautyextra_is_indset(g,m,n,&ss)) {
                         setword ss_comp = 0;
                         int i;
                         for (i=0;i<n;i++)
                           if (! ISELEMENT(&ss,i))
                             ADDELEMENT(&ss_comp,i);
                         /* printf ("%X %X\n", ss, ss_comp); */
                         if (nautyextra_is_indset(g,m,n,&ss_comp)) {
                           /* printf (" found\n"); */
                           return TRUE;
                         }
                       }
                       );
  return FALSE;
}

int
compare_int (const void *p, const void *q)
{
  int pp = *(int*)p;
  int qq = *(int*)q;
  int d = pp - qq;
  return (d<0 ? -1 : d==0 ? 0 : 1);
}

static const int Ramsey_number_table[5][10] =
  /* l= 0  1  2  3  4  5  6  7  8  9  */
  { {   0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, /* k=0 */
    {   0, 1, 1, 1, 1, 1, 1, 1, 1, 1 }, /* k=1 */
    {   0, 1, 2, 3, 4, 5, 6, 7, 8, 9 }, /* k=2 */
    {   0, 1, 3, 6, 9,14,18,23,28,36 }, /* k=3 */
    {   0, 1, 4, 9,18,25,-1,-1,-1,-1 }  /* k=4 */
  };
const int Ramsey_max_k = 4;
const int Ramsey_max_l = 8;
int
Ramsey_number (int k, int l)
{
  /* R(k,l) = R(l,k) by graph complement */
  if (k>l) { int t=k;k=l;l=t; }
  return (k>=0 && k<=4 && l>=0 && l<=9 ? Ramsey_number_table[k][l] : -1);
}

/* --------------------------------------------------------------------------- */

SG_DECL(sg);
SG_DECL(sg2);
int trees_count[64];
int trees_total_vertices[64];
int trees_total_leaf[64];
int trees_total_nonleaf[64];
int trees_count_duplicated_leaf[64];
int trees_count_noduplicated_leaf[64];
int trees_duplicated_leaf_max_indnum[64];
int trees_duplicated_leaf_min_indnum[64];
int trees_duplicated_leaf_max_indnum_count[64];
int trees_duplicated_leaf_min_indnum_count[64];
int trees_duplicated_leaf_indnumCount_eq_1[64];
int trees_duplicated_leaf_indnumCount_eq_2[64];
#define TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_LIMIT  64
#define TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_M  SETWORDSNEEDED(TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_LIMIT)
set trees_duplicated_leaf_indnum_occurring[64][TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_M];
unsigned long long trees_duplicated_leaf_total_indnum[64];
int trees_duplicated_leaf_half_indnum[64];
int trees_noduplicated_leaf_max_indnum[64];
int trees_noduplicated_leaf_min_indnum[64];
int trees_noduplicated_leaf_max_indnum_count[64];
int trees_noduplicated_leaf_min_indnum_count[64];
unsigned long long trees_noduplicated_leaf_total_indnum[64];
int trees_count_deg2with[64];
int trees_count_deg2free[64];
int trees_count_deg3with[64];
int trees_count_deg3free[64];
int tree_degrees_distinct[64];
int tree_degrees_distinct_sorted[64];
int trees_count_withclaw[64];
int trees_count_clawfree[64];
int trees_total_claws[64];
int trees_claws_max[64];
int trees_claws_max_count[64];
int trees_depths_eq_vpar[64];
unsigned long long trees_balance_total[64];
int trees_balance_max[64];
int trees_balance_min[64];
int trees_balance_max_count[64];
int trees_balance_min_count[64];
int trees_balance_zero[64];
int trees_balanced_indnum[64];
int trees_total_diameter[64];
int trees_total_diameter_endpoints[64];
int trees_diameter_unique_endpoints[64];
int trees_total_non_diameter_endpoints[64];
int trees_total_diameter_vertices[64];
int trees_total_non_diameter_vertices[64];
int trees_count_unicentral[64];
int trees_count_bicentral[64];
int trees_total_eccentricities[64];
unsigned long long trees_total_centroid_weight[64];
int trees_count_unicentroid[64];
int trees_count_bicentroid[64];
int trees_count_unicentral_unicentroid[64];
int trees_count_unicentral_bicentroid[64];
int trees_count_bicentral_unicentroid[64];
int trees_count_bicentral_bicentroid[64];
int trees_count_centre_centroid[64];
int trees_count_unicentre_centroid[64];
int trees_count_bicentre_centroid[64];
int trees_count_equal_centre_centroid[64];
int trees_count_equal_unicentral_unicentroid[64];
int trees_count_equal_bicentral_bicentroid[64];
int trees_count_disjoint_centre_centroid[64];
int trees_count_disjoint_unicentral_unicentroid[64];
int trees_count_disjoint_unicentral_bicentroid[64];
int trees_count_disjoint_bicentral_unicentroid[64];
int trees_count_disjoint_bicentral_bicentroid[64];
unsigned long long trees_cc_dist_total[64];
int trees_cc_dist_max[64];
int trees_cc_dist_max_count[64];
int trees_max_bandwidth[64];
int trees_min_bandwidth[64];
int trees_max_bandwidth_count[64];
int trees_min_bandwidth_count[64];
int trees_total_bandwidth[64];
unsigned long long trees_max_indsets[64];
unsigned long long trees_min_indsets[64];
int                trees_max_indsets_count[64];
int                trees_min_indsets_count[64];
unsigned long long trees_total_indsets[64];
int trees_max_indnum[64];
int trees_min_indnum[64];
int trees_max_indnum_count[64];
int trees_min_indnum_count[64];
int trees_indnumCount_eq_1[64];
int trees_indnumCount_eq_2[64];
#define TREES_INDNUM_OCCURRING_LIMIT  64
#define TREES_INDNUM_OCCURRING_M  SETWORDSNEEDED(TREES_INDNUM_OCCURRING_LIMIT)
set trees_indnum_occurring[64][TREES_INDNUM_OCCURRING_M];
unsigned long long trees_total_indnum[64];
int trees_half_indnum[64];
int trees_perfect_matching[64];
int trees_near_perfect_matching[64];
int trees_near_perfect_matching_total[64];
int trees_min_matchnum_ways[64];
int trees_min_matchnum_ways_count[64];
int trees_max_matchnum_ways[64];
int trees_max_matchnum_ways_count[64];
unsigned long long trees_total_matchnum[64];
unsigned long long trees_max_domsets[64];
unsigned long long trees_min_domsets[64];
int                trees_max_domsets_count[64];
int                trees_min_domsets_count[64];
unsigned long long trees_total_domsets[64];
unsigned long long trees_max_minimal_domsets[64];
unsigned long long trees_min_minimal_domsets[64];
int                trees_max_minimal_domsets_count[64];
int                trees_min_minimal_domsets_count[64];
unsigned long long trees_total_minimal_domsets[64];
double             trees_minimal_domsets_frac_min[64];
unsigned long long trees_minimal_domsets_frac_min_num[64];
unsigned long long trees_minimal_domsets_frac_min_den[64];
int                trees_minimal_domsets_frac_min_count[64];
double             trees_minimal_domsets_frac_max[64];
unsigned long long trees_minimal_domsets_frac_max_num[64];
unsigned long long trees_minimal_domsets_frac_max_den[64];
int                trees_minimal_domsets_frac_max_count[64];
int trees_max_domnum[64];
int trees_min_domnum[64];
int trees_max_domnum_count[64];
int trees_min_domnum_count[64];
int trees_min_domnum_ways[64];
int trees_min_domnum_ways_count[64];
int trees_max_domnum_ways[64];
int trees_max_domnum_ways_count[64];
unsigned long long trees_total_domnum[64];
#define TREES_DOMNUM_OCCURRING_LIMIT  64
#define TREES_DOMNUM_OCCURRING_M  SETWORDSNEEDED(TREES_DOMNUM_OCCURRING_LIMIT)
set trees_domnum_occurring[64][TREES_DOMNUM_OCCURRING_M];
int trees_max_disjoint_domnum[64];
int trees_min_disjoint_domnum[64];
int trees_max_disjoint_domnum_count[64];
int trees_min_disjoint_domnum_count[64];
int trees_min_disjoint_domnum_ways[64];
int trees_min_disjoint_domnum_ways_count[64];
int trees_max_disjoint_domnum_ways[64];
int trees_max_disjoint_domnum_ways_count[64];
unsigned long long trees_total_disjoint_domnum[64];
#define TREES_DISJOINT_DOMNUM_OCCURRING_LIMIT  64
#define TREES_DISJOINT_DOMNUM_OCCURRING_M  SETWORDSNEEDED(TREES_DISJOINT_DOMNUM_OCCURRING_LIMIT)
set trees_disjoint_domnum_occurring[64][TREES_DISJOINT_DOMNUM_OCCURRING_M];
int trees_max_totdomnum[64];
int trees_min_totdomnum[64];
int trees_max_totdomnum_count[64];
int trees_min_totdomnum_count[64];
unsigned long long trees_total_totdomnum[64];
#define TREES_TOTDOMNUM_OCCURRING_LIMIT  64
#define TREES_TOTDOMNUM_OCCURRING_M  SETWORDSNEEDED(TREES_TOTDOMNUM_OCCURRING_LIMIT)
set trees_totdomnum_occurring[64][TREES_TOTDOMNUM_OCCURRING_M];
int trees_domsets_minimal_all_minimum_count[64];
unsigned long long trees_total_indomsets[64];
unsigned long long trees_max_indomsets[64];
unsigned long long trees_min_indomsets[64];
int                trees_max_indomsets_count[64];
int                trees_min_indomsets_count[64];
unsigned long long trees_total_indomsets[64];
double             trees_indomsets_frac_min[64];
unsigned long long trees_indomsets_frac_min_num[64];
unsigned long long trees_indomsets_frac_min_den[64];
int                trees_indomsets_frac_min_count[64];
double             trees_indomsets_frac_max[64];
unsigned long long trees_indomsets_frac_max_num[64];
unsigned long long trees_indomsets_frac_max_den[64];
int                trees_indomsets_frac_max_count[64];
int trees_max_indomnum[64];
int trees_min_indomnum[64];
int trees_max_indomnum_count[64];
int trees_min_indomnum_count[64];
unsigned long long trees_total_indomnum[64];
unsigned long long trees_total_leaf_length[64];
#define TREES_LEAF_LENGTH_HISTOGRAM_MAX 8
int trees_leaf_length_eq[64][TREES_LEAF_LENGTH_HISTOGRAM_MAX+1];
int trees_leaf_length_lt[64][TREES_LEAF_LENGTH_HISTOGRAM_MAX+1];
int trees_leaf_length_gt[64][TREES_LEAF_LENGTH_HISTOGRAM_MAX+1];
int trees_asymmetric[64];
int trees_similar_total[64];
int trees_similar_is_unique[64];
int trees_count_stable[64];
int trees_count_semistable[64];
int trees_count_completely_semistable[64];
int trees_total_semistable_vertices[64];

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))
#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

void
my_gentreeg_outproc (FILE *outfile, int *vpar, int n)
{
  if (n==2 && trees_count[2]) return;  /* don't duplicate n=2 single tree */

  /* count_total++; */
  trees_count[n]++;
  trees_total_vertices[n] += n;

  if (option_verbose >= 2) {
    printf("tree %p  n=%d\n", vpar,n);
    nautyextra_print_vpar(vpar,n);
  }

  DYNALLSTAT(graph,g,g_sz);
  int m = SETWORDSNEEDED(n);
  DYNALLOC2(graph,g,g_sz,m,n,"malloc");
  nautyextra_vpar_to_graph(vpar,g,m,n, 0);

  trees_count_stable[n] += nautyextra_is_stable(g,m,n);

  {
    int semistable = 0;
    int completely_semistable = 1;
    int v;
    for (v=0; v<n; v++) {
      if (nautyextra_is_semistable_at_vertex (g,m,n,v)) {
        semistable = 1;
        trees_total_semistable_vertices[n]++;
      } else {
        completely_semistable = 0;
      }
    }
    trees_count_semistable[n] += semistable;
    trees_count_completely_semistable[n] += completely_semistable;
  }

  if (n <= 8) {
    int bandwidth = nautyextra_bandwidth_vpar(vpar,n);
    if (trees_min_bandwidth[n] == 0 || bandwidth < trees_min_bandwidth[n]) {
      trees_min_bandwidth[n] = bandwidth;
      trees_min_bandwidth_count[n] = 0;
    }
    if (bandwidth == trees_min_bandwidth[n])
      trees_min_bandwidth_count[n]++;
    if (trees_max_bandwidth[n] == 0 || bandwidth > trees_max_bandwidth[n]) {
      trees_max_bandwidth[n] = bandwidth;
      trees_max_bandwidth_count[n] = 0;
    }
    if (bandwidth == trees_max_bandwidth[n])
      trees_max_bandwidth_count[n]++;
    trees_total_bandwidth[n] += bandwidth;

    if (0 && n==7 && bandwidth==3) {
      printf("bandwidth %d  ", bandwidth);
      nautyextra_print_vpar(vpar,n);
    }
  }

  int count_deg3 = 0;
  {
    int degrees[64];
    assert(n <= 64);
    nautyextra_degrees_array_vpar(vpar,n,degrees);
    int count_deg2 = 0;
    int v;
    for (v = 1; v <= n; v++) {
      int d = degrees[v];
      count_deg2 += (d==2);
      count_deg3 += (d==3);
    }
    trees_count_deg2free[n] += (count_deg2 == 0);
    trees_count_deg2with[n] += (count_deg2 != 0);
    trees_count_deg3free[n] += (count_deg3 == 0);
    trees_count_deg3with[n] += (count_deg3 != 0);

#define TREE_DEGREES_SEEN_MAX_N   16
    if (n <= TREE_DEGREES_SEEN_MAX_N) {
      static int seen_array[50000][TREE_DEGREES_SEEN_MAX_N + 1];
      int *a = tree_degrees_distinct;
      int i;
      for (i=0; i<a[n]; i++)
        if (memcmp(seen_array[i], degrees, sizeof(int)*(n+1)) == 0)
          goto TREE_DEGREES_SEEN;
      assert(a[n] < numberof(seen_array));
      memcpy (seen_array[a[n]], degrees, sizeof(int)*(n+1));
      a[n]++;
    TREE_DEGREES_SEEN:
      ;
    }

    if (n <= TREE_DEGREES_SEEN_MAX_N) {
      static int seen_array[5000][TREE_DEGREES_SEEN_MAX_N + 1];
      qsort(degrees,n,sizeof(int),compare_int);
      int *a = tree_degrees_distinct_sorted;
      int i;
      for (i=0; i<a[n]; i++)
        if (memcmp(seen_array[i], degrees, sizeof(int)*(n+1)) == 0)
          goto TREE_DEGREES_SORTED_SEEN;
      assert(a[n] < numberof(seen_array));
      memcpy (seen_array[a[n]], degrees, sizeof(int)*(n+1));
      a[n]++;
    TREE_DEGREES_SORTED_SEEN:
      ;
    }
  }
  {
    int count_leaf = nautyextra_leaf_count_vpar(vpar,n);
    int count_nonleaf = nautyextra_nonleaf_count_vpar(vpar,n);
    trees_total_leaf[n] += count_leaf;
    trees_total_nonleaf[n] += count_nonleaf;
    if (option_verbose >= 2) {
      printf("  leaf %d nonleaf %d\n", count_leaf, count_nonleaf);
    }
  }

  int duplicated_leaf = nautyextra_any_duplicated_leaf_vpar(vpar,n);
  {
    trees_count_duplicated_leaf[n]   += duplicated_leaf;
    trees_count_noduplicated_leaf[n] += !duplicated_leaf;
  }

  {
    int claws = nautyextra_claw_count_vpar(vpar,n);
    assert (claws >= count_deg3);
    trees_total_claws[n] += claws;
    trees_count_withclaw[n] += (claws != 0);
    trees_count_clawfree[n] += (claws == 0);

    if (trees_claws_max[n] == 0 || claws > trees_claws_max[n]) {
      trees_claws_max[n] = claws;
      trees_claws_max_count[n] = 0;
    }
    if (claws == trees_claws_max[n])
      trees_claws_max_count[n]++;
  }

  {
    unsigned long long count_indsets = nautyextra_indsets_count_vpar(vpar,n);
    trees_total_indsets[n] += count_indsets;
    if (trees_min_indsets[n] == 0 || count_indsets < trees_min_indsets[n]) {
      trees_min_indsets[n] = count_indsets;
      trees_min_indsets_count[n] = 0;
    }
    if (count_indsets == trees_min_indsets[n])
      trees_min_indsets_count[n]++;
    if (trees_max_indsets[n] == 0 || count_indsets > trees_max_indsets[n]) {
      trees_max_indsets[n] = count_indsets;
      trees_max_indsets_count[n] = 0;
    }
    if (count_indsets == trees_max_indsets[n])
      trees_max_indsets_count[n]++;
  }

  {
    /* 0--1--2--3--4  even diameter = unicentral */
    struct nautyextra_diameter_info  diameter_info;
    nautyextra_diameter_info_vpar(vpar,n, &diameter_info);
    int diameter = diameter_info.diameter;
    trees_total_diameter[n] += diameter;
    trees_total_diameter_endpoints[n]     += diameter_info.num_endpoints;
    trees_diameter_unique_endpoints[n]    += (diameter_info.num_endpoints <= 2);
    trees_total_non_diameter_endpoints[n] += n - diameter_info.num_endpoints;
    trees_total_diameter_vertices[n]     += diameter_info.num_vertices;
    trees_total_non_diameter_vertices[n] += n - diameter_info.num_vertices;
    int bicentral = diameter & 1;
    int unicentral = 1-bicentral;
    trees_count_unicentral[n] += unicentral;
    trees_count_bicentral[n] += bicentral;

    struct nautyextra_centroid_info  centroid_info;
    nautyextra_centroid_info_vpar(vpar,n, &centroid_info);
    trees_total_centroid_weight[n] += centroid_info.weight;
    int unicentroid = (centroid_info.v1 == centroid_info.v2);
    int bicentroid = 1-unicentroid;
    trees_count_unicentroid[n] += unicentroid;
    trees_count_bicentroid[n]  += bicentroid;

    trees_count_unicentral_unicentroid[n] += (unicentral && unicentroid);
    trees_count_unicentral_bicentroid[n]  += (unicentral && bicentroid);
    trees_count_bicentral_unicentroid[n]  += (bicentral  && unicentroid);
    trees_count_bicentral_bicentroid[n]   += (bicentral  && bicentroid);

    /* if (n==6 && !unicentroid) */
    /*   nautyextra_view_vpar(vpar,n); */

    int diameter_c1, diameter_c2;
    nautyextra_path_centres_vpar (vpar,
                                  diameter_info.end_v1, diameter_info.end_v2,
                                  diameter_info.end_ancestor,
                                  &diameter_c1, &diameter_c2);
    if (diameter_c1 > diameter_c2) {
      int t = diameter_c1; diameter_c1 = diameter_c2; diameter_c2 = t;
    }

    int centroid_v1 = centroid_info.v1;
    int centroid_v2 = centroid_info.v2;
    if (centroid_v1 > centroid_v2) {
      int t = centroid_v1; centroid_v1 = centroid_v2; centroid_v2 = t;
    }
    int centre_centroid = (diameter_c1 == centroid_v1
                           && diameter_c2 == centroid_v2);
    trees_count_centre_centroid[n] += centre_centroid;
    trees_count_unicentre_centroid[n] += unicentral && centre_centroid;
    trees_count_bicentre_centroid[n] += bicentral && centre_centroid;

    int equal_centre_centroid = (diameter_c1 == centroid_v1
                                 && diameter_c2 == centroid_v2);
    trees_count_equal_centre_centroid[n] += equal_centre_centroid;
    trees_count_equal_unicentral_unicentroid[n]
      += (equal_centre_centroid && unicentral && unicentroid);
    trees_count_equal_bicentral_bicentroid[n]
      += (equal_centre_centroid && bicentral && bicentroid);

    int disjoint_centre_centroid = (diameter_c1 != centroid_v1
                                    && diameter_c1 != centroid_v2
                                    && diameter_c2 != centroid_v1
                                    && diameter_c2 != centroid_v2);
    trees_count_disjoint_centre_centroid[n] += disjoint_centre_centroid;
    trees_count_disjoint_unicentral_unicentroid[n]
      += (disjoint_centre_centroid && unicentral && unicentroid);
    trees_count_disjoint_unicentral_bicentroid[n]
      += (disjoint_centre_centroid && unicentral && bicentroid);
    trees_count_disjoint_bicentral_unicentroid[n]
      += (disjoint_centre_centroid && bicentral && unicentroid);
    trees_count_disjoint_bicentral_bicentroid[n]
      += (disjoint_centre_centroid && bicentral && bicentroid);

    if (0 && n==12 && disjoint_centre_centroid && bicentral && bicentroid) {
      printf ("n=%d  centre %d,%d  centroid %d,%d\n",
              n, diameter_c1,diameter_c2,
              centroid_v1,centroid_v2);
      nautyextra_view_vpar(vpar,n);
    }

    int cc_dist1 = nautyextra_path_length_vpar(vpar, diameter_c1, centroid_v1);
    int cc_dist2 = nautyextra_path_length_vpar(vpar, diameter_c1, centroid_v2);
    int cc_dist3 = nautyextra_path_length_vpar(vpar, diameter_c2, centroid_v1);
    int cc_dist4 = nautyextra_path_length_vpar(vpar, diameter_c2, centroid_v2);
    int cc_dist = MIN(cc_dist1,cc_dist2);
    cc_dist = MIN(cc_dist, cc_dist3);
    cc_dist = MIN(cc_dist, cc_dist4);

    trees_cc_dist_total[n] += cc_dist;
    if (trees_cc_dist_max[n] == 0 || cc_dist > trees_cc_dist_max[n]) {
      trees_cc_dist_max[n] = cc_dist;
      trees_cc_dist_max_count[n] = 0;
    }
    if (cc_dist == trees_cc_dist_max[n])
      trees_cc_dist_max_count[n]++;

    if (0 && (n==10 && cc_dist==1))
      nautyextra_view_vpar(vpar,n);
  }

  {
    /* eccentricities */
    int eccs[tree_max_vertices+1];
    nautyextra_eccentricities_vpar(vpar,n, eccs);
    int total_eccentricities = 0;
    int v;
    for (v = 1; v <= n; v++) total_eccentricities += eccs[v];
    trees_total_eccentricities[n] += total_eccentricities;
  }

  int indnum = nautyextra_indnum_vpar(vpar,n);
  {
    /* printf("n=%d  indnum = %d\n", n, indnum); */
    trees_total_indnum[n] += indnum;
    if (trees_min_indnum[n] == 0 || indnum < trees_min_indnum[n]) {
      trees_min_indnum[n] = indnum;
      trees_min_indnum_count[n] = 0;
    }
    if (indnum == trees_min_indnum[n])
      trees_min_indnum_count[n]++;
    if (trees_max_indnum[n] == 0 || indnum > trees_max_indnum[n]) {
      trees_max_indnum[n] = indnum;
      trees_max_indnum_count[n] = 0;
    }
    if (indnum == trees_max_indnum[n])
      trees_max_indnum_count[n]++;

    if (2*indnum == n)
      trees_half_indnum[n]++;

    if (indnum >= TREES_INDNUM_OCCURRING_LIMIT) abort();
    ADDELEMENT(trees_indnum_occurring[n],indnum);

    /* if (n==9 && indnum == 5) { */
    /*   nautyextra_vpar_to_sg(vpar,n,&sg); */
    /*   nautyextra_view_sg(&sg); */
    /* } */
  }
  if (duplicated_leaf) {
    trees_duplicated_leaf_total_indnum[n] += indnum;
    if (trees_duplicated_leaf_min_indnum[n] == 0 || indnum < trees_duplicated_leaf_min_indnum[n]) {
      trees_duplicated_leaf_min_indnum[n] = indnum;
      trees_duplicated_leaf_min_indnum_count[n] = 0;
    }
    if (indnum == trees_duplicated_leaf_min_indnum[n])
      trees_duplicated_leaf_min_indnum_count[n]++;
    if (trees_duplicated_leaf_max_indnum[n] == 0 || indnum > trees_duplicated_leaf_max_indnum[n]) {
      trees_duplicated_leaf_max_indnum[n] = indnum;
      trees_duplicated_leaf_max_indnum_count[n] = 0;
    }
    if (indnum == trees_duplicated_leaf_max_indnum[n])
      trees_duplicated_leaf_max_indnum_count[n]++;

    if (2*indnum == n)
      trees_duplicated_leaf_half_indnum[n]++;

    if (indnum >= TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_LIMIT) abort();
    ADDELEMENT(trees_duplicated_leaf_indnum_occurring[n],indnum);

    /* if (n==9 && indnum == 5) { */
    /*   nautyextra_vpar_to_sg(vpar,n,&sg); */
    /*   nautyextra_view_sg(&sg); */
    /* } */
  } else {
    trees_noduplicated_leaf_total_indnum[n] += indnum;
    if (trees_noduplicated_leaf_min_indnum[n] == 0 || indnum < trees_noduplicated_leaf_min_indnum[n]) {
      trees_noduplicated_leaf_min_indnum[n] = indnum;
      trees_noduplicated_leaf_min_indnum_count[n] = 0;
    }
    if (indnum == trees_noduplicated_leaf_min_indnum[n])
      trees_noduplicated_leaf_min_indnum_count[n]++;
    if (trees_noduplicated_leaf_max_indnum[n] == 0 || indnum > trees_noduplicated_leaf_max_indnum[n]) {
      trees_noduplicated_leaf_max_indnum[n] = indnum;
      trees_noduplicated_leaf_max_indnum_count[n] = 0;
    }
    if (indnum == trees_noduplicated_leaf_max_indnum[n])
      trees_noduplicated_leaf_max_indnum_count[n]++;

    /* Jou and Lin n=11 example
       GP-Test  my(n=11); floor((2*n-1)/3) == 7
    */
    if (0 && n==11 && indnum==7) {
      nautyextra_view_synchronous = 0;
      nautyextra_view_vpar(vpar,n);
    }
  }

  if (n <= 14) {
    struct nautyextra_similar_info info;
    nautyextra_similar(g,m,n, &info);

    int similar_pairs
      = nautyextra_num_edges(info.similar,m,n);
    trees_asymmetric[n] += (similar_pairs==0);

    int similar_is_unique = (similar_pairs == 1);
    trees_similar_is_unique[n] += similar_is_unique;
    trees_similar_total[n] += similar_pairs;

    if (0 && similar_pairs==0 && n<=9) {
      printf("n=%d asymmetric  %s", n, ntog6((graph*)g,m,n));
      nautyextra_view(g,m,n);
    }
    if (0 && similar_pairs==0 && n<=11) {
      printf("%s", ntog6((graph*)g,m,n));
    }

    /* if (graphs_pseudosimilar_clique_max[n] == 0 */
    /*     || info.pseudosimilar_clique_size > graphs_pseudosimilar_clique_max[n]){ */
    /*   graphs_pseudosimilar_clique_max[n] = info.pseudosimilar_clique_size; */
    /*   graphs_pseudosimilar_clique_max_count[n] = 0; */
    /* } */
    /* if (info.pseudosimilar_clique_size == graphs_pseudosimilar_clique_max[n]) */
    /*   graphs_pseudosimilar_clique_max_count[n]++; */
    /*  */
    /* if (info.pseudosimilar_clique_size >= 2) { */
    /*   if (graphs_pseudosimilar_min_edges[n] == 0 */
    /*       || num_edges < graphs_pseudosimilar_min_edges[n]){ */
    /*     graphs_pseudosimilar_min_edges[n] = num_edges; */
    /*     graphs_pseudosimilar_min_edges_count[n] = 0; */
    /*   } */
    /*   if (num_edges == graphs_pseudosimilar_min_edges[n]) */
    /*     graphs_pseudosimilar_min_edges_count[n]++; */
    /* } */
    /*  */
    /* int removal_similar_clique_size */
    /*   = nautyextra_clique_number(info.removal_similar,m,n); */
    /* if (graphs_removal_similar_clique_max[n] == 0 */
    /*     || removal_similar_clique_size > graphs_removal_similar_clique_max[n]){ */
    /*   graphs_removal_similar_clique_max[n] = removal_similar_clique_size; */
    /*   graphs_removal_similar_clique_max_count[n] = 0; */
    /* } */
    /* if (removal_similar_clique_size == graphs_removal_similar_clique_max[n]) */
    /*   graphs_removal_similar_clique_max_count[n]++; */
    /*  */
    /* int removal_similar_pairs */
    /*   = nautyextra_num_edges(info.removal_similar,m,n); */
    /* int removal_similar_is_unique = (removal_similar_pairs == 1); */
    /* graphs_removal_similar_is_unique[n] += removal_similar_is_unique; */
    /* graphs_removal_similar_total[n] += removal_similar_pairs; */
    /* if (0 && n==5 && removal_similar_is_unique) { */
    /*   nautyextra_view(g,m,n); */
    /* } */
    /*  */
    /* if (graphs_orbit_max[n] == 0 || info.orbit_max > graphs_orbit_max[n]){ */
    /*   graphs_orbit_max[n] = info.orbit_max; */
    /*   graphs_orbit_max_count[n] = 0; */
    /* } */
    /* if (info.orbit_max == graphs_orbit_max[n]) */
    /*   graphs_orbit_max_count[n]++; */
    /* graphs_numorbits_total[n] += info.numorbits; */
    /*  */
    /* int is_asymmetric = (info.numorbits == n); */
    /* graphs_asymmetric_count[n] += is_asymmetric; */
    /*  */
    /* if (0 && n==6 && is_asymmetric) { */
    /*   printf("%s", ntog6((graph*)g,m,n)); */
    /*   int degrees_sorted[graph_max_vertices]; */
    /*   memcpy(degrees_sorted, degrees, sizeof(degrees)); */
    /*   qsort(degrees_sorted,n,sizeof(degrees_sorted[0]),compare_int); */
    /*   printf("  degrees "); */
    /*   int v; */
    /*   for (v = 0; v < n; v++) printf("%d,",degrees_sorted[v]); */
    /*   printf("\n"); */
    /* } */
  }

  {
    struct nautyextra_indnum_and_count  info;
    nautyextra_indnum_and_count_vpar (vpar,n, &info);
    assert(indnum == info.indnum);
    trees_indnumCount_eq_1[n] += (info.count == 1);
    trees_indnumCount_eq_2[n] += (info.count == 2);

    /* if (n<=6 && info.count == 1) { */
    /*   nautyextra_view_vpar(vpar,n); */
    /* } */
  }

  trees_perfect_matching[n] += nautyextra_has_perfect_matching_vpar(vpar,n);
  {
    struct nautyextra_matchnum_and_count  matchnum_and_count;
    nautyextra_matchnum_and_count_vpar(vpar,n, &matchnum_and_count);
    int matchnum = matchnum_and_count.matchnum;

    if (2*matchnum+1 == n) {
      trees_near_perfect_matching[n]++;
      trees_near_perfect_matching_total[n] += matchnum_and_count.count;
    }

    if (trees_max_matchnum_ways[n] == 0 || matchnum_and_count.count > trees_max_matchnum_ways[n]) {
      trees_max_matchnum_ways[n] = matchnum_and_count.count;
      trees_max_matchnum_ways_count[n] = 0;
    }
    if (matchnum_and_count.count == trees_max_matchnum_ways[n])
      trees_max_matchnum_ways_count[n]++;

    if (trees_min_matchnum_ways[n] == 0 || matchnum_and_count.count < trees_min_matchnum_ways[n]) {
      trees_min_matchnum_ways[n] = matchnum_and_count.count;
      trees_min_matchnum_ways_count[n] = 0;
    }
    if (matchnum_and_count.count == trees_min_matchnum_ways[n])
      trees_min_matchnum_ways_count[n]++;

    trees_total_matchnum[n] += matchnum;

    static const int max_ways[] = {
      0,0, 1,2,3,4,5,8,11,15,21,30,41,56,81,112,153,216 };
    if (0 && n>=4 && matchnum_and_count.count==max_ways[n]) {
      /* printf("n=%d matchnum %d ways %Lu\n", */
      /*        n, matchnum, matchnum_and_count.count); */
      printf("%s", nautyextra_vpar_to_s6(vpar,n));
      /* nautyextra_view_vpar(vpar,n); */
    }
  }

  unsigned long long count_domsets = nautyextra_domsets_count_vpar(vpar,n);
  {
    trees_total_domsets[n] += count_domsets;
    if (trees_min_domsets[n] == 0 || count_domsets < trees_min_domsets[n]) {
      trees_min_domsets[n] = count_domsets;
      trees_min_domsets_count[n] = 0;
    }
    if (count_domsets == trees_min_domsets[n])
      trees_min_domsets_count[n]++;
    if (trees_max_domsets[n] == 0 || count_domsets > trees_max_domsets[n]) {
      trees_max_domsets[n] = count_domsets;
      trees_max_domsets_count[n] = 0;
    }
    if (count_domsets == trees_max_domsets[n])
      trees_max_domsets_count[n]++;

    /*   if (n==4 && count_domsets == 9) { */
    /*     nautyextra_vpar_to_sg(vpar,n,&sg); */
    /*     nautyextra_view_sg(&sg); */
    /*   } */
  }
  unsigned long long count_minimal_domsets
    = nautyextra_minimal_domsets_count_vpar(vpar,n);
  {
    trees_total_minimal_domsets[n] += count_minimal_domsets;
    if (trees_min_minimal_domsets[n] == 0 || count_minimal_domsets < trees_min_minimal_domsets[n]) {
      trees_min_minimal_domsets[n] = count_minimal_domsets;
      trees_min_minimal_domsets_count[n] = 0;
    }
    if (count_minimal_domsets == trees_min_minimal_domsets[n])
      trees_min_minimal_domsets_count[n]++;

    if (trees_max_minimal_domsets[n] == 0 || count_minimal_domsets > trees_max_minimal_domsets[n]) {
      trees_max_minimal_domsets[n] = count_minimal_domsets;
      trees_max_minimal_domsets_count[n] = 0;
    }
    if (count_minimal_domsets == trees_max_minimal_domsets[n])
      trees_max_minimal_domsets_count[n]++;

    /* if (n < 6 && count_minimal_domsets == 2) { */
    /*   nautyextra_vpar_to_sg(vpar,n,&sg); */
    /*   nautyextra_view_sg(&sg); */
    /* } */
  }
  {
    double minimal_domsets_frac = (double) count_minimal_domsets / (double) count_domsets;
    if (trees_minimal_domsets_frac_min_den[n] == 0 || minimal_domsets_frac < trees_minimal_domsets_frac_min[n]) {
      trees_minimal_domsets_frac_min[n] = minimal_domsets_frac;
      trees_minimal_domsets_frac_min_num[n] = count_minimal_domsets;
      trees_minimal_domsets_frac_min_den[n] = count_domsets;
      trees_minimal_domsets_frac_min_count[n] = 0;
    }
    if (minimal_domsets_frac == trees_minimal_domsets_frac_min[n])
      trees_minimal_domsets_frac_min_count[n]++;

    if (trees_minimal_domsets_frac_max_den[n] == 0 || minimal_domsets_frac > trees_minimal_domsets_frac_max[n]) {
      trees_minimal_domsets_frac_max[n] = minimal_domsets_frac;
      trees_minimal_domsets_frac_max_num[n] = count_minimal_domsets;
      trees_minimal_domsets_frac_max_den[n] = count_domsets;
      trees_minimal_domsets_frac_max_count[n] = 0;
    }
    if (minimal_domsets_frac == trees_minimal_domsets_frac_max[n])
      trees_minimal_domsets_frac_max_count[n]++;

    if (0 && ((n==7 && count_minimal_domsets==8 && count_domsets==45)
              || (n==16 && count_minimal_domsets==256 && count_domsets==6561))) {
      printf("greatest minimal/domsets ratio ");
      nautyextra_print_vpar(vpar,n);
    }
  }

  struct nautyextra_domnum_and_count domnum_and_count;
  nautyextra_domnum_and_count_vpar(vpar,n, &domnum_and_count);
  {
    int domnum = domnum_and_count.domnum;
    /* printf("n=%d  domnum = %d\n", n, domnum); */
    trees_total_domnum[n] += domnum;
    if (trees_min_domnum[n] == 0 || domnum < trees_min_domnum[n]) {
      trees_min_domnum[n] = domnum;
      trees_min_domnum_count[n] = 0;
    }
    if (domnum == trees_min_domnum[n])
      trees_min_domnum_count[n]++;

    if (trees_max_domnum[n] == 0 || domnum > trees_max_domnum[n]) {
      trees_max_domnum[n] = domnum;
      trees_max_domnum_count[n] = 0;
    }
    if (domnum == trees_max_domnum[n])
      trees_max_domnum_count[n]++;

    if (trees_max_domnum_ways[n] == 0 || domnum_and_count.count > trees_max_domnum_ways[n]) {
      trees_max_domnum_ways[n] = domnum_and_count.count;
      trees_max_domnum_ways_count[n] = 0;
    }
    if (domnum_and_count.count == trees_max_domnum_ways[n])
      trees_max_domnum_ways_count[n]++;

    if (trees_min_domnum_ways[n] == 0 || domnum_and_count.count < trees_min_domnum_ways[n]) {
      trees_min_domnum_ways[n] = domnum_and_count.count;
      trees_min_domnum_ways_count[n] = 0;
    }
    if (domnum_and_count.count == trees_min_domnum_ways[n])
      trees_min_domnum_ways_count[n]++;

    if (domnum >= TREES_DOMNUM_OCCURRING_LIMIT) abort();
    ADDELEMENT(trees_domnum_occurring[n],domnum);

    trees_domsets_minimal_all_minimum_count[n]
      += (domnum_and_count.count == count_minimal_domsets);

    if (0 && ((n==7 && domnum == 3)
              || (n==8 && domnum == 4))) {
      nautyextra_view_vpar(vpar,n);
    }
    if (0 && n <=8 && domnum_and_count.count == count_minimal_domsets) {
      printf("minimal domsets all minimum: ");
      nautyextra_print_vpar(vpar,n);
      nautyextra_view_vpar(vpar,n);
    }
  }

  {
    struct nautyextra_disjoint_domnum_and_count info;
    nautyextra_disjoint_domnum_and_count_vpar(vpar,n, &info);
    int disjoint_domnum = info.disjoint_domnum;
    {
      trees_total_disjoint_domnum[n] += disjoint_domnum;

      if (trees_min_disjoint_domnum[n] == 0 || disjoint_domnum < trees_min_disjoint_domnum[n]) {
        trees_min_disjoint_domnum[n] = disjoint_domnum;
        trees_min_disjoint_domnum_count[n] = 0;
      }
      if (disjoint_domnum == trees_min_disjoint_domnum[n])
        trees_min_disjoint_domnum_count[n]++;

      if (trees_max_disjoint_domnum[n] == 0 || disjoint_domnum > trees_max_disjoint_domnum[n]) {
        trees_max_disjoint_domnum[n] = disjoint_domnum;
        trees_max_disjoint_domnum_count[n] = 0;
      }
      if (disjoint_domnum == trees_max_disjoint_domnum[n])
        trees_max_disjoint_domnum_count[n]++;

      if (trees_min_disjoint_domnum_ways[n] == 0 || info.count < trees_min_disjoint_domnum_ways[n]) {
        trees_min_disjoint_domnum_ways[n] = info.count;
        trees_min_disjoint_domnum_ways_count[n] = 0;
      }
      if (info.count == trees_min_disjoint_domnum_ways[n])
        trees_min_disjoint_domnum_ways_count[n]++;

      if (trees_max_disjoint_domnum_ways[n] == 0 || info.count > trees_max_disjoint_domnum_ways[n]) {
        trees_max_disjoint_domnum_ways[n] = info.count;
        trees_max_disjoint_domnum_ways_count[n] = 0;
      }
      if (info.count == trees_max_disjoint_domnum_ways[n])
        trees_max_disjoint_domnum_ways_count[n]++;

      if (disjoint_domnum >= TREES_DISJOINT_DOMNUM_OCCURRING_LIMIT) abort();
      ADDELEMENT(trees_disjoint_domnum_occurring[n],disjoint_domnum);

      if (0) {
        static const int disjoint_domnum_max_ways[] = {
          0,0, 1,1,2,2,4,5,8,13,16,33,33,73,73,161,181,337
        };
        if (info.count == disjoint_domnum_max_ways[n]) {
          if(n==-7) nautyextra_view_vpar(vpar,n);
          printf ("n=%d disjoint domnum %d ways %Lu\n",
                  n, disjoint_domnum, info.count);
        }
      }
    }
  }

  {
    int totdomnum = nautyextra_totdomnum_vpar(vpar,n);
    /* printf("n=%d  totdomnum = %d\n", n, totdomnum); */
    trees_total_totdomnum[n] += totdomnum;
    if (trees_min_totdomnum[n] == 0 || totdomnum < trees_min_totdomnum[n]) {
      trees_min_totdomnum[n] = totdomnum;
      trees_min_totdomnum_count[n] = 0;
    }
    if (totdomnum == trees_min_totdomnum[n])
      trees_min_totdomnum_count[n]++;
    if (trees_max_totdomnum[n] == 0 || totdomnum > trees_max_totdomnum[n]) {
      trees_max_totdomnum[n] = totdomnum;
      trees_max_totdomnum_count[n] = 0;
    }
    if (totdomnum == trees_max_totdomnum[n])
      trees_max_totdomnum_count[n]++;

    if (totdomnum >= TREES_TOTDOMNUM_OCCURRING_LIMIT) abort();
    ADDELEMENT(trees_totdomnum_occurring[n],totdomnum);

    /* if ((n==7 && totdomnum == 3) */
    /*     || (n==8 && totdomnum == 4)){ */
    /*   nautyextra_vpar_to_sg(vpar,n,&sg); */
    /*   nautyextra_view_sg(&sg); */
    /* } */
  }

  unsigned long long count_indomsets
    = nautyextra_indomsets_count_vpar(vpar,n);
  {
    trees_total_indomsets[n] += count_indomsets;
    if (trees_min_indomsets[n] == 0 || count_indomsets < trees_min_indomsets[n]) {
      trees_min_indomsets[n] = count_indomsets;
      trees_min_indomsets_count[n] = 0;
    }
    if (count_indomsets == trees_min_indomsets[n])
      trees_min_indomsets_count[n]++;
    if (trees_max_indomsets[n] == 0 || count_indomsets > trees_max_indomsets[n]) {
      trees_max_indomsets[n] = count_indomsets;
      trees_max_indomsets_count[n] = 0;
    }
    if (count_indomsets == trees_max_indomsets[n])
      trees_max_indomsets_count[n]++;

    if (0 && count_indomsets == 32 && n==11) {
      nautyextra_view_vpar(vpar,n);
    }
    if (0 && count_indomsets == 17 && n==10) {
      nautyextra_view_vpar(vpar,n);
    }
  }

  {
    double indomsets_frac = (double) count_indomsets / (double) count_domsets;
    if (trees_indomsets_frac_min_den[n] == 0 || indomsets_frac < trees_indomsets_frac_min[n]) {
      trees_indomsets_frac_min[n] = indomsets_frac;
      trees_indomsets_frac_min_num[n] = count_indomsets;
      trees_indomsets_frac_min_den[n] = count_domsets;
      trees_indomsets_frac_min_count[n] = 0;
    }
    if (indomsets_frac == trees_indomsets_frac_min[n])
      trees_indomsets_frac_min_count[n]++;

    if (trees_indomsets_frac_max_den[n] == 0 || indomsets_frac > trees_indomsets_frac_max[n]) {
      trees_indomsets_frac_max[n] = indomsets_frac;
      trees_indomsets_frac_max_num[n] = count_indomsets;
      trees_indomsets_frac_max_den[n] = count_domsets;
      trees_indomsets_frac_max_count[n] = 0;
    }
    if (indomsets_frac == trees_indomsets_frac_max[n])
      trees_indomsets_frac_max_count[n]++;

    if (0 && n==12 && count_indomsets == 33 && count_domsets == 729) {
      nautyextra_print_vpar(vpar,n);
    }
  }

  {
    struct nautyextra_indomnum_and_count info;
    nautyextra_indomnum_and_count_vpar(vpar,n,&info);
    int indomnum = info.indomnum;
    trees_total_indomnum[n] += indomnum;
    if (trees_min_indomnum[n] == 0 || indomnum < trees_min_indomnum[n]) {
      trees_min_indomnum[n] = indomnum;
      trees_min_indomnum_count[n] = 0;
    }
    if (indomnum == trees_min_indomnum[n])
      trees_min_indomnum_count[n]++;
    if (trees_max_indomnum[n] == 0 || indomnum > trees_max_indomnum[n]) {
      trees_max_indomnum[n] = indomnum;
      trees_max_indomnum_count[n] = 0;
    }
    if (indomnum == trees_max_indomnum[n])
      trees_max_indomnum_count[n]++;
  }

  {
    int depths[tree_max_vertices+1];
    nautyextra_depths_vpar(vpar,n,depths);
    int odd = 0;
    int even = 0;
    int i;
    if (memcmp(vpar+1,depths+1,n*sizeof(int))==0) {
      trees_depths_eq_vpar[n]++;
    }
    for (i = 1; i <= n; i++) {
      int depth = depths[i];
      odd += (depth & 1);
      even += (depth & 1) ^ 1;
    }
    if (option_verbose >= 2) {
      printf("balance odd %d even %d  ", odd, even);
      for (i = 1; i <= n; i++)
        printf(" %d", depths[i]);
      printf("\n");
    }
    int balance = abs(odd - even);
    trees_balance_total[n] += balance;
    trees_balance_zero[n]  += (balance == 0);
    trees_balanced_indnum[n] += (2*indnum + balance == n);

    /* if (n==6 && balance==0) */
    /*   nautyextra_view_vpar(vpar,n); */

    if (trees_balance_min_count[n] == 0 || balance < trees_balance_min[n]) {
      trees_balance_min[n] = balance;
      trees_balance_min_count[n] = 0;
    }
    if (balance == trees_balance_min[n])
      trees_balance_min_count[n]++;

    if (trees_balance_max_count[n] == 0 || balance > trees_balance_max[n]) {
      trees_balance_max[n] = balance;
      trees_balance_max_count[n] = 0;
    }
    if (balance == trees_balance_max[n])
      trees_balance_max_count[n]++;
  }

  {
    int limb_length = nautyextra_limb_length_vpar(vpar,n);
    trees_total_leaf_length[n] += limb_length;
    int i;
    for (i = 0; i <= TREES_LEAF_LENGTH_HISTOGRAM_MAX; i++) {
      trees_leaf_length_eq[n][i] += (limb_length == i);
      trees_leaf_length_lt[n][i] += (limb_length < i);
      trees_leaf_length_gt[n][i] += (limb_length > i);
    }
  }
}

void
try_trees (void)
{
  printf("gentreeg  n=%d to %d\n", tree_min_vertices, tree_max_vertices);
  int num_vertices;
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    if (option_verbose >= 1) {
      printf ("n=%d gentreeg\n", num_vertices);
    }
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int gentreeg_argc = 0;
    char *gentreeg_argv[10];
    gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
    gentreeg_argv[gentreeg_argc++] = "-q";       /* quiet */
    gentreeg_argv[gentreeg_argc++] = num_vertices_str;
    gentreeg_argv[gentreeg_argc] = NULL;
    my_gentreeg_main(gentreeg_argc,gentreeg_argv);
  }

  /* total vertices all connected trees of n vertices
     2,3,8,15,36,77,184,423,1060,2585,6612,16913,44226,116115,309120,826693,2229606,6041145,16461300,45034605,123722632,341045702
     A055543 in all connected trees of n vertices  (n * num trees)
  */
  printf("total vertices ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_vertices[num_vertices]);
  }
  printf("\n");

  /* total leaf vertices in all trees of n vertices
     2,2,5,9,21,43,101,226,556,1333,3365,8500,22007,57258,151264,401761,1077063,2902599,7871250,21440642,58672589,161155637
     A003228 endpoints in trees of n nodes
  */
  printf("total leaf     ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_leaf[num_vertices]);
  }
  printf("\n");

  /* total nonleaf vertices in all connected trees of n vertices
     not in OEIS: 1,3,6,15,34,83,197,504,1252,3247,8413,22219,58857,157856,424932,1152543,3138546,8590050,23593963,65050043,179890065,
  */
  printf("total nonleaf  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_nonleaf[num_vertices]);
  }
  printf("\n");

  /* count with duplicated leaf
     A003426 0,1,1,2,4,8,17,37,85,196,469,1134,2799,6975,17628,44903,115497,299089,780036,
     n =     2 3 4 5 6 7
     stable trees
  */
  printf("count with duplicated leaf  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_duplicated_leaf[num_vertices]);
  }
  printf("\n");

  /* count without duplicated leaf
     A002988 1,1,1,0,1,1,2,3,6,10,21,39,82,167,360,766,1692,3726,8370,18866,43029
     n =     0 1 2 3 4 5 6 7
     trimmed trees
     forbidden limb length 2
     leaf path length <= 2 reaches a branching node, so no limb > 2

     cf.
     A052318 labelled rooted trimmed trees
     forbidden limb length 2
     leaf up 2 reaches branch or root
     A255637 forbidden limb length 7,
     path leaf up 7 reaches root or branching node

     cf
     A002955 rooted trimmed trees
     1,2,4,8,17,36,79,175
     no limbs length >= 2, limbs going up towards the root or branching point
     Jorg Arndt: also rooted trees without x,x in pre-order level sequence,
     by at each >=2 limb change end 2 vertices into V shape
  */
  printf("count without duplicated leaf  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_noduplicated_leaf[num_vertices]);
  }
  printf("\n");


  /* number of degree-2-free trees
     A000014 1,0,1,1,2,2,4,5,10,14,26,42,78,132,249,
     .       2 3 4 5 6 7
     series-reduced trees
  */
  printf("count deg2-free    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_deg2free[num_vertices]);
  }
  printf("\n");

  /* number of degree-2-with trees
     not in OEIS: 0,1,1,2,4,9,19,42,96,221,525,1259,3081,7609,19071,
     .            2 3 4 5 6 7
     series-reduced trees
  */
  printf("count deg2-with    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_deg2with[num_vertices]);
  }
  printf("\n");

  /* number of degree-3-free trees
     not in OEIS: 1,1,1,2,3,5,9,16,29,55,106,209,420,857,1773,
     .            2 3 4 5 6 7
  */
  printf("count deg3-free    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_deg3free[num_vertices]);
  }
  printf("\n");

  /* number of degree-3-with trees
     not in OEIS: 0,0,1,1,3,6,14,31,77,180,445,1092,2739,6884,17547,
     .            2 3 4 5 6 7
     series-reduced trees
  */
  printf("count deg3-with    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_deg3with[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */
  /* Tree Degrees Arrays */


  /* degrees arrays  1,1,2,3,6,11,23,47,106,235,551,1301,3159,7741,19320,
     .           n = 2 3 4 5 6  7  8  9  10
     A000055 count trees, tree determined by degree sequence
  */
  printf("degrees arrays  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",tree_degrees_distinct[num_vertices]);
  }
  printf("\n");

  /* degrees sorted arrays 1,1,2,3,5,7,11,15,22,30,42,56,77,101,135,
     .                 n = 2 3 4 5 6 7  8  9 10
     A000041 partitions
   */
  printf("degrees sorted arrays ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",tree_degrees_distinct_sorted[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */
  /* Pseudosimilar etc */

  /* number of trees with 0 similar pairs
     .   0,0,0,0,0,1,1,3,6,15,29,67,139
     n = 2         7 8 9 10
     A000220 asymmetric trees
  */
  printf("count asymmetric    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_asymmetric[num_vertices]);
  }
  printf("\n");

  /* similar unique pair
     not in OEIS:         1,2,4,8,17,35,78,169,379
     not in OEIS: 1,1,0,1,1,2,4,8,17,35,78,169,379
     n =          2 3 4 5
     smaller than removal similar uniques when get to some pseudosimilars
     which add to the removal similar uniques
  */
  printf("similar unique pair ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,", trees_similar_is_unique[num_vertices]);
  }
  printf("\n");

  /* similar pairs total 1,1,5,9,26,49,118,232,584,1233,3031,7170,18041
     n =                 2 3  4  5   6
  */
  printf("similar pairs total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,", trees_similar_total[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* number of trees with 1 or more claws
     0,0,1,1,3,6,14,31,77,180,445,1092,2739,6884,17547
     2 3 4 5
  */
  printf("count with claw    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_withclaw[num_vertices]);
  }
  printf("\n");

  /* number of claw-free trees
     1,1,1,2,3,5,9,16,29,55,106,209,420,857,1773,
     2 3 4 5
  */
  printf("count claw-free    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_clawfree[num_vertices]);
  }
  printf("\n");

  /* total claws in trees of n vertices
     0,0,1,1,4,8,21,48,130,319,845,2196,5861,15588,42085,
     2 3 4 5
  */
  printf("total claws    ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_claws[num_vertices]);
  }
  printf("\n");

  /* max claws in trees of n vertices, uniquely binomial on star
     tetrahedral
     A000292 0,0,1,4,10,20,35,56,84,120,165,220,286,364,455,
     .       2 3 4 5  6  7  8
     vector(10,n,n++; (n-1)*(n-2)*(n-3)/6)
     count   1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
  */
  printf("max claws       ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_claws_max[num_vertices]);
  }
  printf("\n");
  printf("max claws count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_claws_max_count[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */
  /* depths == vpar
     free trees:
     1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,
   */

  printf("trees depths==vpar ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_depths_eq_vpar[num_vertices]);
  }
  printf("\n");


  /* --------------------------------------------------- */

  /* balance total
     .            2 3 4 5
     not in OEIS: 0,1,2,5,8,21,38,97,198,515,1134,3013,6986,18859,45476

     n=2  *---* only

     n=3  *---*---* only  balance=1

     n=4  *---*---*---*  balance=0

     .        *
     .        |          balance=2
     .    *---*---*
  */
  printf("balance total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_balance_total[num_vertices]);
  }
  printf("\n");

  /* balance zero
     .            2 3 4 5 6
     not in OEIS: 1,0,1,0,3,0,9,0,37,0,168,0,895,0,5097,
     A119856      1,  1,  3,  9,  37,  168,  895,  5097,
     equicolourable unrooted trees of 2n vertices

     n=6  *---*---*---*---*---*  balance=0    path-6

     .        *   *
     .        |   |         balance=0      bi-star 3,3
     .    *---*---*---*

     .            *
     .            |             balance=0
     .    *---*---*---*---*
  */
  printf("balance zero  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_balance_zero[num_vertices]);
  }
  printf("\n");

  /* balance min
     2 3 4 5
     0,1,0,1,0,1,0,1,0,1,0,1,0,1,0,
     count
     .            2 3 4 5
     not in OEIS: 1,1,1,2,3,7,9,28,37,132,168,693,895,3927,5097,
  */
  printf("balance min  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_balance_min[num_vertices]);
  }
  printf("\n");
  printf("       count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_balance_min_count[num_vertices]);
  }
  printf("\n");

  /* balance max  is star
     .     2 3 4 5 6
     max   0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,
     count 1,1,1,1,1,1,1,1,1,1, 1, 1, 1, 1, 1,
  */
  printf("balance max  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_balance_max[num_vertices]);
  }
  printf("\n");
  printf("       count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_balance_max_count[num_vertices]);
  }
  printf("\n");

  /* balanced indnum
     not in OEIS: 1,0,1,0,2,0,5,0,15,0,49,0,180,0,701,
     A005751      1,  1,  2,  5,  15,  49,  180,  701
     matched trees with 2n nodes
  */
  printf("balanced indnum  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_balanced_indnum[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* trees total diameter
     A001851 1,2,5,9,21,44,103,232,571,1368,3441,8653,22263,57518,150733,
  */
  printf("total diameter  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_diameter[num_vertices]);
  }
  printf("\n");

  /* trees total diameter endpoints
     not in OEIS: 2,2,5,9,20,39,85,179,413,931,2210,5270,12883,31736,79558,200972,
     frac of leaf 1.0000, 1.0000, 1.0000, 1.0000, 0.9524, 0.9070, 0.8416, 0.7920, 0.7428, 0.6984, 0.6568, 0.6200, 0.5854, 0.5543, 0.5260, 0.5002

     total non diameter endpoints
     not in OEIS: 0,1,3,6,16,38,99,244,647,1654,4402,11643,31343,84379,229562,625721
     .            2 3 4 5  6
     frac of non-leaf  -nan, 1.0000, 1.0000, 1.0000, 1.0667, 1.1176, 1.1928, 1.2386, 1.2837, 1.3211, 1.3557, 1.3839, 1.4106, 1.4336, 1.4542, 1.4725,
     the not-diameter-endpoints are all non-leaf, plus the leaves which are not
  */
  printf("total diameter endpoints      ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_diameter_endpoints[num_vertices]);
  }
  printf("\n");
  printf(" frac of leaf ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf(" %.4f,",
           (double) trees_total_diameter_endpoints[num_vertices]
           / (double) trees_total_leaf[num_vertices]);
  }
  printf("\n");
  printf("total non diameter endpoints  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_non_diameter_endpoints[num_vertices]);
  }
  printf("\n");
  printf(" frac of non-leaf ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf(" %.4f,",
           (double) trees_total_non_diameter_endpoints[num_vertices]
           / (double) trees_total_nonleaf[num_vertices]);
  }
  printf("\n");

  /* trees where diameter unique endpoints
     num_endpoints <= 2
     not in OEIS: 1,1,1,1,2,3,6,11,24,51,118,271,651,1572,3898,9746,
     n =          2 3 4 5 6 7 8  9
  */
  printf("diameter unique endpoints      ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_diameter_unique_endpoints[num_vertices]);
  }
  printf("\n");

  /* trees total diameter vertices
     not in OEIS: 2,3,8,15,35,73,168,375,912,2158,5359,13327,33897,86664,224952,587144
     frac of vertices 1.0000, 1.0000, 1.0000, 1.0000, 0.9722, 0.9481, 0.9130, 0.8865, 0.8604, 0.8348, 0.8105, 0.7880, 0.7664, 0.7464, 0.7277, 0.7102,

     total non diameter vertices
     not in OEIS: 0,0,0,0,1,4,16,48,148,427,1253,3586,10329,29451,84168,239549
     .            2 3 4 5 6
  */
  printf("total diameter vertices      ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_diameter_vertices[num_vertices]);
  }
  printf("\n");
  printf(" frac of vertices ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf(" %.4f,",
           (double) trees_total_diameter_vertices[num_vertices]
           / (double) trees_total_vertices[num_vertices]);
  }
  printf("\n");
  printf("total non diameter vertices  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_non_diameter_vertices[num_vertices]);
  }
  printf("\n");

  /* trees unicentral
     A000676 0,1,1,2,3,7,12,27,55,127,284,682,1618,3979,9823,
  */
  printf("count unicentral  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_unicentral[num_vertices]);
  }
  printf("\n");

  /* trees bicentral
     A000677 1,0,1,1,3,4,11,20,51,108,267,619,1541,3762,9497,
  */
  printf("count bicentral  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_bicentral[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* trees total eccentricities
     not in OEIS: 2,5,17,38,106,257,684,1725,4698,12334,33725,91590,253038,698624,1948244,5439341,
     n =          2 3  4  5

     n=3 path-3 2+1+2==5
  */
  printf("total eccentricities  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_eccentricities[num_vertices]);
  }
  printf("\n");

  /* --------------------------------------------------- */

  /* trees total centroid weight
     not in OEIS: 1,1,3,5,14,28,74,165,440,1050,2800,7045,19011,49188,133961,
  */
  printf("total centroid weight  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_centroid_weight[num_vertices]);
  }
  printf("\n");

  /* trees unicentroid
     A027416 0,1,1,3,3,11,13,47,61,235,341,1301,1983,7741,12650,
  */
  printf("count unicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_unicentroid[num_vertices]);
  }
  printf("\n");

  /* trees bicentroid
     A102911      1,  1,  3,  10,  45,  210,  1176,  6670,
     not in OEIS: 1,0,1,0,3,0,10,0,45,0,210,0,1176,0,6670,
     .            2   4   6    8
     .            2   4   6    8
  */
  printf("count bicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_bicentroid[num_vertices]);
  }
  printf("\n");


  /* trees unicentral unicentroid
     not in OEIS: 0,1,1,2,2,7,8,27,35,127,184,682,1041,3979,6517,
  */
  printf("count unicentral unicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_unicentral_unicentroid[num_vertices]);
  }
  printf("\n");

  /* trees unicentral bicentroid
     not in OEIS: 0,  0,  1,  4,  20,  100,  577,  3306,
     not in OEIS: 0,0,0,0,1,0,4,0,20,0,100,0,577,0,3306,
     .            2   4   6   8
  */
  printf("count unicentral bicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_unicentral_bicentroid[num_vertices]);
  }
  printf("\n");

  /* trees bicentral unicentroid
     not in OEIS: 0,0,0,1,1,4,5,20,26,108,157,619,942,3762,6133
  */
  printf("count bicentral unicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_bicentral_unicentroid[num_vertices]);
  }
  printf("\n");

  /* trees bicentral bicentroid
     not in OEIS: 1,  1,  2,  6,  25,  110,  599,  3364,
     not in OEIS: 1,0,1,0,2,0,6,0,25,0,110,0,599,0,3364,
     .            2   4   6   8
     n=6 path-6 and bi-star 3,3
  */
  printf("count bicentral bicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_bicentral_bicentroid[num_vertices]);
  }
  printf("\n");

  /* trees centre_centroid    always star, path
     not in OEIS: 1,1,2,2,4,6,12,20,45,84,194,403,976,2170,5411
     .            2 3 4 5 6
  */
  printf("count centre_centroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_centre_centroid[num_vertices]);
  }
  printf("\n");

  /* trees unicentre_centroid
     not in OEIS: 0,1,1,2,2,6,7,20,27,84,125,403,642,2170,3686,
     .            2 3 4 5 6 7
  */
  printf("count unicentre_centroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_unicentre_centroid[num_vertices]);
  }
  printf("\n");

  /* trees bicentre_centroid
     not in OEIS: 1,  1,  2,  5,  18,  69,  334,  1725,
     not in OEIS: 1,0,1,0,2,0,5,0,18,0,69,0,334,0,1725,
     .            2   4   6
  */
  printf("count bicentre_centroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_bicentre_centroid[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* trees equal_centre_centroid
     not in OEIS: 1,1,2,2,4,6,12,20,45,84,194,403,976,2170,5411,12536,31998,76839,199554,
     n =          2 3 4 5 6 7  8  9
  */
  printf("count equal_centre_centroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_equal_centre_centroid[num_vertices]);
  }
  printf("\n");

  /* count equal_unicentral_unicentroid
     not in OEIS: 0,1,1,2,2,6,7,20,27,84,125,403,642,2170,3686,12536,22198,76839,141428,
     not in OEIS:       2,2,6,7,20,27,84,125,403,642,2170,3686,12536,22198,76839,141428,
     n =          2 3 4 5 6 7 8  9 10
     n=7 smallest unrestricted
  */
  printf("count equal_unicentral_unicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_equal_unicentral_unicentroid[num_vertices]);
  }
  printf("\n");

  /* count equal_bicentral_bicentroid
     not in OEIS: 1,0,1,0,2,0,5,0,18,0,69,0,334,0,1725,0,9800,0,58126,
     not in OEIS: 1,  1,  2,  5,  18,  69,  334,  1725,  9800,  58126,
     n =          2 3 4 5 6 7 8 9 10   12
  */
  printf("count equal_bicentral_bicentroid   ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_equal_bicentral_bicentroid[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* trees disjoint_centre_centroid
     not in OEIS: 0,0,0,0,0,1,1,8,10,53,83,363,608,2435,4495,
     .            2 3 4 5 6 7 8 9
     n=7 star-5 with one long tail
     n=8 star-6 with one long tail
  */
  printf("count disjoint_centre_centroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_disjoint_centre_centroid[num_vertices]);
  }
  printf("\n");

  /* count disjoint_unicentral_unicentroid
     not in OEIS: 0,0,0,0,0,1,1,7,8,43,59,279,399,1809,2831,12186,19981,83905,144392
     not in OEIS:           1,1,7,8,43,59,279,399,1809,2831,12186,19981,83905,144392
     n            2 3 4 5 6 7 8 9 10
     n=7 smallest unrestricted
  */
  printf("count disjoint_unicentral_unicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_disjoint_unicentral_unicentroid[num_vertices]);
  }
  printf("\n");

  /* count disjoint_unicentral_bicentroid
     not in OEIS: 0,0,0,0,0,0,0,0,1,0,10,0,82,0,585,0,4306,0,30914,
     not in OEIS:                 1,0,10,0,82,0,585,0,4306,0,30914,
     not in OEIS:                 1,  10,  82,  585,  4306,  30914,
     n            2 3 4 5 6 7 8 9 10
  */
  printf("count disjoint_unicentral_bicentroid   ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_disjoint_unicentral_bicentroid[num_vertices]);
  }
  printf("\n");

  /* count disjoint_bicentral_unicentroid
     not in OEIS: 0,0,0,0,0,0,0,1,1,10,13,84,112,626,937,4717,7424,35166,58655,
     not in OEIS:               1,1,10,13,84,112,626,937,4717,7424,35166,58655,
     n            2 3 4 5 6 7 8 9 10
  */
  printf("count disjoint_bicentral_unicentroid  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_disjoint_bicentral_unicentroid[num_vertices]);
  }
  printf("\n");

  /* count disjoint_bicentral_bicentroid
     not in OEIS: 0,0,0,0,0,0,0,0,0,0,1,0,15,0,142,0,1259,0,10290
     not in OEIS:                     1,0,15,0,142,0,1259,0,10290
     not in OEIS:                     1,  15,  142,  1259,  10290
     n            2 3 4 5 6 7 8 9    12
     starts like A045724 1,15,142,1083 but bigger
  */
  printf("count disjoint_bicentral_bicentroid   ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_disjoint_bicentral_bicentroid[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* centre to centroid distance
     cc_dist total
     not in OEIS: 0,0,0,0,0,1,1,8,10,54,84,379,628,2603,4751,18522,35642,133519,270224,980789,2059221
  */
  printf("cc_dist total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_cc_dist_total[num_vertices]);
  }
  printf("\n");
  /* cc_dist max  star and long tail, and lengthening the star or some hanging
     A002265      0,0,0,0,0, 1,1,1, 1,  2,2, 2, 2, 3,3,3,3,4,4,4,4,
     .            2 3 4 5 6  7
     vector(10,n,n++; floor((n-3)/4))
     count
     not in OEIS: 1,1,1,1,1, 1,1,8,10,  1,1,16,20,  1,1,29,37,  1,1,51,65,
     .            2 3 4 5 6  7 8 9 10  11          15          19      22
  */
  printf("cc_dist_max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_cc_dist_max[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_cc_dist_max_count[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* bandwidth min = 1 path
     bandwidth = 1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,
     count       1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,
  */
  printf("bandwidth min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_bandwidth[num_vertices]);
  }
  printf("\n");
  printf("             count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_bandwidth_count[num_vertices]);
  }
  printf("\n");

  /* bandwidth max
     max   1,1,2,2,3,3,4,4,5,0,0,0,0,0,0,0,
     count 1,1,1,2,1,2,1,2,1,0,0,0,0,0,0,0
     n =   2 3 4 5 6 7 8

     star and also with length=2 arm
     n=8 0,1,2,2,2,2,1
     .          0
     .      2       8
     .  3 4 5 6 7
  */
  printf("bandwidth max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_bandwidth[num_vertices]);
  }
  printf("\n");
  printf("             count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_bandwidth_count[num_vertices]);
  }
  printf("\n");

  /* bandwidth total over free trees
     not in OEIS: 1,1,3,5,12,23,56,112,277
     n =          2 3 4 5  6  7  8   9  10

     n=3 path-3 only, label 0,1,2 dilation=1

     n=4 path-4 label 0,1,2,3 dilation=1
     .   claw   label 2 - 0,1,3 dilation=2
  */
  printf("bandwidth total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_bandwidth[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */
  /* indsets min = F(n+2), unique
     3,5,8,13,21,34,55,89,144,233,377,610,987,1597,2584,4181,6765,10946,17711,28657,46368,75025
     for(n=2,10,print(fibonacci(n+2)))
  */
  printf("indsets min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_min_indsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_indsets_count[num_vertices]);
  }
  printf("\n");

  /* indsets max = 2^n+1, unique
     3,5,9,17,33,65,129,257,513,1025,2049,4097,8193,16385,32769,65537,131073,262145,524289,1048577,2097153,4194305
     for(n=2,10,print(2^(n-1)+1))
  */
  printf("indsets max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_max_indsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_indsets_count[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* indnum min = ceil(n/2) for path, and others
     1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12
     count
     not in OEIS: 1,1,1,2,2,6,5,20,15,76,49,313,180,1361,701,6161,2891,28722,12371,136882,54564,663890
  */
  printf("indnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_indnum_count[num_vertices]);
  }
  printf("\n");

  /* indnum max = n-1 for star, unique
     1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22
  */
  printf("indnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_indnum_count[num_vertices]);
  }
  printf("\n");

  /* indnums occurring 1,1,2,2,3,3,4,4, 5,5,6,6,7,7,8,
     .                 2 3 4 5 6 7 8 9 10
     all values min to max ...
  */
  printf(" occurring ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    int occurring = setsize(trees_indnum_occurring[num_vertices],
                            TREES_INDNUM_OCCURRING_M);
    printf("%d,",occurring);
    if (0) {
      printf("\n%s\n",nautyextra_set_str(trees_indnum_occurring[num_vertices],
                                         TREES_INDNUM_OCCURRING_M));
    }
  }
  printf("\n");

  /* indnum total
     not in OEIS: 1,2,5,10,23,50,117,270,669,1631,4147,10592,27598,72341,192147,513117,1381659,3739058,10176079,27811512,76332960,210238442
  */
  printf("indnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_indnum[num_vertices]);
  }
  printf("\n");

  /* indnum = n/2
     not in OEIS: 1,0,1,0,2,0,5,0,15,0,49,0,180,0,701,
     A005751      1,  1,  2,  5,  15,  49,  180,  701,
     is matched trees, perfect matching ... below
  */
  printf("indnum half ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_half_indnum[num_vertices]);
  }
  printf("\n");

  /* indnumCount = 1
     .            2 3 4 5 6 7 8  9
     not in OEIS: 0,1,1,2,3,6,9,19,35,73,145,314,661,1458,3206,
     indnum attained uniquely
     always star, n odd always path
     n=6 star (indnum=5)
     .   bi-star 3,3 (indnum=4),
     .   path-5 and one extra (indnum=4)
  */
  printf("indnumCount==1  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_indnumCount_eq_1[num_vertices]);
  }
  printf("\n");

  /* indnumCount = 2
     .            2 3 4 5 6 7 8  9
     not in OEIS: 1,0,0,1,1,2,6,11,24,55,126,288,678,1595,3796
     indnum attained in 2 ways
     n=6 bi-star 4,2 (indnum=4)
  */
  printf("indnumCount==2  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_indnumCount_eq_2[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* duplicated leaf indnum min
     indnum min 0,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,
     count
     not in OEIS: 0,1,1,1,3,3,11,10,46,39,209,160,997,702,4929,3177,24928,14830,128262
  */
  printf("duplicated leaf indnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_min_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_min_indnum_count[num_vertices]);
  }
  printf("\n");

  /* duplicated leaf indnum max  for star, unique
     0,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,
     count 0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  */
  printf("duplicated leaf indnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_max_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_max_indnum_count[num_vertices]);
  }
  printf("\n");

  /* duplicated leaf indnums
     occurring 0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9
     n =       2 3 4 5 6 7 8 9 10
  */
  printf(" occurring ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    int occurring = setsize(trees_duplicated_leaf_indnum_occurring[num_vertices],
                            TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_M);
    printf("%d,",occurring);
    if (0) {
      printf("\n%s\n",nautyextra_set_str(trees_duplicated_leaf_indnum_occurring[num_vertices],
                                         TREES_DUPLICATED_LEAF_INDNUM_OCCURRING_M));
    }
  }
  printf("\n");

  /* duplicated leaf indnum total
     not in OEIS: 0,2,3,7,17,38,92,220,558,1395,3622,9409,24895,66106,177586,478835,1300522,3545341,9712309
  */
  printf("duplicated leaf indnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_duplicated_leaf_total_indnum[num_vertices]);
  }
  printf("\n");

  /* duplicated leaf indnum half = n/2 never
     0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  */
  printf("duplicated leaf indnum half ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_half_indnum[num_vertices]);
  }
  printf("\n");

  /* indnumCount = 1 never
     0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  */
  printf("duplicated leaf indnumCount==1  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_indnumCount_eq_1[num_vertices]);
  }
  printf("\n");

  /* indnumCount = 2 never
     0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  */
  printf("indnumCount==2  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_duplicated_leaf_indnumCount_eq_2[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* noduplicated leaf indnum min = path and others
     not in OEIS: 1,0,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,
     count
     not in OEIS: 1,0,1,1,2,3,5,10,15,37,49,153,180,659,701,2984,2891,13892,12371
     n =          2   3 4 5 6 7  8  9 10
  */
  printf("noduplicated leaf indnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_noduplicated_leaf_min_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_noduplicated_leaf_min_indnum_count[num_vertices]);
  }
  printf("\n");

  /* noduplicated leaf indnum max
     not in OEIS: 1,0,2,3,3,4,5,5, 6, 7, 7, 8, 9,9,10,11,11,12,13
     n =          2 3 4 5 6 7 8 9 10 11 12 13 14
     none noduplicated of n=3
     GP-Test  my(v=[1,0,2,3,3,4,5,5,6,7,7,8,9,9,10,11,11,12,13]); \
     GP-Test  vector(#v,n,n++; if(n==3,0, floor((2*n-1)/3))) == v

     count
     not in OEIS: 1,0,1,1,2,3,1,10, 6, 2,33,14, 3,107,34, 6,328,83,11
     n =          2 3 4 5 6 7 8  9 10 11 12 13 14  15 16 17  18 19 20

     n == 2 mod 3:            1,       2,       3,        6,       11,     23
     A000055 count all trees, at least so far
     not in OEIS:     1,    3,      6,      14,       34,       83,     207
     not in OEIS:         2,    10,      33,      107,      328     \\
  */
  printf("noduplicated leaf indnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_noduplicated_leaf_max_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_noduplicated_leaf_max_indnum_count[num_vertices]);
  }
  printf("\n");

  /* noduplicated leaf indnum total
     not in OEIS: 1,0,2,3,6,12,25,50,111,236,525,1183,2703,6235,14561,34282
  */
  printf("noduplicated leaf indnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_noduplicated_leaf_total_indnum[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* perfect matching */

  /* trees with perfect matchings
     .            2 3 4 5 6 7
     not in OEIS: 1,0,1,0,2,0,5,0,15,0,49,0,180,0,701
     A005751      1,  1,  2,  5,  15,  49,  180,  701
  */
  printf("perfect matchings ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_perfect_matching[num_vertices]);
  }
  printf("\n");

  /* trees with near perfect matchings
     .            2 3 4 5 6 7
     not in OEIS: 0,1,0,2,0,6,0,20,0,76,0,313,0,1361,0,
     not in OEIS:   1,  2,  6,  20,  76,  313,  1361

     total number of such matchings
     not in OEIS: 0,2,0,5,0,17,0,59,0,231,0,969,0,4272,0,
     not in OEIS:   2,  5,  17,  59,  231,  969,  4272,
  */
  printf("near perfect matchings ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_near_perfect_matching[num_vertices]);
  }
  printf("\n");
  printf("near perfect total     ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_near_perfect_matching_total[num_vertices]);
  }
  printf("\n");

  /* matchnum min ways 1,2,1,2,1,2,1,2,1,2,1,2,1,2,1,2,
     not in OEIS: 1,1,1,1,2,3,5,10,15,39,49,160,180,702,701,3177,
     not in OEIS:           3,5,10,15,39,49,160,180,702,701,3177,
  */
  printf("matchnum min ways ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_matchnum_ways[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_matchnum_ways_count[num_vertices]);
  }
  printf("\n");

  /* matchnum max ways
     n          = 2   4   6    8    10    12    14      16
     A333347      1,2,3,4,5,8,11,15,21,30,41,56,81,112,153,216,
     matchnum         1 1 x 2  2  2  3  3  3  3  4   4   4   5
     ways 1,1,1,1,2,1,1,1,1,1,1,1,1,1,1,1,
     unique other than n=6
     recurrence_guess([11,15,21,30,41,56,81,112,153,216])
     n=4 matchnum 1 ways 3     4-star
     n=5 matchnum 1 ways 4     5-star
     n=6 matchnum 2 ways 5     6-star
     n=6 matchnum 1 ways 5    *--*--*--*--*
     .                                  \-*
     n=7 matchnum 2 ways 8    *--*--*--*--*
     .                        *-/       \-*

     n=8 matchnum 2 ways 11             /-*
     .                        *--*--*--*--*
     .                        *-/       \-*

     n=9 matchnum 2 ways 15   *-\       /-*
     .                        *--*--*--*--*
     .                        *-/       \-*

     n=10 matchnum 3 ways 21  *--*--*--*--*--*--*
     .                        *-/      |      \-*
     .                                 *

     n=11 matchnum 3 ways 30  *--*--*--*--*--*--*
     .                        *-/     / \     \-*
     .                               *   *

     .                         *--\
     n=12 matchnum 3 ways 41   *---*--*--*--*--*---*
     .                         *--/     / \     \--*
     .                                 *   *

     .                         *--\             /--*
     n=13 matchnum 3 ways 56   *---*--*--*--*--*---*
     .                         *--/     / \     \--*
     .                                 *   *

     n=14 matchnum 4 ways 81          * ---*
     .                             /  |  \
     .                           *    *    *
     .                           |    |    |
     .                           *    *    *
     .                          / \  / \  / \
     .                         *  *  * *  *  *

     n=15 matchnum 4 ways 112     *--*--*--*--*--*--*
     .                           /|    /|     |\    |\
     .                          * *   * *     * *   * *
     3^4 == 81

     n=16 matchnum 4 ways 153   *--*--*--*--*--*--*--*
     .                            /|    /|     |\    |\
     .                           * *   * *     * *   * *

     n=17 matchnum 5 ways 216             *
     .                               /  |   |  \
     .                             *    *   *    *
     .                             |    |   |    |
     .                             *    *   *    *
     .                            / \  / \ / \  / \
     .                           *  *  * * * *  *  *
     4*binomial(4,4)*2^4 + 3*binomial(4,3)*2^3 + 2*binomial(4,2)*2^2 + 1*binomial(4,1)*2 == 216
     3^4*3 - 3^3 == 216
  */
  printf("matchnum max ways ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_matchnum_ways[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_matchnum_ways_count[num_vertices]);
  }
  printf("\n");

  /* matchnum total
     not in OEIS: 1,1,3,5,13,27,67,153,391,954,2465,6321,16628,43774,116973
  */
  printf("matchnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_matchnum[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* domsets */

  /* domsets min = 1 for star, unique
     ???????
     domsets min
     A005517      3,5,9,15,25,45,75,125,225,375,625,1125,1875,3125,5625,9375,15625,28125,46875,78125,140625,234375,
     count
     not in OEIS: 1,1,2,1,1,3,2,1,8,4,2,19,9,3,52,20,6,136,48,11,377,115,
  */
  printf("domsets min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_min_domsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_domsets_count[num_vertices]);
  }
  printf("\n");

  /* domsets max
     max 2^(n-1) + 1
     3,5,9,17,33,65,129,257,513,1025,2049,4097,8193,16385,32769,65537,131073,262145,524289,1048577,2097153,4194305,
     3,5,9,17,33,65,129,257,513,1025,2049,4097,8193,16385,32769,
     2 3 4  5  6  7   8   9  10
     count
     1,1,2, 2, 1, 1,  1,  1,  1,1,1,1,1,1,1,

     n=4  star-4 claw and path-4
     n=5  star-5 and path-5
     otherwise star only
  */
  printf("domsets max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_max_domsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_domsets_count[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* minimal domsets */

  /* minimal_domsets min = 2 for star (centre or all outsides), unique
     min   2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,
     count 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  */
  printf("minimal_domsets min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_min_minimal_domsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_minimal_domsets_count[num_vertices]);
  }
  printf("\n");

  /* minimal_domsets max
     count
     n     2 3 4 5 6 7  8    10    12     14      16      18       20        22
     max   2,2,4,4,8,9,16,19,32,41,64,85,128,177,256,361,512,737,1024,1489,2048,3009,
     2^n/2 2   4   8   16
     count 1,1,1,2,1,1, 2, 1, 3, 1, 6, 1, 11,  1, 23,  1, 47,  1, 106,   1, 235,   1,
     count         1,   2,    3,    6,    11,     23,     47,     106,      235

     frac of 2^(n/2)  1.0000, 0.7071, 1.0000, 0.7071, 1.0000, 0.7955, 1.0000, 0.8397, 1.0000, 0.9060, 1.0000, 0.9391, 1.0000, 0.9778, 1.0000, 0.9971,

     even max 2^(n/2)
     even count = A000055 num trees of n/2 vertices
     n                  4     5     6     7     8    9    10

     n=2  *---*   count=2   n=1 extended by 1 vertex
     n=4  *-*---*-*   count=4   n=2 extended by 1 vertex each

     2^(7/2)  \\   11.31
     2^(11/2) \\   45.25
     2^(13/2) \\   90.50
     2^(15/2) \\  181.01
     2^(17/2) \\  362.03
     2^(19/2) \\  724.07
     2^(21/2) \\ 1448.15
     2^(23/2) \\ 2896.30

     A306692      2,2,4,4,8,9,16,19,32,41,64,85
     not in OEIS: 9,19,41,85,177,361,737,1489,3009,

     vector_modulo(v,i) = v[(i% #v)+1];
     NMD(n) = vector_modulo([1, 3/2],n) * 2^floor(n/2) \
     - vector_modulo([0, 3/2, 0, 2],n) * 2^floor(n/4) \
     + vector_modulo([0, 1],n)
     vector(24,n,n--; NMD(n))
     forstep(n=27,999,27, print(NMD(n) / 12161^(n/27) * 1.0));
     NMD(27) == 12161
     NMD(27*2) == 134217728
     12161^2   == 147889921

     to_binary(n)=fromdigits(binary(n))
     vector(24,n,n--; to_binary(n))

  */
  printf("minimal_domsets max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_max_minimal_domsets[num_vertices]);
  }
  printf("\n");
  printf(" frac of 2^(n/2) ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf(" %.4f,",
           (double) trees_max_minimal_domsets[num_vertices]
           / (double) pow(2.0, (double)num_vertices / 2.0));
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_minimal_domsets_count[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */
  /* minimal_domsets frac */

  /* minimal_domsets_frac min
     2/3, 2/5, 2/9, 2/17, 2/33, 2/65, 2/129, 2/257, 2/513, 2/1025, 2/2049, 2/4097, 2/8193, 2/16385, 2/32769, 2/65537,
     count 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  */
  printf("minimal_domsets_frac min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%.4f, ",trees_minimal_domsets_frac_min[num_vertices]);
  }
  printf("\n");
  printf("     ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu/%Lu, ",trees_minimal_domsets_frac_min_num[num_vertices],trees_minimal_domsets_frac_min_den[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_minimal_domsets_frac_min_count[num_vertices]);
  }
  printf("\n");

  /* minimal_domsets_frac max
     2/3, 2/5, 4/9, 4/15, 8/27, 8/45, 16/81, 16/135, 32/243, 32/405, 64/729, 64/1215, 128/2187, 128/3645, 256/6561, 361/15201
     n=2    3    4     5     6     7      8       9      10      11      12       13        14        15        16  17
     count 1,1,1,1,1,2,2,4,3,9,6,20,11,48,23,1
     n =   2 3 4 5 6 7 8 9  11   13 14 15 16 17

  */
  printf("minimal_domsets_frac max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%.4f, ",trees_minimal_domsets_frac_max[num_vertices]);
  }
  printf("\n");
  printf("     ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu/%Lu, ",trees_minimal_domsets_frac_max_num[num_vertices],trees_minimal_domsets_frac_max_den[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_minimal_domsets_frac_max_count[num_vertices]);
  }
  printf("\n");

  printf("\n");


  /* --------------------------------------------------- */

  /* domnum min = 1 for star, unique
   */
  printf("domnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_domnum_count[num_vertices]);
  }
  printf("\n");

  /* domnum max 1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,
     2 3 4 5 6 ...
     = floor(n/2) for various, inc 2-arm star
     count
     not in OEIS: 1,1,1,2,1,5,2,11,3,27,6,65,11,163,23,
  */
  printf("domnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_domnum_count[num_vertices]);
  }
  printf("\n");

  /* domnum min ways 2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
     uniquely attained
     count trees with this many ways
     not in OEIS: 1,1,1,1,3,4,6,11,22,38,75,153,308,616,1310,2776,
     not in OEIS:         3,4,6,11,22,38,75,153,308,616,1310,2776,
  */
  printf("domnum min ways ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_domnum_ways[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_domnum_ways_count[num_vertices]);
  }
  printf("\n");

  /* domnum max ways
     n          = 2   4   6    8    10    12     14      16
     not in OEIS: 2,1,4,3,8,8,16,18,32,40,64,84,128,176,256,360,
     not in OEIS:             16,18,32,40,64,84,128,176,256,360,
     vector(16,n,n++; if(n%2==0, 2^(n/2), 3*2^((n-3)/2) - if(n%4==1,3,2)*2^((n-2)\4)))
     count trees with this many ways
     n =          2 3 4 5 6 7 8  10       14
     not in OEIS: 1,1,1,1,1,1,2,1,3,1,6,1,11,1,23,1,
     .                        2,  3,  6,  11,  23
     A000055 free trees
  */
  printf("domnum max ways ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_domnum_ways[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_domnum_ways_count[num_vertices]);
  }
  printf("\n");

  /* domnums occurring 1,1,2,2,3,3,4,4,5,5,6,6,7,7,8
     .                 2 3 4 5 6 7 8 9 10
     all values min to max ...
  */
  printf(" occurring ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    int occurring = setsize(trees_domnum_occurring[num_vertices],
                            TREES_DOMNUM_OCCURRING_M);
    printf("%d,",occurring);
    if (0) {
      printf("\n%s\n",nautyextra_set_str(trees_domnum_occurring[num_vertices],
                                         num_vertices));
    }
  }
  printf("\n");

  /* domnum total for all trees of n vertices
     not in OEIS: 1,1,3,5,12,26,62,142,357,870,2225,5691,14881,39058,103965,
  */
  printf("domnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_domnum[num_vertices]);
  }
  printf("\n");

  /* minimal domsets all minimum
     count 1,0,1,0,1,0,2,0,3,0,6,0,11,0,23,0
     n =   2 3 4 5 6   8  10  12   14   16
     all trees with corona of leaves
  */
  printf("minimal domsets all minimum ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%u,",trees_domsets_minimal_all_minimum_count[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* disjoint_domnum min 2,3,4,4,5,6,6,7,8,8,9,10,10,11,12,12,13,14,14,
     A004523
     count
     not in OEIS: 1,1,2,1,2,6,2,6,24,5,22,104,15,89,503,49,382,2499,180,
  */
  printf("disjoint_domnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_disjoint_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_disjoint_domnum_count[num_vertices]);
  }
  printf("\n");

  /* disjoint_domnum max 2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,
     n                 = 2 3 4 ...
     max = n for star and various others
     count
     not in OEIS: 1,1,2,2,4,5,10,14,27,43,82,140,269,486,939,1765,3446,6652,13137,
     not in OEIS:           5,10,14,27,43,82,140,269,486,939,1765,3446,6652,13137,
  */
  printf("disjoint_domnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_disjoint_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_disjoint_domnum_count[num_vertices]);
  }
  printf("\n");

  /* disjoint_domnum min ways 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
     count trees with 1 ways
     not in OEIS: 1,1,1,2,2,3,5,8,12,22,37,65,117,219,399,754,1434,2751,5293,
     not in OEIS:           3,5,8,12,22,37,65,117,219,399,754,1434,2751,5293,
  */
  printf("disjoint_domnum min ways ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_disjoint_domnum_ways[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_disjoint_domnum_ways_count[num_vertices]);
  }
  printf("\n");

  /* disjoint_domnum set pairs max ways
     n          = 2   4   6
     not in OEIS: 1,1,2,2,4,5,8,13,16,33,33,73,73,161,181,337,473,705,1089,
     count trees with this many ways
     n =          2 3 4 5 6 7 8  10  12  14  16  18  20
     not in OEIS: 1,1,1,1,1,1,2,1,3,1,1,1,2,1,1,1,1,1,1,
     not A128267 gcd of stuff, seems unlikely

     n=2,3,4,7 path only
     n=5     *--*--*--*       disjoint domnum 5 in 2 ways
     .          |
     .
     my(n=5); 2^(n/2-1)     \\ 2.82

     n=6     *--*--2--*--*    disjoint domnum 6 in 4 ways
     .             |
     .             1          leaf must 1,2 then others each way for 4
     .

     n=7 disjoint domnum 6 ways 5       <- smaller
     path-7
     my(n=7); 2^(n/2-1)     \\ 5.65

     n=8 disjoint domnum 8 ways 8
     n=8 disjoint domnum 8 ways 8
     n=8         *            disjoint domnum 8 in 8 ways
     .        / / \ \
     .      *  *   *  *         the two n=4 trees corona of singletons
     .      |  |   |            claw and path-4
     .      *  *   *
     n=8      *--*--*--*      disjoint domnum 8 in 8 ways
     .        |  |  |  |
     .        *  *  *  *
     vpar_count_free(4) == 2

     n=9 disjoint domnum 8 ways 13       <- smaller
     n=9   1--2--3--4--5--6--7--8    disjoint domnum 8 in 13 ways
     .           |
     .           9
     my(n=9); 2^(n/2-1)     \\ 11.31

     n=10 disjoint domnum 10 ways 16
     n=10 disjoint domnum 10 ways 16
     n=10 disjoint domnum 10 ways 16
     n=10        *             disjoint domnum 10 in 16 ways
     .        / / \ \  \
     .      *  *   *  *  *
     .      |  |   |  |
     .      *  *   *  *
     n=10   *--*--*--*--*      disjoint domnum 10 in 16 ways
     .      |  |  |  |  |
     .      *  *  *  *  *
     .         *--*
     .         |
     n=10   *--*--*--*         disjoint domnum 10 in 16 ways
     .      |  |  |  |
     .      *  *  *  *
     GP-Test  my(n=10); 2^(n/2-1) == 16
     vpar_count_free(5) == 3

     n=12 disjoint domnum 11 ways 33
     n=12   *--*--*--*--*--*
     .      |  |     |  |
     .      *  *     *  *
     .         |     |
     .         *     *
     GP-Test  my(n=12); 2^(n/2-1) == 32
     https://hog.grinvin.org/ViewGraphInfo.action?id=31088

     n=2 disjoint domnum 2 ways 1
     n=3 disjoint domnum 3 ways 1
     n=4 disjoint domnum 4 ways 2
     n=5 disjoint domnum 5 ways 2
     n=6 disjoint domnum 6 ways 4

     n=11 disjoint domnum 10 ways 33      rest smaller
     n=13 disjoint domnum 12 ways 73
     n=14 disjoint domnum 13 ways 73
     n=14 disjoint domnum 13 ways 73
     n=15 disjoint domnum 14 ways 161
     n=16 disjoint domnum 14 ways 181
     n=17 disjoint domnum 16 ways 337
  */
  printf("disjoint_domnum max ways ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_disjoint_domnum_ways[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_disjoint_domnum_ways_count[num_vertices]);
  }
  printf("\n");

  /* disjoint_domnums occurring 1,1,1,2,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,
     .                          2 3 4 5 6 7 8 9 10
  */
  printf(" occurring ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    int occurring = setsize(trees_disjoint_domnum_occurring[num_vertices],
                            TREES_DISJOINT_DOMNUM_OCCURRING_M);
    printf("%d,",occurring);
    if (0) {
      printf("\n%s\n",nautyextra_set_str(trees_disjoint_domnum_occurring[num_vertices],
                                         num_vertices));
    }
  }
  printf("\n");

  /* disjoint_domnum total for all trees of n vertices
     not in OEIS: 2,3,8,14,34,71,169,384,957,2319,5906,15040,39197,102587,272394,726743,1956008,5289951,14390559,

     mean fraction of vertices
     1.000, 1.000, 1.000, 0.933, 0.944, 0.922, 0.918, 0.908, 0.903, 0.897, 0.893, 0.889, 0.886, 0.883, 0.881, 0.879, 0.877, 0.876, 0.874, 
  */
  printf("disjoint_domnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_disjoint_domnum[num_vertices]);
  }
  printf("\n");
  printf("disjoint_domnum total fraction ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    double mean = (double) trees_total_disjoint_domnum[num_vertices]
      / ((double) num_vertices * trees_count[num_vertices]);
    printf("%.3f, ",mean);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* totdomnum min = 2 for bistars
     count 1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,
  */
  printf("totdomnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_totdomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_totdomnum_count[num_vertices]);
  }
  printf("\n");

  /* totdomnum max for path n
     2,2,2,3,4,4,5,6,6,7,8,8,9,10,10,11,
     maybe A189819
     count
     not in OEIS: 1,1,2,1,1,4,2,1,12,4,2,28,9,3,77,20,
  */
  printf("totdomnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_totdomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_totdomnum_count[num_vertices]);
  }
  printf("\n");

  /* totdomnums occurring 1,1,1,2,3,3,4,5,5,6,7,7,8,9,9,10,
     .                    2 3 4 5 6 7 8 9 10
     all values min to max ... by path with star at end maybe
  */
  printf(" num occurring ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    int occurring = setsize(trees_totdomnum_occurring[num_vertices],
                            TREES_TOTDOMNUM_OCCURRING_M);
    printf("%d,",occurring);
    if (1) {
      printf(" set %s\n",
             nautyextra_set_str(trees_totdomnum_occurring[num_vertices],
                                num_vertices));
    }
  }
  printf("\n");

  /* totdomnum total for all trees of n vertices
     not in OEIS: 2,2,4,7,16,34,80,184,461,1122,2867,7328,19147,50229,133627,357127,
     n =          2 3 4 5  6  7
  */
  printf("totdomnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_totdomnum[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */

  /* indomsets count (maximal indsets) min = 2, unique star-n
     min = 2,2,2,2,2,2,2,2,2   star
     count 1,1,1,1,1,1,1,1,1
  */
  printf("indomsets min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_min_indomsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_indomsets_count[num_vertices]);
  }
  printf("\n");

  /* indomsets count (maximal indsets) max = 2^ceil(n/2)+1-(n%2)   per Wilf
     A228693 2,2,3,4,5,8,9,16,17,32,33,64,65,128,129,256,257,512,513,1024,1025
     f(n) = if(n==0,1, n%2==0,2^(n/2-1)+1, 2^((n-1)/2));
     vector(11,n,n++; f(n)) == [2,2,3,4,5,8,9,16,17,32,33]
     count
     A057979 1,1,1,1,2,1,3,1, 4,1,5,1,6,1,7,1,8,1,9,1,10,1,
     n =     2 3 4 5 6 7 8 9 10
     odd unique
     even (n/2)-1 trees attaining
  */
  printf("maximal indsets max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_max_indomsets[num_vertices]);
  }
  printf("\n");
  printf("      count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_indomsets_count[num_vertices]);
  }
  printf("\n");

  /* indomsets (maximal indsets) total
     not in OEIS: 2,2,5,9,22,52,133,344,967,2690,7879,23308,70817,217257,678828,2138978,6821117,21919928,71038089,231715198,760725767,2511014849
     for(n=2,10,print(2^(n-1)+1))
  */
  printf("indomsets (maximal indsets) total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_indomsets[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */
  /* indomnum */

  /* indomnum min = 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1 for star
     count 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  */
  printf("indomnum min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_indomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_min_indomnum_count[num_vertices]);
  }
  printf("\n");

  /* indomnum max
     indomnum max 1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,
     count
     not in OEIS: 1,1,1,2,2,6,4,17,9,53,21,167,55,546,146,1806,
     n =          2 3 4 5 6 7 8  9   11 12  13 14  15  16   17
  */
  printf("indomnum max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_indomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_max_indomnum_count[num_vertices]);
  }
  printf("\n");

  /* indomnum total for all trees of n vertices
     not in OEIS: 1,1,3,5,13,27,66,150,379,921,2361,6028,15773,41376,110142,294403,
     n =          2 3 4 5  6  7
  */
  printf("indomnum total ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_indomnum[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */
  /* indomsets frac */

  /* indomsets_frac min
     2/3, 2/5, 2/9, 2/17, 2/33, 2/65, 2/129, 2/257, 2/513, 2/1025, 2/2049, 2/4097, 2/8193, 2/16385, 2/32769, 2/65537,
     count 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
  */
  printf("indomsets_frac min ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%.4f, ",trees_indomsets_frac_min[num_vertices]);
  }
  printf("\n");
  printf("     ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu/%Lu, ",trees_indomsets_frac_min_num[num_vertices],trees_indomsets_frac_min_den[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_indomsets_frac_min_count[num_vertices]);
  }
  printf("\n");

  /* indomsets_frac max
     2/3, 2/5, 3/9, 4/17, 5/27, 8/53, 9/81, 16/161, 17/243, 32/485, 33/729, 64/1457, 65/2187, 128/4373, 129/6561, 256/13121,
     n=2    3    4     5     6     7     8       9      10      11      12
     count 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,

     vpar n=12 0,1,2,1,4,1,6,1,8,1,10,1
     star arms length 2 which is most indomsets maybe ...
  */
  printf("indomsets_frac max ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%.4f, ",trees_indomsets_frac_max[num_vertices]);
  }
  printf("\n");
  printf("     ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu/%Lu, ",trees_indomsets_frac_max_num[num_vertices],trees_indomsets_frac_max_den[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_indomsets_frac_max_count[num_vertices]);
  }
  printf("\n");

  printf("\n");


  /* --------------------------------------------------- */

  /* total limb length
     not in OEIS: 1,2,4,7,14,27,57,120,273,619,1471,3532,8695,21624,54669,139408,
  */
  printf("total limb length ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%Lu,",trees_total_leaf_length[num_vertices]);
  }
  printf("\n");

  /* total limb length
     limb length = 0  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

     limb length = 1
     A002988 1,0,1,1,2,3,6,10,21,39,82,167,360,766,1692,3726
     trimmed

     limb length = 2
     not in OEIS: 0,1,0,1,2,4,8,18,40,92,215,511,1232,3004,7404,18395

     limb length = 3
     not in OEIS: 0,0,1,0,1,2,5,10,25,57,142,348,880,2227,5739,14843

     limb length = 4  0,0,0,1,0,1,2,5,11,27,64,161,404,1036,2675,6997,
     limb length = 5  0,0,0,0,1,0,1,2,5,11,28,66,168,423,1094,2837,
     limb length = 6  0,0,0,0,0,1,0,1,2,5,11,28,67,170,430,1113,
     limb length = 7  0,0,0,0,0,0,1,0,1,2,5,11,28,67,171,432,
     limb length = 8  0,0,0,0,0,0,0,1,0,1,2,5,11,28,67,171,

     ----
     limb length < 0  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
     limb length < 1  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,

     limb length < 2
     A002988 1,0,1,1,2,3,6,10,21,39,82,167,360,766,1692,3726,
     trimmed

     limb length < 3
     A002989 1,1,1,2,4,7,14,28,61,131,297,678,1592,3770,9096,22121,
     forbidden limb length 3

     limb length < 4
     A002990 1,1,2,2,5,9,19,38,86,188,439,1026,2472,5997,14835,36964,
     forbidden limb length 4

     limb length < 5
     A002991 1,1,2,3,5,10,21,43,97,215,503,1187,2876,7033,17510,43961,
     forbidden limb length 5

     limb length < 6
     A002992 1,1,2,3,6,10,22,45,102,226,531,1253,3044,7456,18604,46798,
     forbidden limb length 6

     limb length < 7
     not in OEIS: 1,1,2,3,6,11,22,46,104,231,542,1281,3111,7626,19034,47911,

     limb length < 8
     not in OEIS: 1,1,2,3,6,11,23,46,105,233,547,1292,3139,7693,19205,48343,

     -------
     limb length > 0
     A000055 1,1,2,3,6,11,23,47,106,235,551,1301,3159,7741,19320,48629,
     all trees

     limb length > 1
     A003426 0,1,1,2,4,8,17,37,85,196,469,1134,2799,6975,17628,44903,
     stable trees

     limb length > 2
     not in OEIS: 0,0,1,1,2,4,9,19,45,104,254,623,1567,3971,10224,26508,

     limb length > 3
     not in OEIS: 0,0,0,1,1,2,4,9,20,47,112,275,687,1744,4485,11665,

     limb length > 4
     not in OEIS: 0,0,0,0,1,1,2,4,9,20,48,114,283,708,1810,4668,

     limb length > 5  0,0,0,0,0,1,1,2,4,9,20,48,115,285,716,1831,
     limb length > 6  0,0,0,0,0,0,1,1,2,4,9,20,48,115,286,718,
     limb length > 7  0,0,0,0,0,0,0,1,1,2,4,9,20,48,115,286,
     limb length > 8  0,0,0,0,0,0,0,0,1,1,2,4,9,20,48,115,
  */
  {
    int l;
    for (l = 0; l <= TREES_LEAF_LENGTH_HISTOGRAM_MAX; l++) {
      printf("limb length = %d  ", l);
      for (num_vertices = tree_min_vertices;
           num_vertices <= tree_max_vertices;
           num_vertices++) {
        printf("%d,",trees_leaf_length_eq[num_vertices][l]);
      }
      printf("\n");
    }
    for (l = 0; l <= TREES_LEAF_LENGTH_HISTOGRAM_MAX; l++) {
      printf("limb length < %d  ", l);
      for (num_vertices = tree_min_vertices;
           num_vertices <= tree_max_vertices;
           num_vertices++) {
        printf("%d,",trees_leaf_length_lt[num_vertices][l]);
      }
      printf("\n");
    }
    for (l = 0; l <= TREES_LEAF_LENGTH_HISTOGRAM_MAX; l++) {
      printf("limb length > %d  ", l);
      for (num_vertices = tree_min_vertices;
           num_vertices <= tree_max_vertices;
           num_vertices++) {
        printf("%d,",trees_leaf_length_gt[num_vertices][l]);
      }
      printf("\n");
    }
  }
  printf("\n");

  /* --------------------------------------------------- */

  /* number of stable trees
     count 1,1,1,2,4,8,17,37,85,
     n   = 2 3 4 5 6 7 8
     A003426
  */
  printf("trees stable      ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_stable[num_vertices]);
  }
  printf("\n");

  /* number of semi-stable trees
     count 1,1,1,2,5,9,22,46,105,234,550,1300,3158,7740,19319,48628,123866,317954,823064,
     n   = 2 3 4 5 6 7 8
     not in OEIS: 2,5,9,22,46,105,234,550,1300,3158,7740,19319,

     number of not semi-stable trees
     count 0,0,1,1,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,
     n   = 2 3 4 5 6 7 8
     all path-n, and at n=7 also the asymmetric star-1,2,3
  */
  printf("trees semistable      ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_semistable[num_vertices]);
  }
  printf("\n");

  printf("trees not semistable  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",
           trees_count[num_vertices] - trees_count_semistable[num_vertices]);
  }
  printf("\n");

  /* number of completely semi-stable trees
     count 1,0,1,1,2,1,4,2,9,7,17,19,50,55,134,187,415,640,1365,
     n   = 2 3 4 5 6 7 8
     not in OEIS: 1,4,2,9,7,17,19,50,55,134,187,415,640,1365,
  */
  printf("trees completely semistable  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_count_completely_semistable[num_vertices]);
  }
  printf("\n");

  /* trees total number of semistable vertices
     total 2,1,4,6,18,39,101,230,616,1510,3962,10253,27254,72084,193845,521498,1415286,3851965,10541378,
     n   = 2 3 4 5  6  7   8   9
     not in OEIS: 18,39,101,230,616,1510,3962,10253,27254,72084,193845,521498,1415286,3851965,10541378,
  */
  printf("trees total semistable vertices  ");
  for (num_vertices = tree_min_vertices;
       num_vertices <= tree_max_vertices;
       num_vertices++) {
    printf("%d,",trees_total_semistable_vertices[num_vertices]);
  }
  printf("\n");


}


int graphs_count[64];
int graphs_count_biconnected[64];
int graphs_count_notbiconnected[64];
int graphs_max_Szeged[64];
int graphs_min_Szeged[64];
int graphs_max_Szeged_count[64];
int graphs_min_Szeged_count[64];
int graphs_total_Szeged[64];
int graphs_max_bandwidth[64];
int graphs_min_bandwidth[64];
int graphs_max_bandwidth_count[64];
int graphs_min_bandwidth_count[64];
int graphs_total_bandwidth[64];
int graphs_max_bandwidth_reg3[64];
int graphs_min_bandwidth_reg3[64];
int graphs_max_bandwidth_reg3_count[64];
int graphs_min_bandwidth_reg3_count[64];
int graphs_total_bandwidth_reg3[64];
int graphs_total_cut_edges[64];
int graphs_count_any_cut_edge[64];
int graphs_count_no_cut_edge[64];
int graphs_count_transitive[64];
int graphs_count_transitive_symmetric[64];
int graphs_count_transitive_unsymmetric[64];
int graphs_count_nontransitive[64];
unsigned long long graphs_total_cycles[64];
long graphs_max_cycles[64];
int  graphs_max_cycles_count[64];
unsigned long long graphs_total_indcycles[64];
long graphs_max_indcycles[64];
int  graphs_max_indcycles_count[64];
int graphs_total_claws[64];
int graphs_count_withclaw[64];
int graphs_count_clawfree[64];
int graphs_count_deg2free[64];
int graphs_count_deg3free[64];
int graphs_count_cubics[64];
int graphs_claws_max[64];
int graphs_claws_max_count[64];
int graphs_claws_max_cyclic[64];
int graphs_claws_max_cyclic_count[64];
int graphs_similar_total[64];
int graphs_similar_is_unique[64];
int graphs_pseudosimilar_clique_max[64];
int graphs_pseudosimilar_clique_max_count[64];
int graphs_pseudosimilar_clique_min_edges[64];
int graphs_pseudosimilar_min_edges[64];
int graphs_pseudosimilar_min_edges_count[64];
int graphs_removal_similar_clique_max[64];
int graphs_removal_similar_clique_max_count[64];
int graphs_removal_similar_is_unique[64];
int graphs_removal_similar_total[64];
int graphs_asymmetric_count[64];
int graphs_orbit_max[64];
int graphs_orbit_max_count[64];
int graphs_numorbits_total[64];
int graphs_total_median_size[64];
int graphs_median_size_max[64];
int graphs_median_size_max_count[64];
int count_regular[64];
int count_notregular[64];
int count_distance_regular[64];
int count_not_distance_regular[64];
int count_bipartites[64];
int count_not_bipartites[64];
int count_cyclics[64];
int count_acyclics[64];
int count_linegraphs[64];
int count_nonlinegraphs[64];
int count_selfcomplements[64];
int graphs_total_triangles[64];
int graphs_count_withtriangle[64];
int graphs_count_onetriangle[64];
int graphs_count_trianglefree[64];
int graphs_total_diamonds[64];
int graphs_count_withdiamond[64];
int graphs_count_onediamond[64];
int graphs_count_diamondfree[64];
int graphs_total_induced_diamonds[64];
int graphs_count_with_induced_diamond[64];
int graphs_count_one_induced_diamond[64];
int graphs_count_induced_diamond_free[64];
/* int graphs_total_complete4s[64]; */
int graphs_count_with_complete4[64];
/* int graphs_count_one_complete4[64]; */
int graphs_count_complete4_free[64];
int sum_girths[64];
int sum_maxcliques[64];
int total_clique_number[64];
unsigned long long graphs_total_edges[64];
int graphs_max_indnum[64];
int graphs_min_indnum[64];
int graphs_max_indnum_count[64];
int graphs_min_indnum_count[64];
unsigned long long graphs_total_indnum[64];
int graphs_indnum_ways_max[64];
int graphs_indnum_ways_max_count[64];
int graphs_half_indnum[64];
int graphs_indcomp[64];
int graphs_max_domnum[64];
int graphs_min_domnum[64];
int graphs_max_domnum_count[64];
int graphs_min_domnum_count[64];
int graphs_domnum_ways_max[64];
int graphs_domnum_ways_max_count[64];
unsigned long long graphs_total_domnum[64];
#define DOMNUM_COUNT_MAX  64
int graphs_domnum_count[64][DOMNUM_COUNT_MAX+1];
int graphs_domnum_perfect[64];
int graphs_domnum_perfect_cyclic[64];
int graphs_domnum_perfect_smaller[64];
int cyclics_max_domnum[64];
int cyclics_min_domnum[64];
int cyclics_max_domnum_count[64];
int cyclics_min_domnum_count[64];
unsigned long long cyclics_total_domnum[64];
unsigned long long cyclics_total_diameter[64];
unsigned long long graphs_total_diameter[64];
int graphs_max_totdomnum[64];
int graphs_min_totdomnum[64];
int graphs_max_totdomnum_count[64];
int graphs_min_totdomnum_count[64];
unsigned long long graphs_max_totdomnum_count_ways[64];
unsigned long long graphs_min_totdomnum_count_ways[64];
unsigned long long graphs_total_totdomnum[64];
int graphs_diameter_le_3_totdomnum_gt_2[64];
#define DIAMETER_COUNT_MAX  3
int graphs_diameter_count[64][DIAMETER_COUNT_MAX+1];
int cyclics_diameter_count[64][DIAMETER_COUNT_MAX+1];
int graphs_max_indomnum[64];
int graphs_min_indomnum[64];
int graphs_max_indomnum_count[64];
int graphs_min_indomnum_count[64];
unsigned long long graphs_max_indomnum_count_ways[64];
unsigned long long graphs_min_indomnum_count_ways[64];
unsigned long long graphs_total_indomnum[64];
int graphs_indomnum_ways_max[64];
int graphs_indomnum_ways_max_count[64];
int graphs_indomnum_eq_domnum[64];
int graphs_indomnum_eq_domnum_withclaw[64];
int graphs_indomnum_eq_domnum_clawfree[64];
int graphs_max_invdomnum[64];
int graphs_min_invdomnum[64];
int graphs_max_invdomnum_count[64];
int graphs_min_invdomnum_count[64];
unsigned long long graphs_max_invdomnum_count_ways[64];
unsigned long long graphs_min_invdomnum_count_ways[64];
unsigned long long graphs_total_invdomnum[64];
int graphs_invdomnum_ways_max[64];
int graphs_invdomnum_ways_max_count[64];
int graphs_invdomnum_eq_domnum[64];
int graphs_invdomnum_eq_domnum_withclaw[64];
int graphs_invdomnum_eq_domnum_clawfree[64];
int graphs_count_wellcovered[64];
int graphs_count_stable[64];
int graphs_total_semistable_vertices[64];
int graphs_count_completely_semistable[64];
int graphs_count_completely_semistable_irregular[64];
int graphs_count_completely_semistable_intransitive[64];

void
my_geng_outproc (FILE *outfile, graph *g, int n)
{
  graphs_count[n]++;

  int m = SETWORDSNEEDED(n);
  int is_cyclic    = nautyextra_is_cyclic(g,m,n);
  int is_linegraph = nautyextra_is_linegraph(g,m,n);
  int num_edges    = nautyextra_num_edges(g,m,n);
  nauty_to_sg(g,&sg,m,n);

  if (option_verbose >= 2) {
    printf("graph %p m,n=%d,%d   cyclic=%d\n", g,m,n, is_cyclic);
    nautyextra_print_matrix(g,m,n);
    /* nautyextra_view(g,m,n); */
  }

  count_cyclics[n]  += is_cyclic;
  count_acyclics[n] += ! is_cyclic;

  count_linegraphs[n] += is_linegraph;
  count_nonlinegraphs[n] += ! is_linegraph;
  if (0 && n==5 && is_linegraph) {
    printf("%s", ntog6((graph*)g,m,n));
  }

  int regdegree = nautyextra_regular_degree(g,m,n);
  int is_regular = (regdegree >= 0);
  count_regular[n]    += is_regular;
  count_notregular[n] += ! is_regular;

  int bandwidth = -1;
  if (n <= 6) {
    bandwidth = nautyextra_bandwidth(g,m,n);
    if (graphs_min_bandwidth[n] == 0 || bandwidth < graphs_min_bandwidth[n]) {
      graphs_min_bandwidth[n] = bandwidth;
      graphs_min_bandwidth_count[n] = 0;
    }
    if (bandwidth == graphs_min_bandwidth[n])
      graphs_min_bandwidth_count[n]++;
    if (graphs_max_bandwidth[n] == 0 || bandwidth > graphs_max_bandwidth[n]) {
      graphs_max_bandwidth[n] = bandwidth;
      graphs_max_bandwidth_count[n] = 0;
    }
    if (bandwidth == graphs_max_bandwidth[n])
      graphs_max_bandwidth_count[n]++;
    graphs_total_bandwidth[n] += bandwidth;

    if (0 && n==6 && bandwidth==5) {
      printf("bandwidth %d\n", bandwidth);
      nautyextra_view(g,m,n);
    }
  }
  if (n <= 8 && regdegree == 3) {
    if (bandwidth == -1)
      bandwidth = nautyextra_bandwidth(g,m,n);

    if (graphs_min_bandwidth_reg3[n] == 0 || bandwidth < graphs_min_bandwidth_reg3[n]) {
      graphs_min_bandwidth_reg3[n] = bandwidth;
      graphs_min_bandwidth_reg3_count[n] = 0;
    }
    if (bandwidth == graphs_min_bandwidth_reg3[n])
      graphs_min_bandwidth_reg3_count[n]++;
    if (graphs_max_bandwidth_reg3[n] == 0 || bandwidth > graphs_max_bandwidth_reg3[n]) {
      graphs_max_bandwidth_reg3[n] = bandwidth;
      graphs_max_bandwidth_reg3_count[n] = 0;
    }
    if (bandwidth == graphs_max_bandwidth_reg3[n])
      graphs_max_bandwidth_reg3_count[n]++;
    graphs_total_bandwidth_reg3[n] += bandwidth;

    if (0 && n==8 && bandwidth==4) {
      printf("bandwidth %d\n", bandwidth);
      nautyextra_view(g,m,n);
    }
  }

  {
    unsigned long long Szeged = nautyextra_Szeged_index(g,m,n);
    graphs_total_Szeged[n] += Szeged;
    if (graphs_min_Szeged[n] == 0 || Szeged < graphs_min_Szeged[n]) {
      graphs_min_Szeged[n] = Szeged;
      graphs_min_Szeged_count[n] = 0;
    }
    graphs_min_Szeged_count[n] += (Szeged == graphs_min_Szeged[n]);

    if (graphs_max_Szeged[n] == 0 || Szeged > graphs_max_Szeged[n]) {
      graphs_max_Szeged[n] = Szeged;
      graphs_max_Szeged_count[n] = 0;
    }
    graphs_max_Szeged_count[n] += (Szeged == graphs_max_Szeged[n]);

    if (0 && n==6 && Szeged == 81) {     /* bipartite 3,3 */
      nautyextra_view(g,m,n);
    }
  }

  int transitivity;  /* 0=not, 1=not symmetric, 2=symmetric */
  int is_transitive;
  {
    const int m = SETWORDSNEEDED(graph_max_vertices);
    setword g2[graph_max_vertices][m];
    transitivity = istransitive(g,m,n, (graph*)g2);
    is_transitive = (transitivity!=0);
    graphs_count_transitive[n]             += is_transitive;
    graphs_count_nontransitive[n]          += ! is_transitive;
    graphs_count_transitive_symmetric[n]   += (transitivity == 2);
    graphs_count_transitive_unsymmetric[n] += (transitivity == 1);
  }

  graphs_count_stable[n] += nautyextra_is_stable(g,m,n);
  {
    int completely_semistable = 1;  /* stable at all vertices */
    int v;
    for (v=0; v<n; v++) {
      if (nautyextra_is_semistable_at_vertex (g,m,n,v)) {
        graphs_total_semistable_vertices[n]++;
      } else {
        completely_semistable = 0;
      }
    }
    /* Per Holton and Grant, all regulars are completely semi-stable. */
    if (is_regular) assert(completely_semistable);
    graphs_count_completely_semistable[n] += completely_semistable;

    int completely_semistable_irregular
      = (completely_semistable && !is_regular);
    graphs_count_completely_semistable_irregular[n]
      += completely_semistable_irregular;

    if (0 && completely_semistable_irregular && n <= 7) {
      printf("completely semistable irregular %d\n", regdegree);
      nautyextra_view_synchronous = 1;
      nautyextra_view(g,m,n);
    }
    if (0 && completely_semistable_irregular && n <= 6) {
      printf("%s", ntog6((graph*)g,m,n));
    }

    int completely_semistable_intransitive
      = (completely_semistable && !is_transitive);
    graphs_count_completely_semistable_intransitive[n]
      += completely_semistable_intransitive;

}

  int intersection_array[2*graph_max_vertices];
  int is_distance_regular
    = nautyextra_is_distance_regular(g,m,n, intersection_array);
  count_distance_regular[n]     += is_distance_regular;
  count_not_distance_regular[n] += ! is_distance_regular;
  if (0 && is_distance_regular) {
    printf ("n=%d ",n);
    int i;
    for (i = 0; i < n; i++) {
      if (intersection_array[i]==-1) break;
      if (i) printf (",");
      printf ("%d",intersection_array[i]);
    }
    printf ("; ");
    for (i = 0; i < n; i++) {
      if (intersection_array[i+n]==-1) break;
      if (i) printf (",");
      printf ("%d",intersection_array[i+n]);
    }
    printf ("\n");
  }

  int median_size = nautyextra_median_size(g,m,n, NULL);
  graphs_total_median_size[n] += median_size;
  if (graphs_median_size_max[n] == 0
      || median_size > graphs_median_size_max[n]) {
    graphs_median_size_max[n] = median_size;
    graphs_median_size_max_count[n] = 0;
  }
  if (median_size == graphs_median_size_max[n])
    graphs_median_size_max_count[n]++;
  if (0 && (n==8||n==9) && (median_size==n) != is_regular) {
    printf("%s  n=%d regular %d != median=%d max, cf transitivity %d\n",
           ntog6((graph*)g,m,n), n, is_regular, median_size, transitivity);
    /* nautyextra_view(g,m,n); */
  }

  int degrees[graph_max_vertices];
  {
    assert(n <= graph_max_vertices);
    nautyextra_degrees_array(g,m,n, degrees);
    int count_deg2 = 0;
    int count_deg3 = 0;
    int v;
    for (v = 0; v < n; v++) {
      int deg = degrees[v];
      count_deg2 += (deg==2);
      count_deg3 += (deg==3);
    }
    graphs_count_deg2free[n] += (count_deg2 == 0);
    graphs_count_deg3free[n] += (count_deg3 == 0);
    graphs_count_cubics[n] += (count_deg3 == n);
  }

  if (0) {
    /* Bondy and Murty exercies 7.3.1
       Many graphs with this degree sequence. */
    int degrees_sorted[graph_max_vertices];
    memcpy(degrees_sorted, degrees, sizeof(degrees));
    qsort(degrees_sorted,n,sizeof(degrees_sorted[0]),compare_int);

    const int want_degrees[] = {2, 4,4, 5,5,5,5, 6,6};
    if (memcmp(degrees_sorted,want_degrees,sizeof(want_degrees))==0) {
      printf("degrees found %s", ntog6((graph*)g,m,n));
    }
  }

  if (n <= 8) {
    struct nautyextra_similar_info info;
    nautyextra_similar(g,m,n, &info);

    int similar_pairs
      = nautyextra_num_edges(info.similar,m,n);
    int similar_is_unique = (similar_pairs == 1);
    graphs_similar_is_unique[n] += similar_is_unique;
    graphs_similar_total[n] += similar_pairs;

    if (graphs_pseudosimilar_clique_max[n] == 0
        || info.pseudosimilar_clique_size > graphs_pseudosimilar_clique_max[n]){
      graphs_pseudosimilar_clique_max[n] = info.pseudosimilar_clique_size;
      graphs_pseudosimilar_clique_max_count[n] = 0;
    }
    if (info.pseudosimilar_clique_size == graphs_pseudosimilar_clique_max[n])
      graphs_pseudosimilar_clique_max_count[n]++;
    if (0 && n==8 && info.pseudosimilar_clique_size > 1) {
      printf("%s  pseudosimilar\n", ntog6((graph*)g,m,n));
      if (0 && is_linegraph) {
        printf("%s  linegraph %d\n", ntog6((graph*)g,m,n), is_linegraph);
        nautyextra_view(g,m,n);
      }
    }

    if (info.pseudosimilar_clique_size >= 2) {
      if (graphs_pseudosimilar_min_edges[n] == 0
          || num_edges < graphs_pseudosimilar_min_edges[n]){
        graphs_pseudosimilar_min_edges[n] = num_edges;
        graphs_pseudosimilar_min_edges_count[n] = 0;
      }
      if (num_edges == graphs_pseudosimilar_min_edges[n])
        graphs_pseudosimilar_min_edges_count[n]++;
    }

    int removal_similar_clique_size
      = nautyextra_clique_number(info.removal_similar,m,n);
    if (graphs_removal_similar_clique_max[n] == 0
        || removal_similar_clique_size > graphs_removal_similar_clique_max[n]){
      graphs_removal_similar_clique_max[n] = removal_similar_clique_size;
      graphs_removal_similar_clique_max_count[n] = 0;
    }
    if (removal_similar_clique_size == graphs_removal_similar_clique_max[n])
      graphs_removal_similar_clique_max_count[n]++;

    int removal_similar_pairs
      = nautyextra_num_edges(info.removal_similar,m,n);
    int removal_similar_is_unique = (removal_similar_pairs == 1);
    graphs_removal_similar_is_unique[n] += removal_similar_is_unique;
    graphs_removal_similar_total[n] += removal_similar_pairs;
    if (0 && n==5 && removal_similar_is_unique) {
      nautyextra_view(g,m,n);
    }

    if (graphs_orbit_max[n] == 0 || info.orbit_max > graphs_orbit_max[n]){
      graphs_orbit_max[n] = info.orbit_max;
      graphs_orbit_max_count[n] = 0;
    }
    if (info.orbit_max == graphs_orbit_max[n])
      graphs_orbit_max_count[n]++;
    graphs_numorbits_total[n] += info.numorbits;

    int is_asymmetric = (info.numorbits == n);
    graphs_asymmetric_count[n] += is_asymmetric;

    if (0 && n==6 && is_asymmetric) {
      printf("%s", ntog6((graph*)g,m,n));
      int degrees_sorted[graph_max_vertices];
      memcpy(degrees_sorted, degrees, sizeof(degrees));
      qsort(degrees_sorted,n,sizeof(degrees_sorted[0]),compare_int);
      printf("  degrees ");
      int v;
      for (v = 0; v < n; v++) printf("%d,",degrees_sorted[v]);
      printf("\n");
    }
  }

  int is_clawfree;
  int is_withclaw;
  {
    int claws = nautyextra_claw_count(g,m,n);
    graphs_total_claws[n] += claws;
    is_withclaw = (claws != 0);
    is_clawfree = (claws == 0);
    graphs_count_withclaw[n] += is_withclaw;
    graphs_count_clawfree[n] += is_clawfree;

    if (graphs_claws_max[n] == 0 || claws > graphs_claws_max[n]) {
      graphs_claws_max[n] = claws;
      graphs_claws_max_count[n] = 0;
    }
    if (claws == graphs_claws_max[n])
      graphs_claws_max_count[n]++;

    if (is_cyclic) {
      if (graphs_claws_max_cyclic[n] == 0 || claws > graphs_claws_max_cyclic[n]) {
        graphs_claws_max_cyclic[n] = claws;
        graphs_claws_max_cyclic_count[n] = 0;
      }
      if (claws == graphs_claws_max_cyclic[n])
        graphs_claws_max_cyclic_count[n]++;
    }

    if (0 && claws==112 && n==10) {
      nautyextra_view_synchronous = 0;
      nautyextra_view(g,m,n);
      printf("%d  %s%s", is_cyclic, GRAPH6_HEADER, ntog6((graph*)g,m,n));
    }
  }

  if (n <= 9) {
    if (isbipartite(g,m,n))
      count_bipartites[n]++;
    else
      count_not_bipartites[n]++;
  }



  if (n <= 9) {
    int is_selfcomplement = nautyextra_is_selfcomplement (g,m,n);
    if (is_selfcomplement)
      count_selfcomplements[n]++;
  }

  if (n <= 9) {
    int triangles = numtriangles(g,m,n);
    graphs_total_triangles[n]    += triangles;
    graphs_count_withtriangle[n] += (triangles != 0);
    graphs_count_trianglefree[n] += (triangles == 0);
    graphs_count_onetriangle[n]  += (triangles == 1);
  }

  if (n <= 9) {
    int diamonds = nautyextra_diamond_count(g,m,n, 0);
    graphs_total_diamonds[n]    += diamonds;
    graphs_count_withdiamond[n] += (diamonds != 0);
    graphs_count_diamondfree[n] += (diamonds == 0);
    graphs_count_onediamond[n]  += (diamonds == 1);

    int induced_diamonds = nautyextra_diamond_count(g,m,n, 1);
    graphs_total_induced_diamonds[n]    += induced_diamonds;
    graphs_count_with_induced_diamond[n] += (induced_diamonds != 0);
    graphs_count_induced_diamond_free[n] += (induced_diamonds == 0);
    graphs_count_one_induced_diamond[n]  += (induced_diamonds == 1);

    int has_complete4 = (diamonds != induced_diamonds);
    graphs_count_with_complete4[n] +=   has_complete4;
    graphs_count_complete4_free[n] += ! has_complete4;
  }

  {
    int gg = girth(g,m,n);
    sum_girths[n] += gg;
  }
  {
    int cc = maxcliques(g,m,n);  /* number of maximal cliques */
    sum_maxcliques[n] += cc;
  }

  {
    int bic = isbiconnected(g,m,n);
    graphs_count_biconnected[n]    += (bic != 0);
    graphs_count_notbiconnected[n] += (bic == 0);
  }
  {
    int cut_edge_count = nautyextra_cut_edge_count(g,m,n);
    graphs_total_cut_edges[n] += cut_edge_count;
    graphs_count_no_cut_edge[n] += (cut_edge_count == 0);
    graphs_count_any_cut_edge[n] += (cut_edge_count != 0);
  }

  if (n <= 9) {
    int cycles = cyclecount(g,m,n);
    graphs_total_cycles[n] += cycles;

    if (graphs_max_cycles[n] == 0 || cycles > graphs_max_cycles[n]) {
      graphs_max_cycles[n] = cycles;
      graphs_max_cycles_count[n] = 0;
    }
    if (cycles == graphs_max_cycles[n])
      graphs_max_cycles_count[n]++;
  }
  if (n <= 9) {
    int indcycles = indcyclecount(g,m,n);
    graphs_total_indcycles[n] += indcycles;

    if (graphs_max_indcycles[n] == 0 || indcycles > graphs_max_indcycles[n]) {
      graphs_max_indcycles[n] = indcycles;
      graphs_max_indcycles_count[n] = 0;
    }
    if (indcycles == graphs_max_indcycles[n])
      graphs_max_indcycles_count[n]++;
  }

  graphs_total_edges[n] += num_edges;

  int radius, diameter;
  diamstats(g, m,n, &radius, &diameter);
  {
    graphs_total_diameter[n] += diameter;
    if (is_cyclic)
      cyclics_total_diameter[n] += diameter;
    if (diameter <= DIAMETER_COUNT_MAX) {
      graphs_diameter_count[n][diameter]++;
      if (is_cyclic)
        cyclics_diameter_count[n][diameter]++;
    }
  }

  unsigned long edges = 99999999;
  int mindeg=-1, mincount=-1, maxdeg=-1, maxcount=-1, eulerian=-1;
  degstats(g,m,n, &edges,&mindeg,&mincount,&maxdeg,&maxcount,&eulerian);

  int indnum = -1;
  if (n <= 10) {
    struct nautyextra_indnum_and_count  info;
    nautyextra_indnum_and_count(g,m,n, &info);
    indnum = info.indnum;

    /* printf("n=%d  indnum = %d\n", n, indnum); */
    graphs_total_indnum[n] += info.indnum;
    if (graphs_min_indnum[n] == 0 || info.indnum < graphs_min_indnum[n]) {
      graphs_min_indnum[n] = info.indnum;
      graphs_min_indnum_count[n] = 0;
    }
    if (info.indnum == graphs_min_indnum[n])
      graphs_min_indnum_count[n]++;

    if (graphs_max_indnum[n] == 0 || info.indnum > graphs_max_indnum[n]) {
      graphs_max_indnum[n] = info.indnum;
      graphs_max_indnum_count[n] = 0;
    }
    if (info.indnum == graphs_max_indnum[n])
      graphs_max_indnum_count[n]++;

    if (graphs_indnum_ways_max[n] == 0 || info.count > graphs_indnum_ways_max[n]) {
      graphs_indnum_ways_max[n] = info.count;
      graphs_indnum_ways_max_count[n] = 0;
    }
    if (info.count == graphs_indnum_ways_max[n])
      graphs_indnum_ways_max_count[n]++;
    if (0 && ((n==5 && info.count == 5)
              || (n==6 && info.count == 8)
              || (n==7 && info.count == 11)
              || (n==8 && info.count == 15)
              || (n==9 && info.count == 22))) {
      printf("n=%d indnum %d ways %Lu\n",
             n, info.indnum, info.count);
      printf("%s%s", GRAPH6_HEADER, ntog6((graph*)g,m,n));
      /* nautyextra_view(g,m,n); */
    }

    if (2*info.indnum == n)
      graphs_half_indnum[n]++;

    if (has_indcomp(g,m,n)) {
      graphs_indcomp[n]++;
      /* if (n==4) { */
      /*   nautyextra_view(g,m,n); */
      /*   nautyextra_print_matrix(g,m,n); */
      /* } */
    }
  }

  int clique_number = nautyextra_clique_number(g,m,n);
  total_clique_number[n] += clique_number;
  if (0) {
    int k,l;
    for (k = 1; k <= Ramsey_max_k; k++) {
      for (l = 1; l <= Ramsey_max_l; l++) {
        int r = Ramsey_number(k,l);
        if (n == r-1 && clique_number < k && indnum < l) {
          printf("Ramsey n=%d %d,%d\n", n,k,l);

          if (0 && k==3 && l==3) {
            nautyextra_view(g,m,n);
            printf("%s%s", GRAPH6_HEADER, ntog6((graph*)g,m,n));
          }
        }
      }
    }
  }

  int domnum = -1;
  if (n <= 9) {
    struct nautyextra_domnum_and_count  info;
    nautyextra_domnum_and_count_sg(&sg, &info);
    domnum = info.domnum;

    /* if (n==5 && info.domnum==1) */
    /*   nautyextra_view(g,m,n); */

    graphs_total_domnum[n] += info.domnum;
    graphs_domnum_count[n][info.domnum]++;
    if (graphs_min_domnum[n] == 0 || info.domnum < graphs_min_domnum[n]) {
      graphs_min_domnum[n] = info.domnum;
      graphs_min_domnum_count[n] = 0;
    }
    if (info.domnum == graphs_min_domnum[n])
      graphs_min_domnum_count[n]++;
    if (graphs_max_domnum[n] == 0 || info.domnum > graphs_max_domnum[n]) {
      graphs_max_domnum[n] = info.domnum;
      graphs_max_domnum_count[n] = 0;
    }
    if (info.domnum == graphs_max_domnum[n])
      graphs_max_domnum_count[n]++;

    /* domnum is by domnum many vertices each covering maxdeg+1 */
    int is_domnum_perfect = (info.domnum*(maxdeg+1) == n);
    graphs_domnum_perfect[n] += is_domnum_perfect;
    graphs_domnum_perfect_cyclic[n] += is_domnum_perfect && is_cyclic;
    graphs_domnum_perfect_smaller[n] += is_domnum_perfect && maxdeg<n-1;
    if (0 && n==6 && is_domnum_perfect && maxdeg<n-1) {
      nautyextra_view(g,m,n);
    }
    if (0 && n==8 && is_domnum_perfect && maxdeg<n-1) {
      printf("n=%d  domnum=%d maxdeg=%d\n", n, info.domnum, maxdeg);
    }

    if (graphs_domnum_ways_max[n] == 0 || info.count > graphs_domnum_ways_max[n]) {
      graphs_domnum_ways_max[n] = info.count;
      graphs_domnum_ways_max_count[n] = 0;
    }
    if (info.count == graphs_domnum_ways_max[n])
      graphs_domnum_ways_max_count[n]++;
    if (0 && ((n==4 && info.count == 6)
              || (n==5 && info.count == 9)
              || (n==6 && info.count == 15)
              || (n==7 && info.count == 22)
              || (n==8 && info.count == 36))) {
      printf("n=%d domnum %d ways %Lu\n",
             n, info.domnum, info.count);
      printf("%s%s", GRAPH6_HEADER, ntog6((graph*)g,m,n));
      /* nautyextra_view(g,m,n); */
    }

    if (is_cyclic) {
      cyclics_total_domnum[n] += info.domnum;
      if (cyclics_min_domnum[n] == 0 || info.domnum < cyclics_min_domnum[n]) {
        cyclics_min_domnum[n] = info.domnum;
        cyclics_min_domnum_count[n] = 0;
      }
      if (info.domnum == cyclics_min_domnum[n])
        cyclics_min_domnum_count[n]++;
      if (cyclics_max_domnum[n] == 0 || info.domnum > cyclics_max_domnum[n]) {
        cyclics_max_domnum[n] = info.domnum;
        cyclics_max_domnum_count[n] = 0;
        DEBUG(printf ("new high cyclics_max_domnum[%d] = %d\n",
                      n, info.domnum));
      }
      if (info.domnum == cyclics_max_domnum[n])
        cyclics_max_domnum_count[n]++;

      if (0 && n==8 && info.domnum==4) {
        /* nautyextra_view_synchronous = 1; */
        /* nautyextra_view(g,m,n); */
        printf("%s%s", GRAPH6_HEADER, ntog6((graph*)g,m,n));
      }
    }
  }

  if (n <= 10) {
    struct nautyextra_totdomnum_and_count  info;
    nautyextra_totdomnum_and_count_sg(&sg, &info);

    if (0 && n==5 && !(info.totdomnum==2 && diameter <= 3)) {
      printf("%s%s", GRAPH6_HEADER, ntog6((graph*)g,m,n));
      printf("  totdomnum %d ways %Lu  cyclic %d diameter %d\n",
             info.totdomnum, info.count, is_cyclic, diameter);
      nautyextra_view(g,m,n);
    }
    if (0 && n==8 && info.totdomnum==5) {
      printf("%s%s", GRAPH6_HEADER, ntog6((graph*)g,m,n));
      printf("  totdomnum %d ways %Lu  cyclic %d\n",
             info.totdomnum, info.count, is_cyclic);
      nautyextra_view(g,m,n);
    }

    if (info.totdomnum == 2)
      assert(diameter <= 3);

    graphs_total_totdomnum[n] += info.totdomnum;
    if (graphs_min_totdomnum[n] == 0 || info.totdomnum < graphs_min_totdomnum[n]) {
      graphs_min_totdomnum[n] = info.totdomnum;
      graphs_min_totdomnum_count[n] = 0;
      graphs_min_totdomnum_count_ways[n] = 0;
    }
    if (info.totdomnum == graphs_min_totdomnum[n]) {
      graphs_min_totdomnum_count[n]++;
      graphs_min_totdomnum_count_ways[n] += info.count;
    }
    if (graphs_max_totdomnum[n] == 0 || info.totdomnum > graphs_max_totdomnum[n]) {
      graphs_max_totdomnum[n] = info.totdomnum;
      graphs_max_totdomnum_count[n] = 0;
      graphs_max_totdomnum_count_ways[n] = 0;
    }
    if (info.totdomnum == graphs_max_totdomnum[n]) {
      graphs_max_totdomnum_count[n]++;
      graphs_max_totdomnum_count_ways[n] += info.count;
    }

    if (diameter<=3 && info.totdomnum > 2) {
      graphs_diameter_le_3_totdomnum_gt_2[n]++;
    }
  }

  {
    struct nautyextra_indomnum_and_count  info;
    nautyextra_indomnum_and_count_sg(&sg, &info);

    graphs_total_indomnum[n] += info.indomnum;
    if (graphs_min_indomnum[n] == 0 || info.indomnum < graphs_min_indomnum[n]) {
      graphs_min_indomnum[n] = info.indomnum;
      graphs_min_indomnum_count[n] = 0;
      graphs_min_indomnum_count_ways[n] = 0;
    }
    if (info.indomnum == graphs_min_indomnum[n]) {
      graphs_min_indomnum_count[n]++;
      graphs_min_indomnum_count_ways[n] += info.count;
    }
    if (graphs_max_indomnum[n] == 0 || info.indomnum > graphs_max_indomnum[n]) {
      graphs_max_indomnum[n] = info.indomnum;
      graphs_max_indomnum_count[n] = 0;
      graphs_max_indomnum_count_ways[n] = 0;
    }
    if (info.indomnum == graphs_max_indomnum[n]) {
      graphs_max_indomnum_count[n]++;
      graphs_max_indomnum_count_ways[n] += info.count;
    }

    if (0 && info.indomnum==1) {
      printf("n=%d  indomnum %d ways %Lu  cyclic %d\n",
             n, info.indomnum, info.count, is_cyclic);
      nautyextra_view_synchronous = 1;
      nautyextra_view(g,m,n);
    }
    if (0 && n==9 && info.indomnum==5) {
      printf("n=%d  indomnum %d ways %Lu  cyclic %d\n",
             n, info.indomnum, info.count, is_cyclic);
      nautyextra_view(g,m,n);
    }

    if (graphs_indomnum_ways_max[n] == 0 || info.count > graphs_indomnum_ways_max[n]) {
      graphs_indomnum_ways_max[n] = info.count;
      graphs_indomnum_ways_max_count[n] = 0;
    }
    if (info.count == graphs_indomnum_ways_max[n])
      graphs_indomnum_ways_max_count[n]++;

    int indomnum_eq_domnum = (info.indomnum == domnum);
    graphs_indomnum_eq_domnum[n] += indomnum_eq_domnum;
    graphs_indomnum_eq_domnum_withclaw[n]
      += (indomnum_eq_domnum && is_withclaw);
    graphs_indomnum_eq_domnum_clawfree[n]
      += (indomnum_eq_domnum && is_clawfree);

    /* Allan and Laskar all claw-free graphs have indomnum==domnum */
    if (is_clawfree && domnum>=0)
      assert(indomnum_eq_domnum);

    if (0 && n==5 && indomnum_eq_domnum && is_withclaw) {
      printf("n=%d  indomnum %d domnum %d  cyclic %d\n",
             n, info.indomnum, domnum, is_cyclic);
      nautyextra_view(g,m,n);
    }
  }

  {
    struct nautyextra_invdomnum_and_count  info;
    nautyextra_invdomnum_and_count_sg(&sg, &info);

    graphs_total_invdomnum[n] += info.invdomnum;
    if (graphs_min_invdomnum[n] == 0 || info.invdomnum < graphs_min_invdomnum[n]) {
      graphs_min_invdomnum[n] = info.invdomnum;
      graphs_min_invdomnum_count[n] = 0;
      graphs_min_invdomnum_count_ways[n] = 0;
    }
    if (info.invdomnum == graphs_min_invdomnum[n]) {
      graphs_min_invdomnum_count[n]++;
      graphs_min_invdomnum_count_ways[n] += info.count;
    }
    if (graphs_max_invdomnum[n] == 0 || info.invdomnum > graphs_max_invdomnum[n]) {
      graphs_max_invdomnum[n] = info.invdomnum;
      graphs_max_invdomnum_count[n] = 0;
      graphs_max_invdomnum_count_ways[n] = 0;
    }
    if (info.invdomnum == graphs_max_invdomnum[n]) {
      graphs_max_invdomnum_count[n]++;
      graphs_max_invdomnum_count_ways[n] += info.count;
    }

    if (0 && info.invdomnum==1) {
      printf("n=%d  invdomnum %d ways %Lu  cyclic %d\n",
             n, info.invdomnum, info.count, is_cyclic);
      nautyextra_view_synchronous = 1;
      nautyextra_view(g,m,n);
    }
    if (0 && n==9 && info.invdomnum==5) {
      printf("n=%d  invdomnum %d ways %Lu  cyclic %d\n",
             n, info.invdomnum, info.count, is_cyclic);
      nautyextra_view(g,m,n);
    }

    if (graphs_invdomnum_ways_max[n] == 0 || info.count > graphs_invdomnum_ways_max[n]) {
      graphs_invdomnum_ways_max[n] = info.count;
      graphs_invdomnum_ways_max_count[n] = 0;
    }
    if (info.count == graphs_invdomnum_ways_max[n])
      graphs_invdomnum_ways_max_count[n]++;

    int invdomnum_eq_domnum = (info.invdomnum == domnum);
    graphs_invdomnum_eq_domnum[n] += invdomnum_eq_domnum;
    graphs_invdomnum_eq_domnum_withclaw[n]
      += (invdomnum_eq_domnum && is_withclaw);
    graphs_invdomnum_eq_domnum_clawfree[n]
      += (invdomnum_eq_domnum && is_clawfree);

    /* Allan and Laskar all claw-free graphs have indomnum==domnum
       also invdomnum ?
    */
    if (is_clawfree && domnum>=0)
      assert(invdomnum_eq_domnum);

    if (0 && n==6 && info.invdomnum == 3) {
      printf("n=%d  invdomnum %d domnum %d  cyclic %d  hog %d\n",
             n, info.invdomnum, domnum, is_cyclic,
             nautyextra_hog_grep(g,m,n));
      nautyextra_view(g,m,n);
    }
  }

  graphs_count_wellcovered[n] += nautyextra_is_well_covered_sg(&sg);
}

void
try_graphs (void)
{
  printf("geng\n");
  /* option_verbose = 1; */

  int num_vertices;
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    if (option_verbose >= 1) {
      printf ("n=%d geng\n", num_vertices);
    }
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = "geng";
    geng_argv[geng_argc++] = "-q";     /* quiet */
    geng_argv[geng_argc++] = "-c";     /* connected graphs */
    /* geng_argv[geng_argc++] = "-d3"; */   /* min degree */
    /* geng_argv[geng_argc++] = "-D3"; */   /* max degree */
    geng_argv[geng_argc++] = num_vertices_str;
    geng_argv[geng_argc] = NULL;
    my_geng_main(geng_argc,geng_argv);
  }

  /* --------------------------------------------------- */

  /* number of connected graphs
     A001349 1,2,6,21,112,853,11117,261080,11716571,
     .       2 3 4  5   6   7     8      9       10
  */
  printf("count graphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count[num_vertices]);
  }
  printf("\n");

  /* number of cyclic graphs
     A241841 0,1,4,18,106,842,11094,261033,11716465,
     .       2 3 4  5
  */
  printf("count cyclic  graphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_cyclics[num_vertices]);
  }
  printf("\n");

  /* number of acyclic   graphs, ie trees
     A000055 1,1,2,3,6,11,23,47,106
     .       2 3 4 5
  */
  printf("count acyclic graphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_acyclics[num_vertices]);
  }
  printf("\n");

  /* number of bipartite graphs
     A005142 1,1,3,5,17,44,182,730,
     .       2 3 4 5
  */
  printf("count bipartite graphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_bipartites[num_vertices]);
  }
  printf("\n");

  /* number of non-bipartite graphs
     A157051 0,1,3,16,95,809,10935,260350,
     .       2 3 4  5
  */
  printf("count not-bipartite graphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_not_bipartites[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* biconnected count    0,1,3,10,56,468,7123,194066,
     A002218
  */
  printf("biconnected count    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_biconnected[num_vertices]);
  }
  printf("\n");

  /* nonbiconnected count 1,1,3,11,56,385,3994,67014,
     A052442  1,1,3,11,56,385,3994,67014,
  */
  printf("nonbiconnected count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_notbiconnected[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* bandwidth min = 1 path
     bandwidth = 1,1,1,1,1,0,0,
     count       1,1,1,1,1,0,0,
  */
  printf("bandwidth min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_bandwidth[num_vertices]);
  }
  printf("\n");
  printf("             count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_bandwidth_count[num_vertices]);
  }
  printf("\n");

  /* bandwidth max = n-1 complete-n always has 0,n-1 adjacent
     max   1,2,3,4,5,0,0,
     count 1,1,1,1,1,0,0,
     n =   2 3 4 5 6 7 8
  */
  printf("bandwidth max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_bandwidth[num_vertices]);
  }
  printf("\n");
  printf("             count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_bandwidth_count[num_vertices]);
  }
  printf("\n");

  /* bandwidth total
     not in OEIS: 1,3,12,50,313,0,0,
     n =          2 3  4  5   6
  */
  printf("bandwidth total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_bandwidth[num_vertices]);
  }
  printf("\n");

  /*-------*/

  /* bandwidth_reg3 min
     bandwidth = 0,0,3,0,3,0,3,
     count       0,0,1,0,1,0,2,
     n =         2 3 4 5 6 7 8

     n=8  two diamonds cross connected
     and  ...
  */
  printf("bandwidth_reg3 min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_bandwidth_reg3[num_vertices]);
  }
  printf("\n");
  printf("             count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_bandwidth_reg3_count[num_vertices]);
  }
  printf("\n");

  /* bandwidth_reg3 max
     max   0,0,3,0,4,0,4,
     count 0,0,1,0,1,0,3,
     n =   2 3 4 5 6 7 8

     n=8  cube
     and  ...
  */
  printf("bandwidth_reg3 max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_bandwidth_reg3[num_vertices]);
  }
  printf("\n");
  printf("             count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_bandwidth_reg3_count[num_vertices]);
  }
  printf("\n");

  /* bandwidth_reg3 total
     n =          2 3  4  5   6  7
     not in OEIS: 0,0,3,0,7,0,18,0
     not A128114 middle match
  */
  printf("bandwidth_reg3 total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_bandwidth_reg3[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* Szeged min = 1,3,6,10,15,21,28, for complete-n, sum 1*1 each
     count        1,1,1,1,1,1,1,
  */
  printf("Szeged min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_Szeged[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_Szeged_count[num_vertices]);
  }
  printf("\n");

  /* Szeged max = 1,4,16,36,81,144,256,
     n =          2 3  4  5
     count 1,1,1,1,1,1,1,
     bipartite n,n or n,n+1
  */
  printf("Szeged max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_Szeged[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_Szeged_count[num_vertices]);
  }
  printf("\n");

  /* Szeged total
     not in OEIS: 1,7,58,403,4061,51412,1065283,
     n =          2 3  4

  */
  printf("Szeged total    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_Szeged[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* number of linegraph
     count   1,2,5,12,30,79,227,
     A003089 1,2,5,12,30,79,227,710,2322
     count   1,2,5,14,48,170,711,
     n =     2 3 4  5  6
  */
  printf("count linegraphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_linegraphs[num_vertices]);
  }
  printf("\n");

  /* number of non linegraphs
     not in OEIS: 0,0,1,9,82,774,10890
     n =          2 3 4 5  6
  */
  printf("count nonlinegraphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_nonlinegraphs[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* total cut edges
     not in OEIS: 1,2,7,22,104,619,5688,85064,
  */
  printf("total cut edges    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_cut_edges[num_vertices]);
  }
  printf("\n");

  /* count any cut edge    1,1,3,10,52,351,3714,63638,
     A052446 simple connected bridged graphs
  */
  printf("count any cut edge    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_any_cut_edge[num_vertices]);
  }
  printf("\n");

  /* count no cut edge
     0,1,3,11,60,502,7403,197442,
     A007146 simple connected bridgeless graphs
  */
  printf("count no cut edge     ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_no_cut_edge[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* transitive count    1,1,2,2,5,3,10,7,
     A006800
  */
  printf("transitive count    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_transitive[num_vertices]);
  }
  printf("\n");

  /* nontransitive count
     not in OEIS: 0,1,4,19,107,850,11107,261073,
  */
  printf("nontransitive count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_nontransitive[num_vertices]);
  }
  printf("\n");

  /* transitive_symmetric   count 1,1,2,2,4,2,5,4,
     A241814 maybe
  */
  printf("transitive_symmetric   count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_transitive_symmetric[num_vertices]);
  }
  printf("\n");

  /* transitive_unsymmetric count 0,0,0,0,1,1,5,3,
     A121021 maybe
  */
  printf("transitive_unsymmetric count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_transitive_unsymmetric[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */

  /* total cycles
     not in OEIS: 0,1,12,129,1836,37534,1442495,116445013,
     .            2 3  4
  */
  printf("total cycles    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_cycles[num_vertices]);
  }
  printf("\n");

  /* cycles max 0,1,7,37,197,1172,8018,62814
     A002807
     count      1,1,1,1,1,1,1,1
  */
  printf("cycles max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%ld,",graphs_max_cycles[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_cycles_count[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* total indcycles
     not in OEIS: 0,1,8,57,526,6481,133512,4884503,
     .            2 3 4  5
  */
  printf("total indcycles    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_indcycles[num_vertices]);
  }
  printf("\n");

  /* indcycles max 0,1,4,10,20,35,56,84,
     count         1,1,1,1,1,1,1,1
     A000292 tetrahedrals ?
     A101552 sum some binomials
  */
  printf("indcycles max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%ld,",graphs_max_indcycles[num_vertices]);
  }
  printf("\n");
  printf("        count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_indcycles_count[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* number of graphs with 1 or more claws
     not in OEIS: 0,0,1,7,62,662,10236,256586,11690182
     .            2 3 4 5
  */
  printf("count with claw    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_withclaw[num_vertices]);
  }
  printf("\n");

  /* number of claw-free graphs
     . 1,2,5,14,50,191,881,4494,26389
     . 2 3 4  5
     A022562 connected claw-free unlabeled graphs on n nodes
     1,1,2,5,14,50,191,881,4494,26389,184749,1728404,23805256,491544474,14491876320,
  */
  printf("count claw-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_clawfree[num_vertices]);
  }
  printf("\n");

  /* total claws in graphs of n vertices
     not in OEIS: 0,0,1,13,164,2434,56134,2210796,159330519
     .            2 3 4  5
  */
  printf("total claws    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_claws[num_vertices]);
  }
  printf("\n");

  /* number of degree-3-free graphs
     not in OEIS: 1,2,2,7,23,136,1563,41190
     .            2 3 4 5
  */
  printf("count deg3-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_deg3free[num_vertices]);
  }
  printf("\n");

  /* number of cubic (all degree-3) graphs
     not in OEIS: 0,0,1,0,2,0,5,0,19,0,85   \\ not with 0s
     .            2 3 4 5 6 7 8 9 10
     A002851          1,  2,  5,  19,  85  cubics on 2n

     cf A002829 cubics labelled
     A164919 cubic non-hamiltonians
     A165653 disconnected cubics
     A000109 cubic planars, maybe
     A204186 cubic non-planars
     A204326 cubics with chromatic number 3
     A204328 cubics with non-trivial automorphism group
  */
  printf("count cubics       ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_cubics[num_vertices]);
  }
  printf("\n");

  /* number of degree-2-free graphs
     A005636 1,0,2,5,32,234,3638,106147
     .       2 3 4 5  6   7
  */
  printf("count deg2-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_deg2free[num_vertices]);
  }
  printf("\n");

  /* max claws in graphs of n vertices
     .        0,0,1,4,10,20,40,70,112
     count    1,1,1,1, 1, 3, 2, 2,  3
     .        2 3 4 5  6  7  8  9  10
     cf
     A000292  0,0,1,4,10,20,35,56, 84,120,165,220,286,364,455,
     = trees tetrahedral numbers smaller (1 tree star-n each)
     at n=8 have more claws in cyclic, so the star-n tree is not included
     not A049032    0, 0, 1, 4, 10, 20, 40, 70, 112, 168, 252, 360, 495, 660,
     not A100354 0, 0, 0, 1, 4, 10, 20, 40, 70, 112, 168, 252, 360, 495, 661,
     their n=11 168 smaller than bipartite 176

     cyclics: 0,0,0,2, 8,20,40,70,112
     .        2 3 4 5  6  7  9  9  10
     A007290  2*binomial(n,3).
     count    0,1,1,3, 2, 2, 2, 2,  3

     A049032 = A049031/2  A049031 = maximization integer cubes

     n=7 star-7 tree
     .   complete bipartite 5,2
     .   complete bipartite 5,2 edge between 2

     n=8 complete bipartite 6,2
     .   complete bipartite 6,2 edge between 2

     n=9 complete bipartite 7,2
     .   complete bipartite 7,2 edge between 2

     n=10 complete bipartite 8,2
     .    complete bipartite 8,2 edge between 2
     .    complete bipartite 7,3

     GP-DEFINE  bipartite_claws(n,m) = m*binomial(n,3) + n*binomial(m,3);
     GP-DEFINE  bipartite_max_claws(n) = vecmax(vector(n,m, bipartite_claws(n-m,m)));
     one side only is A049032
     vector(20,n,n+=7; bipartite_max_claws(n))
     not in OEIS: 40,70,112,176,261,372,520,704,935,1220,1560,1976
     .             8  9  10  11

     one side only is at floor((n+1)/4) per A049031
     vector(20,n, my(m=(n+1)\4); bipartite_claws(n-m,m))

     for(n=8,20, \
     my(target=bipartite_max_claws(n)); \
     for(m=1,n, \
     if(bipartite_claws(n-m,m)==target, \
     print("n="n"  "n-m" "m"  "target))));

     for(n=2,10,for(m=2,n-2,print1(bipartite_claws(n-m,m)",")))
  */
  printf("max claws              ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_claws_max[num_vertices]);
  }
  printf("\n");
  printf("max claws count        ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_claws_max_count[num_vertices]);
  }
  printf("\n");
  printf("cyclic max claws       ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_claws_max_cyclic[num_vertices]);
  }
  printf("\n");
  printf("cyclic max claws count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_claws_max_cyclic_count[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* number of graphs with 1 or more triangles
     A128240 0,1,3,15,93,794,10850,259700,
     .       2 3 4  5
     girth 3
  */
  printf("count with triangle    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_withtriangle[num_vertices]);
  }
  printf("\n");

  /* number of graphs with exactly 1 triangle
     not in OEIS: 0,1,1,4,14,70,395,2967
     A000226 one cycle length 3 ?
  */
  printf("count one triangle     ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_onetriangle[num_vertices]);
  }
  printf("\n");

  /* number of triangle-free graphs
     A024607 1,1,3,6,19,59,267,1380,
     .       2 3 4  5
  */
  printf("count triangle-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_trianglefree[num_vertices]);
  }
  printf("\n");

  /* total triangles in graphs of n vertices
     not in OEIS: 0,1,7,48,401,4589,86317,2878659,
     .            2 3 4  5
  */
  printf("total triangles    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_triangles[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* number of graphs with 1 or more diamonds
     A243250 0,0,2,10,73,688,10150,
     n =     2 3 4 5
  */
  printf("count with diamond    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_withdiamond[num_vertices]);
  }
  printf("\n");

  /* number of graphs with exactly 1 diamond
     not in OEIS: 0,0,1,3,18,110,908,
     n =          2 3 4 5
  */
  printf("count one diamond     ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_onediamond[num_vertices]);
  }
  printf("\n");

  /* number of diamond-free graphs
     A242790 1,2,4,11,39,165,967,
     n =     2 3 4  5
     A079573 no induced diamond
  */
  printf("count diamond-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_diamondfree[num_vertices]);
  }
  printf("\n");

  /* total diamonds in graphs of n vertices
     not in OEIS: 0,0,7,71,701,9016,187788,
     n =          2 3 4  5
  */
  printf("total diamonds    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_diamonds[num_vertices]);
  }
  printf("\n");

  /* ----- */

  /* number of graphs with 1 or more induced_diamonds
     not in OEIS: 0,0,1,8,66,659,9979,
     n =          2 3 4 5
     girth 3
  */
  printf("count with induced_diamond    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_with_induced_diamond[num_vertices]);
  }
  printf("\n");

  /* number of graphs with exactly 1 induced_diamond
     not in OEIS: 0,0,1,3,18,113,947,
     n =          2 3 4 5
  */
  printf("count one induced_diamond     ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_one_induced_diamond[num_vertices]);
  }
  printf("\n");

  /* number of induced_diamond-free graphs
     count 1,2,5,13,46,194,1138,
     n =   2 3 4  5
     A079573 number of no induced diamond
  */
  printf("count induced_diamond-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_induced_diamond_free[num_vertices]);
  }
  printf("\n");

  /* total induced_diamonds in graphs of n vertices
     not in OEIS: 0,0,1,17,227,3460,80988,
     n =          2 3 4  5
  */
  printf("total induced_diamonds    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_induced_diamonds[num_vertices]);
  }
  printf("\n");

  /* ----- */

  /* number of graphs with 1 or more complete-4s
     A243244 connected graphs with >=1 complete-4
     count 0,0,1,4,30,317,5511,
     n =   2 3 4 5
     girth 3
  */
  printf("count with complete4    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_with_complete4[num_vertices]);
  }
  printf("\n");

  /* number of complete4-free graphs
     count 1,2,5,17,82,536,5606,
     n =   2 3 4  5
     A079574 number with no complete4
  */
  printf("count complete4-free    ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_complete4_free[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* sum girths of connected graphs of n vertices
     not in OEIS: 3,13,58,334,2584,33561,784556
  */
  printf("sum girths ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",sum_girths[num_vertices]);
  }
  printf("\n");

  /* --------------------------------------------------- */

  /* total number of maximal cliques in connected graphs of n vertices
     not in OEIS: 1,3,15,70,496,4751,77394,2245335,
  */
  printf("sum maxcliques ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",sum_maxcliques[num_vertices]);
  }
  printf("\n");


  /* --------------------------------------------------- */

  /* total clique number
   */
  printf("total clique number ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",total_clique_number[num_vertices]);
  }
  printf("\n");


  /* --------------------------------------------------- */

  /* Ramsey graphs
     R(3,3)=6  5-cycle no 3-clique no 3-indset
     R(4,2)=4  3-cycle no 4-clique no 2-indset

     R(3,4)=9

     .   0--3-6--1
     .   |   X   |
     .   5--2-7--4
  */
  printf("total clique number ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",total_clique_number[num_vertices]);
  }
  printf("\n");


  /* --------------------------------------------------- */

  /* sum edges of connected graphs of n vertices
     A094602 1,5,25,130,951,9552,160220,4756703
  */
  printf("total edges ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_edges[num_vertices]);
  }
  printf("\n");

  int graph_min_vertices_cyclic = MAX(graph_min_vertices,3);

  /* --------------------------------------------------- */

  /* graphs with diameter d
     graphs diameter=2 A241706 0,1,4,14,59,373,4154,91518
     graphs diameter=3 A241707 0,0,1, 5,43,387,5797,148229
  */
  {
    int d;
    for (d = 0; d <= DIAMETER_COUNT_MAX; d++) {
      printf("graphs diameter=%d count ", d);
      for (num_vertices = graph_min_vertices;
           num_vertices <= graph_max_vertices;
           num_vertices++) {
        printf("%d,",graphs_diameter_count[num_vertices][d]);
      }
      printf("\n");
    }

    for (d = 0; d <= DIAMETER_COUNT_MAX; d++) {
      printf("graphs diameter<=%d count ", d);
      for (num_vertices = graph_min_vertices;
           num_vertices <= graph_max_vertices;
           num_vertices++) {
        int i;
        int cumulative = 0;
        for (i = 0; i <= d; i++)
          cumulative += graphs_diameter_count[num_vertices][i];
        printf("%d,",cumulative);
      }
      printf("\n");
    }
  }

  /* cyclics with diameter
     diameter=1
     count        1,1, 1, 1,  1,   1,     1
     diameter=2
     not in OEIS: 0,3,13,58,372,4153, 91517
     diameter=3
     not in OEIS: 0,0, 4,41,385,5794,148226

     n =                      2 3 4 5 6 7 8 9
     graphs diameter<=0 count 0,0,0,0,0,0,0,0,
     graphs diameter<=1 count 1,1,1,1,1,1,1,1,
     graphs diameter<=2 count
     not in OEIS: 1,2,5,15,60,374,4155,91519,
     graphs diameter<=3 count
     not in OEIS: 1,2,6,20,103,761,9952,239748,
  */
  {
    int d;
    for (d = 0; d <= DIAMETER_COUNT_MAX; d++) {
      printf("cyclics diameter=%d count ", d);
      for (num_vertices = graph_min_vertices_cyclic;
           num_vertices <= graph_max_vertices;
           num_vertices++) {
        printf("%d,",cyclics_diameter_count[num_vertices][d]);
      }
      printf("\n");
    }
  }

  /* graphs diameter total
     not in OEIS: 1,3,12,48,285,2287,30512,715273,
  */
  printf("graphs diameter total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_diameter[num_vertices]);
  }
  printf("\n");

  /* cyclics diameter total
     not in OEIS: 1,7,39,264,2243,30409,715041
  */
  printf("cyclics diameter total ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",cyclics_total_diameter[num_vertices]);
  }
  printf("\n");
  printf("\n");

  /* --------------------------------------------------- */

  /* indnum min = 1,1,1,1,1,1,1 for complete
     count        1,1,1,1,1,1,1
  */
  printf("indnum min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_indnum_count[num_vertices]);
  }
  printf("\n");

  /* indnum max = 1,2,3,4,5,6,7  n-1 for star, unique
     count        1,1,1,1,1,1,1
  */
  printf("indnum max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_indnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_indnum_count[num_vertices]);
  }
  printf("\n");

  /* indnum total
     not in OEIS: 1,3,12,51,315,2700,38360
  */
  printf("indnum total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_indnum[num_vertices]);
  }
  printf("\n");

  /* indnum = n/2 number of graphs
     not in OEIS: 1,0,4,0,63,0,4308
     not in OEIS: 1,  4,  63,  4308
  */
  printf("indnum half ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_half_indnum[num_vertices]);
  }
  printf("\n");

  /* indcomp is all bipartites
     A005142 1,1,3,5,17,44,182
     .       2   4    6      8
  */
  printf("indcomp ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indcomp[num_vertices]);
  }
  printf("\n");


  /* indnum ways max
     not in OEIS: 2,3,4,5,8,11,15,22
     n =          2 3 4 5
     count graphs
     not in OEIS: 1,1,1,3,1, 1, 1, 1
     n =          2 3 4 5 6  7  8, 9

     n=5, indnum attained in 5 ways
     triangle with path-2  indnum=2   https://hog.grinvin.org/ViewGraphInfo.action?id=330
     5-cycle               indnum=2
     complete-5            indnum=1

     n=6 in 8 ways  complete-3 -- complete-3
     >>graph6<<EQjO

     n=7 in 11 ways complete-4 -- complete-3
     >>graph6<<FQhVO

     n=8 in 15 ways complete-4 -- complete-4
     >>graph6<<GQhTUg

     n=9 in 22 ways triangle--triangle--triangle, at common vertex of middle
     >>graph6<<HCOcaRc
     2*3*3 without middle

     2*2 with middle
     2*3*3 + 2*2 == 22

     disconnected graphs
     v=[1,2,3,4,5,8,11,15,22];
     for(n=2,9, my(m=0,pm); forpart(p=n, my(t=prod(i=1,#p,v[p[i]])); if(t>m,m=t;pm=p)); print("n="n" max "m" partition "Vec(pm)" factor ",factor(m),"  cf "v[n]));
  */
  printf("indnum ways max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indnum_ways_max[num_vertices]);
  }
  printf("\n");
  printf("          count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indnum_ways_max_count[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* graphs domnum min 1,1,1,1,1,1,1,
     any graph n-1 with a further vertex connected to all others
     count 1,2,4,11,34,156,1044
     n     2 3 4  5
     A000088 all graphs starting n=0 1,1,2,4,11 (incl disconnected)
  */
  printf("graphs domnum min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_domnum_count[num_vertices]);
  }
  printf("\n");

  /* graphs domnum max 1,1,2, 2,3, 3,4,  4
     count graphs
     not in OEIS:      1,2,2,10,2,42,6,191
     n =               2 3 4  5 6  7 8   9
  */
  printf("graphs domnum max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_domnum_count[num_vertices]);
  }
  printf("\n");

  /* graphs domnum total
     not in OEIS: 1,2,8,31,192,1592,22107,538916
  */
  printf("graphs domnum total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_domnum[num_vertices]);
  }
  printf("\n");

  /* graphs domnum count by d
     1,2,0,4,2,0,11,10,0,0,34,76,2,0,0,156,655,42,0,0,0,1044,9162,905,6,0,0,0,

     1, 1, 1, 2, 1, 1, 4, 5, 1, 1, 11, 16, 5, 1, 1, 34, 94, 21, 5, 1, 1,
  */
  printf("graphs domnum by d ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    int d;
    for (d=1; d<num_vertices; d++)
      printf("%d,",graphs_domnum_count[num_vertices][d]);
  }
  printf("\n");

  /* graphs domnum perfect maxdeg+1
     not in OEIS: 1,2,4,11,36,156,1096,12348
     n =          2 3 4  5  6   7    8     9
     cyclic       0,1,3,10,34,155,1094,12346
     vertex of degree n-1 edges to everything else is domnum=1

     maxdeg-1 < n
     not in OEIS: 0,0,0,0,2,0,52,2
     n =          2 3 4 5 6 7  8 9
     n=6 is path-6 and 6-cycle
  */
  printf("graphs domnum perfect ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_domnum_perfect[num_vertices]);
  }
  printf("\n");
  printf("graphs domnum perfect cyclic ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_domnum_perfect_cyclic[num_vertices]);
  }
  printf("\n");
  printf("graphs domnum perfect smaller ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_domnum_perfect_smaller[num_vertices]);
  }
  printf("\n");

  /* domnum ways max
     not in OEIS: 2,3,6,9,15,22,36
     n =          2 3 4 5  6  7  8
     count graphs 1,1,1,1, 1, 1, 1

     n=4  4-cycle
     n=5
     .       0---2      3
     .      / \ /|    /   \
     .     3   / |   0--2--1
     .      \ / \|    \ | /
     .       1---4      4
     >>graph6<<D]w

     n=6  complete-6 less 3 edges for 4-regular
     GP-Test  binomial(6,2) == 15
     >>graph6<<E]~o

     n=7  22 ways
     >>graph6<<FCZbg

     n=8  36 ways
     >>graph6<<GCxvBo

     v=[1,2,3,6,9,15,22,36];
     for(n=2,#v, my(m=0,pm); forpart(p=n, my(t=prod(i=1,#p,v[p[i]])); if(t>m,m=t;pm=p)); print("n="n" max "m" partition "Vec(pm)" factor ",factor(m),"  cf "v[n]));
  */
  printf("domnum ways max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_domnum_ways_max[num_vertices]);
  }
  printf("\n");
  printf("          count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_domnum_ways_max_count[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* cyclics domnum min 1,1, 1,1,1,1,1 = 1 for complete
     count
     not in OEIS:       1,3,10,33,155,1043,12345
     .                  3 4  5  6   7    8     9
     dominated by 1 is diameter <= 3, but not all such
  */
  printf("cyclics domnum min ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",cyclics_min_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",cyclics_min_domnum_count[num_vertices]);
  }
  printf("\n");

  /* cyclics domnum max 1,2,2,3, 3,4,  4
     count
     not in OEIS:       1,1,8,1,37,4,180
     .                  3 4 5 6  7 8   9

     n=8 tetrahedral with extra vertex each
     .   square with extra vertex each
     .   triangle with 3x1 and 1x2 hanging
     .   square with cross edge and extra vertex each
  */
  printf("cyclics domnum max ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",cyclics_max_domnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",cyclics_max_domnum_count[num_vertices]);
  }
  printf("\n");

  /* cyclics domnum total
     not in OEIS: 1,5,26,180,1566,22045,538774
  */
  printf("cyclics domnum total ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",cyclics_total_domnum[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */

  /* graphs totdomnum min 2,2,2,2,2,2,2,2,
     count
     not in OEIS: 1,2,6,19,96,670,8191,183149,
     n            2 3 4  5

     ways
     not in OEIS: 1,5,22,88,467,3206,36262,737119,

     n=2 triangle  any 2 of 3 = 3 ways
     .   path  *--*--.  or .--*--*

     all vertices neighbours of two adjacent
     must diameter<=3, but some diameter<=3 have bigger totdomnum,
     eg. n=5 cycle is diameter=2 but totdomnum=3
  */
  printf("graphs totdomnum min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_totdomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_totdomnum_count[num_vertices]);
  }
  printf("\n");
  printf("     ways  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_min_totdomnum_count_ways[num_vertices]);
  }
  printf("\n");

  /* graphs diameter <= 3 but totdomnum > 2
     not in OEIS: 0,0,0,1,7,91,1761,56599,
     n            2 3 4 5 6  7    8     9
  */
  printf("graphs diameter<=3 but totdomnum > 2 ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_diameter_le_3_totdomnum_gt_2[num_vertices]);
  }
  printf("\n");


  /* graphs totdomnum max
     not in OEIS: 2,2, 2,3, 4,  4,5,6,
     n            2 3  4 5  6   7 8 9
     count graphs
     not in OEIS: 1,2, 6,2, 2, 20,3,2,
     n            2 3  4 5  6   7 8 9
     ways
     not in OEIS: 1,5,22,6,13,140,8,
     n            2 3  4 5  6   7 8

     n=8 3 of
     triangle with arms 1,2,2      3 sets
     star arms 2,2,3               2 sets
     star arms 1,3,3               3 sets
  */
  printf("graphs totdomnum max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_totdomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_totdomnum_count[num_vertices]);
  }
  printf("\n");
  printf("     ways  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_max_totdomnum_count_ways[num_vertices]);
  }
  printf("\n");

  /* graphs totdomnum total
     not in OEIS: 2,4,12,44,242,1909,25506,607470,
     n            2 3  4  5   6    7     8      9
  */
  printf("graphs totdomnum total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_totdomnum[num_vertices]);
  }
  printf("\n");

  printf("\n");


  /* --------------------------------------------------- */

  /* graphs indomnum min 1,1,1,1,1,1,1,1,
     any graph n-1 with a further vertex connected to all others
     count graphs        1,2,4,11,34,156,1044,12346,274668
     A000088
     n =                 2 3 4  5  6   7    8     9
     number of ways      2,4,8,19,53,209,1253,13599,288267
  */
  printf("graphs indomnum min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_indomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_indomnum_count[num_vertices]);
  }
  printf("\n");
  printf("     ways  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_min_indomnum_count_ways[num_vertices]);
  }
  printf("\n");


  /* graphs indomnum max 1,1,2, 2, 3,  3,  4, 5,    5,
     n =                 2 3 4  5  6   7   8  9    10
     count graphs
     not in OEIS:        1,2,2,10, 6, 82, 28, 1,  213
     n                   2 3 4  5  6   7   8  9    10
     total indomnum sets
     not in OEIS:        2,4,5,30,20,354,128, 3, 1245
  */
  printf("graphs indomnum max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_indomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_indomnum_count[num_vertices]);
  }
  printf("\n");
  printf("     ways  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_max_indomnum_count_ways[num_vertices]);
  }
  printf("\n");

  /* graphs indomnum total
     not in OEIS: 1,2,8,31,196,1632,22865,560841,
     n            2 3  4  5   6    7     8      9
  */
  printf("graphs indomnum total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_indomnum[num_vertices]);
  }
  printf("\n");

  /* indomnum ways max -- same as indnum ways, graph shown in Jou and Chang
     not in OEIS: 2,3,4,5,8,11,15,22
     count        1,1,1,3,1, 1, 1, 1
  */
  printf("indomnum ways max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indomnum_ways_max[num_vertices]);
  }
  printf("\n");
  printf("          count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indomnum_ways_max_count[num_vertices]);
  }
  printf("\n");

  /* graphs indomnum == domnum
     not in OEIS: 1,2,6,21,108,813,10370
     n            2 3  4  5   6    7     8      9
  */
  printf("graphs indomnum = domnum ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indomnum_eq_domnum[num_vertices]);
  }
  printf("\n");

  /* graphs indomnum == domnum, claw free
     A022562 number of connected claw-free
     = 1,2,5,14,50,191,881
     n 2 3 4  5  6   7   8      9
  */
  printf("graphs indomnum = domnum clawfree ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indomnum_eq_domnum_clawfree[num_vertices]);
  }
  printf("\n");

  /* graphs indomnum == domnum, with claw
     not in OEIS: 0,0,1,7,58,622,9489
     n            2 3 4 5  6   7    8      9
     n=4 is claw itself
     n=5 tree claw plus 1 longer arm
     tree star-5
     bipartite 5,2, and with "2" joined
     4-cycle and 1 hanging
     triangle and 2 hanging from one vertex

     Vaidya and Pandit
     duplication each vertex is indomnum=domnum
     duplication of edge is new vertex joined to each end is indomnum=domnum
     duplication of edges of path-n is indomnum=domnum
     for path-3 gives two triangles -- claw-free?
     duplication of edges of bipartite r,s is indomnum=domnum (unchanged)
  */
  printf("graphs indomnum = domnum withclaw ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_indomnum_eq_domnum_withclaw[num_vertices]);
  }
  printf("\n");

  printf("\n");


  /* --------------------------------------------------- */

  /* graphs invdomnum min 1,1,1,1,1,1,1,
     any graph n-1 with a further vertex connected to all others
     count graphs        1,2,4,11,34,156,1044
     A000088
     n =                 2 3 4  5  6   7    8     9
     number of ways      2,4,8,19,53,209,1253,
     same as domnum
  */
  printf("graphs invdomnum min ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_invdomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_min_invdomnum_count[num_vertices]);
  }
  printf("\n");
  printf("     ways  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_min_invdomnum_count_ways[num_vertices]);
  }
  printf("\n");


  /* graphs invdomnum max 1,1, 2, 2, 3,  3,  4,
     n =                  2 3  4  5  6   7   8  9    10
     count graphs
     not in OEIS:         1,2, 2,10, 2, 42,  6,
     n =                  2 3  4  5  6   7   8  9    10
     total invdomnum sets
     not in OEIS:         2,4,10,52,16,462, 96,

     count smaller than domnum
     n=6 invdomnum=3
     E graph
     triangle with hanging leaves
  */
  printf("graphs invdomnum max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_invdomnum[num_vertices]);
  }
  printf("\n");
  printf("     count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_max_invdomnum_count[num_vertices]);
  }
  printf("\n");
  printf("     ways  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_max_invdomnum_count_ways[num_vertices]);
  }
  printf("\n");

  /* graphs invdomnum total
     not in OEIS: 1,2,8,31,192,1592,22107,
     n            2 3 4  5   6    7     8      9
  */
  printf("graphs invdomnum total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%Lu,",graphs_total_invdomnum[num_vertices]);
  }
  printf("\n");

  /* invdomnum ways max
     not in OEIS: 2,3,6,9,15,22,36,
     count        1,1,1,1, 1, 1, 1,
  */
  printf("invdomnum ways max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_invdomnum_ways_max[num_vertices]);
  }
  printf("\n");
  printf("          count ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_invdomnum_ways_max_count[num_vertices]);
  }
  printf("\n");

  /* graphs invdomnum == domnum
     .   1,2, 6,21,112, 853,11117,
     n = 2 3  4  5   6    7     8      9
     number of all graphs ... ?
  */
  printf("graphs invdomnum = domnum ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_invdomnum_eq_domnum[num_vertices]);
  }
  printf("\n");

  /* graphs invdomnum == domnum, claw free .. it seems
     A022562 number of connected claw-free
     = 1,2,5,14,50,191,881,
     n 2 3 4  5  6   7   8      9
  */
  printf("graphs invdomnum = domnum clawfree ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_invdomnum_eq_domnum_clawfree[num_vertices]);
  }
  printf("\n");

  /* graphs invdomnum == domnum, with claw
     not in OEIS: 0,0,1,7,62,662,10236
     n            2 3 4 5  6   7     8      9
  */
  printf("graphs invdomnum = domnum withclaw ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_invdomnum_eq_domnum_withclaw[num_vertices]);
  }
  printf("\n");

  printf("\n");


  /* --------------------------------------------------- */

  /* number of well covered graphs
     count 1,1,3,6,27,108,788,9035
     n   = 1 2 3 4  5   6   7
     A222625
  */
  printf("graphs well covered  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_wellcovered[num_vertices]);
  }
  printf("\n");


  /* --------------------------------------------------- */

  /* number of stable connected graphs
     count 1,2,5,16,79,544,6007,
     n   = 2 3 4  5  6   7    8
     not in OEIS: 1,2,5,16,79,544,6007,

     connected graphs not stable
     count 0,0,1,5,33,309,5110,
     n   = 2 3 4 5  6   7    8
     not in OEIS: 1,5,33,309,5110,

     Holton, "A Report on Stable Graphs",
     n=5,  5 not stable
     n=6,  graph codes list of connected not stables = 33 of
     1, 4, 9, 12, 14, 18, 19, 24, 30, 31, 32, 36, 37, 44, 45, 46, 53, 54,
     55, 59, 69, 70, 71, 74, 77, 79, 85, 88, 91, 93, 100, 102, 106
  */
  printf("graphs stable      ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_stable[num_vertices]);
  }
  printf("\n");

  printf("graphs not stable  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count[num_vertices] - graphs_count_stable[num_vertices]);
  }
  printf("\n");

  /* number of stable graphs connected or disconnected
     count 2,4,10,28,116,688,6830,
     n   = 2 3  4  5   6   7    8
     not in OEIS: 2,4,10,28,116,688,6830,
  */
  printf("graphs stable      ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_stable[num_vertices]);
  }
  printf("\n");

  /* --------------------------------------------------- */

  /* graphs total number of semistable vertices
     total 2,4,15,50,343,3152,55087
     n   = 2 3  4  5   6    7     8
  */
  printf("graphs total semistable vertices  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_total_semistable_vertices[num_vertices]);
  }
  printf("\n");

  /* number of completely semi-stable graphs
     count 1,1,3,5,23,63,766
     n   = 2 3 4 5  6  7   8
  */
  printf("graphs completely semistable  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_completely_semistable[num_vertices]);
  }
  printf("\n");

  /* number of completely semi-stable graphs, irregular
     not in OEIS: 1,3,18,59,749,
     n   =       2 3 4 5  6  7   8
     count       0,0,1,3,18,59,749,
     cf
     not regular 0,1,4,19,107,849,11100,
     regular     1,1,2,2,5,4,17,

     n=4 claw
     n=5 star, triangles common edge, triangles common vertex
     n=6 star, bistar 3,3, complete bipartite 4,2,
     triangle 3 hanging leaves (= net graph)
     n=6 triangle and square common vertex (fish graph)
  */
  printf("graphs completely semistable irregular  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_completely_semistable_irregular[num_vertices]);
  }
  printf("\n");

  /* number of completely semi-stable graphs, intransitive
     not in OEIS: 1,3,18,60,756,
     n   =       2 3 4 5  6  7   8
     count       0,0,1,3,18,60,756,
  */
  printf("graphs completely semistable intransitive  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_count_completely_semistable_intransitive[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */

  /* selfcomplements
     A000171  0,1,2,0,0,10,36
  */
  printf("selfcomplements ");
  for (num_vertices = graph_min_vertices_cyclic;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_selfcomplements[num_vertices]);
  }
  printf("\n");
  printf("\n");


  /* --------------------------------------------------- */

  /* regular
     A005177 1,1,2,2,5,4,17,22,
  */
  printf("regular     ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_regular[num_vertices]);
  }
  printf("\n");

  /* not regular
     A241839 0,1,4,19,107,849,11100,
  */
  printf("not regular ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_notregular[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */

  /* median size max 2,3,4,5,6,7,8,9,
     n =             2 3 4 5 6 7 8
     count           1,1,2,2,5,4,15,23,
     cf regulars     1,1,2,2,5,4,17,22,

     max n for complete-n all same distances, and other vertex transitive
  */
  printf("median size max ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_median_size_max[num_vertices]);
  }
  printf("\n");

  printf("count           ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_median_size_max_count[num_vertices]);
  }
  printf("\n");

  printf("\n");


  /* --------------------------------------------------- */

  /* distance regular
     A241814 1,1,2,2,4,2,5,
     n =     2 3 4 5 6
     n=3 triangle
     n=4 complete-4 {3;1} and 4-cycle {2,1; 1;2}
     n=5 complete-5 {4;1} and 5-cycle {2,1; 1;1}
  */
  printf("distance regular     ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_distance_regular[num_vertices]);
  }
  printf("\n");

  /* not distance regular
     A241840 0,1,4,19,108,851,11112,
  */
  printf("not distance_regular ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",count_not_distance_regular[num_vertices]);
  }
  printf("\n");

  printf("\n");

  /* --------------------------------------------------- */
  /* similar, pseudosimilar etc */

  /* biggest set of mutually pseudosimilar u,v
     max pseudosimilar clique 1,1,1, 1,  1,  1, 2,  2,
     count graphs             1,2,6,21,112,853,44,410,
     n =                      2 3 4  5   6   7  8   9
     not in OEIS: 44,410
     not A183413 nxk binary arrays seems unlikely

     min edges                 0,0,0,0,0,0,9,10,
     count graphs              0,0,0,0,0,0,1,1,
     Harary and Palmer n=8 unique minimum 9 edges

     possibly disconnected graphs:
     max pseudosimilar clique  1,1,1,1,1,1,2,
     count graphs              2,4,11,34,156,1044,44,
     min edges                 0,0,0,0,0,0,9,
     count graphs              0,0,0,0,0,0,1,
  */
  printf("max pseudosimilar clique  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_pseudosimilar_clique_max[num_vertices]);
  }
  printf("\n");
  printf("count graphs              ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_pseudosimilar_clique_max_count[num_vertices]);
  }
  printf("\n");
  printf("min edges                 ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_pseudosimilar_min_edges[num_vertices]);
  }
  printf("\n");
  printf("count graphs              ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_pseudosimilar_min_edges_count[num_vertices]);
  }
  printf("\n");

  /* (biggest set of mutually removal similar u,v)
     is all vertices of each vertex-transitive
     A006800 count vertex-transitive graphs
     max removal_similar clique 2,3,4,5,6,7, 8, 9,
     count                      1,1,2,2,5,3,10, 7,
     n =                        2 3 4 5 6 7  8  9

     possibly disconnected:
     max removal_similar clique  2,3,4,5,6,7,8,
     count graphs                2,2,4,3,8,4,14,
     A006799 2,2,4,3,8,4,14 number of vertex transitive
  */
  printf("max removal_similar clique  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_removal_similar_clique_max[num_vertices]);
  }
  printf("\n");
  printf("count graphs                ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_removal_similar_clique_max_count[num_vertices]);
  }
  printf("\n");


  /* similar pairs total 1,4,20,72,394
     n =                 2 3  4  5   6
     same as removal similar total, as no pseudosimilars until n=8
  */
  printf("similar pairs total         ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_similar_total[num_vertices]);
  }
  printf("\n");

  /* removal similar pairs total 1,4,20,72,394,
     n =                         2 3  4  5   6
     not in OEIS: 1,4,20,72,394
     n=3 triangle 3 pairs, path-3 1 pair = 4 pairs
  */
  printf("removal similar pairs total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_removal_similar_total[num_vertices]);
  }
  printf("\n");


  /* similar unique pair 1,1,1,5,21,193,2825,
     not in OEIS: 1,1,1,5,21,193,2825,
     n =          2 3 4 5
     smaller than removal similar uniques when get to some pseudosimilars
     which add to the removal similar uniques
  */
  printf("similar unique pair ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_similar_is_unique[num_vertices]);
  }
  printf("\n");

  /* removal similar unique pair 1,1,1,5,21,193,2853,
     not in OEIS: 1,1,1,5,21,193,2853
     n =          2 3 4 5

     n=2 path-2 only
     n=3 path-3 only, ends removal similar
     n=4 triangle and hanging, ends of the triangle removal similar
     n=5 square and leaf
     .   triangle and 2-tail
     .   4-star and tail
     .   2-triangles and tail at 2 places

     possibly disconnected:
     removal similar unique pair 2,2,2,6,26,214,3062,
     not in OEIS: 2,2,2,6,26,214,3062
  */
  printf("removal similar unique pair ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_removal_similar_is_unique[num_vertices]);
  }
  printf("\n");


  /* max orbit  2,3,4,5,6,7,8,9,
     count      1,1,2,2,5,3,10,7,
     n =        2 3 4 5  6   7    8
     all vertices in orbit of a vertex transitive
     A006800 count connected vertex-transitive graphs
     1,1,2,2,5,3,10,7,

     possibly disconnected:
     max orbit  2,3,4,5,6,7,8,
     count      2,2,4,3,8,4,14,
     A006799 2,2,4,3,8,4,14 number of vertex transitive
  */
  printf("max orbit  ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,",graphs_orbit_max[num_vertices]);
  }
  printf("\n");
  printf("count      ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_orbit_max_count[num_vertices]);
  }
  printf("\n");

  /* numorbits total 1,3,11,58,407
     n =             2 3  4  5   6
     A126100 num connected rooted graphs
  */
  printf("numorbits total ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_numorbits_total[num_vertices]);
  }
  printf("\n");

  /* identity (single automorphism, asymmetric)
     A124059  0,0,0,0,8,144,3552
     n =      2 3 4 5 6   7    8
     n=3 triangle 3 pairs, path-3 1 pair = 4 pairs
  */
  printf("identity graphs ");
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    printf("%d,", graphs_asymmetric_count[num_vertices]);
  }
  printf("\n");

  printf("\n");
}

int ltrees_count[64];
int ltrees_total_eccentricities[64];
unsigned long long ltrees_total_indomsets[64];
unsigned long long ltrees_total_height[64];
unsigned long long ltrees_total_2height[64];
int ltrees_ahead_same[64];
int ltrees_ahead_different[64];
void
try_ltrees (void)
{
  const int min_n = 2;
  const int max_n = 7;

  int forest;
  for (forest = 0; forest <= 1; forest++) {
    const char *str = (forest ? "forest" : "tree");

    memset(ltrees_count,          '\0',sizeof(ltrees_count));
    memset(ltrees_total_eccentricities,'\0',sizeof(ltrees_total_eccentricities));
    memset(ltrees_total_indomsets,'\0',sizeof(ltrees_total_indomsets));
    memset(ltrees_total_height   ,'\0',sizeof(ltrees_total_height));
    memset(ltrees_total_2height  ,'\0',sizeof(ltrees_total_2height));
    memset(ltrees_ahead_same     ,'\0',sizeof(ltrees_ahead_same));
    memset(ltrees_ahead_different,'\0',sizeof(ltrees_ahead_different));

    /* loop over all labelled trees */
    int n;
    for (n = 2; n <= max_n; n++) {
      int vec[max_n];
      nautyextra_can_init(vec,n, forest);
      do {
        DEBUG({ int i;
                printf("vec ");
                for (i = 0; i < n; i++) printf(" %d", vec[i]);
                printf("\n");
              });

        int vpar[max_n+1];
        nautyextra_from_upascend(vec,n,0, vpar);
        nautyextra_validate_vpar(vpar,n);
        ltrees_count[n]++;

        {
          int vpar_ahead[max_n+1];
          nautyextra_from_upascend(vec,n,1, vpar_ahead);
          nautyextra_validate_vpar(vpar_ahead,n);
          int ahead_same = nautyextra_aresame_vpar(vpar, vpar_ahead, n);
          ltrees_ahead_same[n]      += ahead_same;
          ltrees_ahead_different[n] += !ahead_same;

          if (0) {
            printf("ahead=0  ");
            nautyextra_print_vpar(vpar,n);
            printf("ahead=1  ");
            nautyextra_print_vpar(vpar_ahead,n);
            printf("same %d\n", ahead_same);
          }
        }
        {
          int eccs[tree_max_vertices+1];
          nautyextra_eccentricities_vpar(vpar,n, eccs);
          int total_eccentricities = 0;
          int v;
          for (v = 1; v <= n; v++) total_eccentricities += eccs[v];
          ltrees_total_eccentricities[n] += total_eccentricities;
        }
        if (n>=1) ltrees_total_height[n] += nautyextra_height_vpar(vpar,n);
        if (n>=2) ltrees_total_2height[n] += nautyextra_2height_vpar(vpar,n);
        ltrees_total_indomsets[n] += nautyextra_indomsets_count_vpar(vpar,n);
      } while (nautyextra_can_next(vec,n, forest));
    }

    /* --------------------------------------------------- */

    /* labelled trees count
       A000169 1,1,2,9,64,625,7776,117649,2097152,
       n =     0 1 2 3  4   5

       labelled forests count
       A000272 1,1,3,16,125,1296,16807,262144,4782969
    */
    printf("labelled %s count  ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%d,",ltrees_count[n]);
    printf("\n");

    /* labelled trees total eccentricities
       not in OEIS: 0,0,4,45,592,8925,156636,3147907,71417536
       n =          0 1 2  3   4    5

       labelled forests total eccentricities
       not really sensible for forest, comes out as sum within tree, maybe
       not in OEIS: 0,0,4,57,844,13875,258726,5443893,128083320
    */
    printf("labelled %s total eccentricities  ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%d,",ltrees_total_eccentricities[n]);
    printf("\n");

    /* --------- */
    /* labelled trees total indomsets
       total 1,1,4,18,176,2150,34092,657188,14906432
       n =   0 1 2  3   4    5
       A058924 starting n=1

       labelled forests total indomsets
       not in OEIS: 1,1,5,31,321,4211,69853,1389697,32333857
    */
    printf("labelled %s total indomsets  ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%Lu,",ltrees_total_indomsets[n]);
    printf("\n");

    /* --------- */
    /* labelled trees total height
       A001854 0,0,2,15,148,1785,26106,449701,8927192
       n =     0 1 2  3   4    5

       labelled forests total height
       total   0,0,2,21,232,3055,47436,853755,17533440
       n =     0 1 2  3   4    5
       not in OEIS: 232,3055,47436,853755,17533440
    */
    printf("labelled %s total height  ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%Lu,",ltrees_total_height[n]);
    printf("\n");

    /* labelled trees total 2height
       total 0,0,0,9,100,1365,20946,372211,7562696,
       n =   0 1 2 3   4    5     6      7       8
       not in OEIS: 100,1365,20946,372211,7562696

       n=3 trees 6+3=9 of, second highest is always at depth=1
       7562696/8^7.0 == 3.60
       my(n=8); sqrt(Pi*n) - 7/6
       + n^(-1/2+1)

       labelled forests total 2height
       total 0,0,0,9,148,2195,36366,683193,14454952
       n =   0 1 2 3   4    5
       not in OEIS: 148,2195,36366,683193,14454952
    */
    printf("labelled %s total 2height ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%Lu,",ltrees_total_2height[n]);
    printf("\n");

    /* -------------------------------------------- */
    /* ahead trees
       same
       not in OEIS: 2,9,64,585,6576,88375,
       not in OEIS: 1,3,16,117,1096,12625,
       n =          2 3  4   5    6     7
       not in OEIS: 0,0, 0, 40,1200,29274,
       not in OEIS: 0,0, 0,  8, 200, 4182,
       different

       tree ahead same      0,2,15,116,1095,12624,173313,
       labelled tree ahead different 0,0,0,40,1200,29274,710640,
       tree ahead different 0,0,0,8,200,4182,88830,

       ahead forests
       same
       not in OEIS: 3,16,117,1096,12625,173314,
       not in OEIS: 1, 5, 29, 219, 2104, 24759,
       n =          0 1 2  3   5    5
       not in OEIS: 0,0,8,200,4182,88830,
       not in OEIS: 0,0,2, 40, 697,12690,
       different
    */
    printf("labelled %s ahead same      ", str);
    for (n = min_n; n <= max_n; n++) {
      printf("%d,",ltrees_ahead_same[n]);
      assert(ltrees_ahead_same[n] + ltrees_ahead_different[n]
             == ltrees_count[n]);
    }
    printf("\n");
    printf("         %s ahead same      ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%d,",ltrees_ahead_same[n]/(n?n:1));
    printf("\n");
    printf("labelled %s ahead different ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%d,",ltrees_ahead_different[n]);
    printf("\n");
    printf("         %s ahead different ", str);
    for (n = min_n; n <= max_n; n++)
      printf("%d,",ltrees_ahead_different[n]/(n?n:1));
    printf("\n");

    printf("\n");
  }
  printf("\n");
}

void
try_various (void)
{
  {
    /* binomial tree, n vertices */

    DYNALLSTAT(int,vpar,vpar_sz);
    const int max_n = 32;
    int  diameter[max_n+1];
    int  indnum[max_n+1];
    unsigned long long  indsets_count[max_n+1];
    unsigned long long  indomsets_count[max_n+1];
    int  domnum[max_n+1];
    unsigned long long  domsets_count[max_n+1];
    unsigned long long  minimal_domsets_count[max_n+1];
    unsigned long long  totdomsets_count[max_n+1];
    memset(totdomsets_count,'\0',sizeof(totdomsets_count));
    int  matchnum[max_n+1];
    unsigned long long  Wiener[max_n+1];
    unsigned long long  terminal_Wiener[max_n+1];

    int n;
    for (n = 1; n <= max_n; n++) {
      DYNALLOC1(int,vpar,vpar_sz,n+1,"malloc");
      nautyextra_make_binomial_tree_vpar(vpar,n);
      nautyextra_vpar_to_sg (vpar,n, &sg, 0); /* undirected */

      diameter[n]              = nautyextra_diameter_vpar(vpar,n);
      indnum[n]                = nautyextra_indnum_vpar(vpar,n);
      indsets_count[n]         = nautyextra_indsets_count_vpar(vpar,n);
      matchnum[n]              = nautyextra_matchnum_vpar(vpar,n);
      domnum[n]                = nautyextra_domnum_vpar(vpar,n);
      domsets_count[n]         = nautyextra_domsets_count_vpar(vpar,n);
      minimal_domsets_count[n] = nautyextra_minimal_domsets_count_vpar(vpar,n);
      indomsets_count[n]       = nautyextra_indomsets_count_vpar(vpar,n);
      if (n <= 20)
        totdomsets_count[n]    = nautyextra_totdomsets_count_sg(&sg);

      struct nautyextra_Wiener_info info;
      nautyextra_Wiener_index_vpar(vpar,n, 0, &info);
      Wiener[n]          = info.W;
      nautyextra_Wiener_index_vpar(vpar,n, 1, &info);
      terminal_Wiener[n] = info.W;
    }

    /* binomial tree  diameter
       0,1,2,3,3,4,4,5,5,5,5,6,6,6,6,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,9,
       A126236
    */
    printf("binomial tree   diameter ");
    for (n = 1; n <= max_n; n++)
      printf("%d,",diameter[n]);
    printf("\n");

    /* binomial tree  indsets_count
       not in OEIS: 2,3,5,8,14,22,38,60,108,168,288,456,816,1272,2184,3456,6336,9792,16704,26496,47232,73728,126720,200448,366336,566784,967680,1534464,2737152,4271616,7340544,11612160,
    */
    printf("binomial tree   indsets_count ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",indsets_count[n]);
    printf("\n");

    /* binomial tree  indomsets_count
       not in OEIS: 1,2,2,3,3,5,5,8,8,14,14,22,22,38,38,60,60,108,108,168,168,288,288,456,456,816,816,1272,1272,2184,2184,3456,
       unduplicated
       not in OEIS: 1, 2, 3, 5, 8, 14, 22, 38, 60, 108, 168, 288, 456, 816, 1272, 2184, 3456
    */
    printf("binomial tree   indomsets_count ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",indomsets_count[n]);
    printf("\n");

    /* binomial tree  indnum
       1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15,16,16,
       floor(n/2)
       A004526
    */
    printf("binomial tree   indnum ");
    for (n = 1; n <= max_n; n++)
      printf("%d,",indnum[n]);
    printf("\n");

    /* binomial tree  domnum
       1,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15,16,
       A065033 1 appears three times, other numbers twice.
    */
    printf("binomial tree   domnum ");
    for (n = 1; n <= max_n; n++)
      printf("%d,",domnum[n]);
    printf("\n");

    /* binomial tree  domsets_count
       .       1 2 3 4  5  6  7  8   9  10
       A083658 1,3,5,9,15,27,45,81,135,243,405,729,1215,2187,3645,6561,10935,19683,32805,59049,98415,177147,295245,531441,885735,1594323,2657205,4782969,7971615,14348907,23914845,43046721,
       a(2n) = 3^n, a(2n+1) = 5*3^(n-1) for n>0, with a(0)=1, a(1)=1.

       GP-DEFINE  vector_modulo(v,i) = v[(i% #v)+1];
       GP-DEFINE  binomial_tree_domsets_count(n) = if(n==0,1, n==1,1, \
       GP-DEFINE    vector_modulo([1, 5/3],n) * 3^floor(n/2));

       GP-Test  my(v=[1,3,5,9,15,27,45,81,135,243,405,729,1215,2187,3645,6561,10935,19683,32805,59049,98415,177147,295245,531441,885735,1594323,2657205,4782969,7971615,14348907,23914845,43046721]); \
       GP-Test  vector(#v,n, binomial_tree_domsets_count(n)) == v

       vector(10,n, vector_modulo([1, 5],n) * 3^(floor(n/2)-(n%2)))
    */
    printf("binomial tree   domsets_count ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",domsets_count[n]);
    printf("\n");

    /* binomial tree  minimal_domsets_count
       0 1 2 3 4 5 6 7  8  9 10
       1,2,2,4,4,8,8,16,16,32,32,64,64,128,128,256,256,512,512,1024,1024,2048,2048,4096,4096,8192,8192,16384,16384,32768,32768,65536,
       A016116  2^floor(n/2)
    */
    printf("binomial tree  minimal_domsets_count ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",minimal_domsets_count[n]);
    printf("\n");

    /* binomial tree  matchnum
       0,1,1,2,2,3,3,4,4,5,5,6,6,7,7,8,8,9,9,10,10,11,11,12,12,13,13,14,14,15,15,16,
       floor(n/2)
       A004526
    */
    printf("binomial tree   matchnum ");
    for (n = 1; n <= max_n; n++)
      printf("%d,",matchnum[n]);
    printf("\n");

    /* binomial tree   totdomsets_count
       0,1,3,4,8,8,16,16,32,32,64,64,128,128,256,256,512,512,1024,1024
       1 2 3 4 5 6  7  8  9 10
    */
    printf("binomial tree   totdomsets_count ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",totdomsets_count[n]);
    printf("\n");

    /* binomial tree  Wiener
       not in OEIS: 0,1,4,10,18,31,46,68,88,117,148,190,226,275,326,392,440,505,572,658,730,823,918,1036,1120,1229,1340,1478,1594,1739,1886,2064,
    */
    printf("binomial tree  Wiener ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",Wiener[n]);
    printf("\n");

    /* binomial tree  terminal_Wiener
       not in OEIS: 0,1,2,3,8,10,19,22,34,38,56,61,82,88,117,124,152,160,198,207,248,258,311,322,370,382,444,457,522,536,617,632,
    */
    printf("binomial tree  terminal_Wiener ");
    for (n = 1; n <= max_n; n++)
      printf("%Lu,",terminal_Wiener[n]);
    printf("\n");

    printf("\n");
  }

  {
    /* binomial tree, 2^k vertices */

    DYNALLSTAT(int,vpar,vpar_sz);
    const int max_k = 10;
    int  diameter[max_k+1];
    int  indnum[max_k+1];
    unsigned long long  indsets_count[max_k+1];
    unsigned long long  indomsets_count[max_k+1];
    int  domnum[max_k+1];
    unsigned long long  domsets_count[max_k+1];
    unsigned long long  minimal_domsets_count[max_k+1];
    int  matchnum[max_k+1];
    int  perfect_matching[max_k+1];
    unsigned long long  Wiener[max_k+1];
    unsigned long long  terminal_Wiener[max_k+1];

    int k;
    for (k = 0; k <= max_k; k++) {
      int n = 1 << k;
      DYNALLOC1(int,vpar,vpar_sz,n+1,"malloc");
      nautyextra_make_binomial_tree_vpar(vpar,n);

      diameter[k]              = nautyextra_diameter_vpar(vpar,n);
      indnum[k]                = nautyextra_indnum_vpar(vpar,n);
      if (k <= 6) {
        indsets_count[k]        = nautyextra_indsets_count_vpar(vpar,n);
        indomsets_count[k]      = nautyextra_indomsets_count_vpar(vpar,n);
        domsets_count[k]        = nautyextra_domsets_count_vpar(vpar,n);
        minimal_domsets_count[k]= nautyextra_minimal_domsets_count_vpar(vpar,n);
      }
      domnum[k]                = nautyextra_domnum_vpar(vpar,n);
      matchnum[k]              = nautyextra_matchnum_vpar(vpar,n);
      perfect_matching[k]      = nautyextra_has_perfect_matching_vpar(vpar,n);

      struct nautyextra_Wiener_info info;
      nautyextra_Wiener_index_vpar(vpar,n, 0, &info);
      Wiener[k]                = info.W;
      nautyextra_Wiener_index_vpar(vpar,n, 1, &info);
      terminal_Wiener[k]       = info.W;
    }

    /* binomial tree k   diameter
       diameter 1,3,5,7,9,11,
    */
    printf("binomial tree   diameter ");
    for (k = 0; k <= max_k; k++)
      printf("%d,",diameter[k]);
    printf("\n");

    /* binomial tree k   indsets_count
       3,8,60,3456,11612160,132090377011200,
       A347289 binomial tree order n num indsets
    */
    printf("binomial tree   indsets_count ");
    for (k = 0; k <= max_k; k++)
      printf("%Lu,",indsets_count[k]);
    printf("\n");

    /* binomial tree k   indomsets_count
       not in OEIS: 2,3,8,60,3456,11612160,
    */
    printf("binomial tree   indomsets_count ");
    for (k = 0; k <= max_k; k++)
      printf("%Lu,",indomsets_count[k]);
    printf("\n");

    /* binomial tree k   indnum
       1,2,4,8,16,32,
    */
    printf("binomial tree   indnum ");
    for (k = 0; k <= max_k; k++)
      printf("%d,",indnum[k]);
    printf("\n");

    /* binomial tree k   domnum
       1,2,4,8,16,32,
    */
    printf("binomial tree   domnum ");
    for (k = 0; k <= max_k; k++)
      printf("%d,",domnum[k]);
    printf("\n");

    /* binomial tree k   domsets_count
       3,9,81,6561,43046721,1853020188851841,
       3^(2^n)
       A011764
    */
    printf("binomial tree   domsets_count ");
    for (k = 0; k <= max_k; k++)
      printf("%Lu,",domsets_count[k]);
    printf("\n");

    /* binomial tree k   minimal_domsets_count
       2,4,16,256,65536,4294967296,
       A001146 a(n) = 2^(2^n).
    */
    printf("binomial tree  minimal_domsets_count ");
    for (k = 0; k <= max_k; k++)
      printf("%Lu,",minimal_domsets_count[k]);
    printf("\n");

    /* binomial tree k  matchnum
       1,2,4,8,16,32,
       perfect matching
    */
    printf("binomial tree   matchnum ");
    for (k = 0; k <= max_k; k++)
      printf("%d,",matchnum[k]);
    printf("\n");
    printf("binomial tree   perfect matching ");
    for (k = 0; k <= max_k; k++)
      printf("%d,",perfect_matching[k]);
    printf("\n");


    /* binomial tree k  Wiener
       1,10,68,392,2064,10272,
       A192021
    */
    printf("binomial tree  Wiener ");
    for (k = 0; k <= max_k; k++)
      printf("%Lu,",Wiener[k]);
    printf("\n");

    /* binomial tree k  terminal_Wiener
       not in OEIS: 0,1,3,22,124,632,3056,14304,65472,294784,1310464
       vector(11,k,k--; 1/8*k*4^k - 1/4*2^k)
       vector(11,k,k--; 1/2*( k*4^(k-1) - 2^(k-1)))

       .   k*4^(k-1) - 2^(k-1)
       .   ------------------
       .           2
    */
    printf("binomial tree  terminal_Wiener ");
    for (k = 0; k <= max_k; k++)
      printf("%Lu,",terminal_Wiener[k]);
    printf("\n");

    printf("\n");
  }

  {
    /* paths */

    DYNALLSTAT(int,vpar,vpar_sz);
    const int max_n = 20;
    unsigned long long  domsets_count[max_n+1];
    unsigned long long  minimal_domsets_count[max_n+1];
    unsigned long long  totdomsets_count[max_n+1];
    unsigned long long  indomsets_count[max_n+1];

    int n;
    for (n = 0; n <= max_n; n++) {
      DYNALLOC1(int,vpar,vpar_sz,n+1,"malloc");
      nautyextra_make_path_vpar(vpar,n);
      nautyextra_vpar_to_sg (vpar,n, &sg, 0); /* undirected */

      domsets_count[n]         = nautyextra_domsets_count_sg(&sg);
      minimal_domsets_count[n] = nautyextra_minimal_domsets_count_sg(&sg);
      totdomsets_count[n]      = nautyextra_totdomsets_count_sg(&sg);
      indomsets_count[n]       = nautyextra_indomsets_count_sg(&sg);
    }

    /* path          domsets_count
       0 1 2 3 4  5  6  7   8   9  10
       1,1,3,5,9,17,31,57,105,193,355,653,1201,2209,4063,7473,13745,25281,46499,85525,157305,
       A000213 Tribonacci

       1--2--3    2     1,2
       1,3   2,3
       1,2,3
       P(n) = P(n-1)        with n
       + Pwith(n-1)    without n, must n-1 to dom
       Pwith(n-1) = Pany(n-2)
       Pany(n-2) = Pwith(n-2) + Pwithout(n-2)
       = Pwith(n-2) + (Pany(n-2)-Pwith(n-2))

       Pwith(n)    = n>=1||error(); if(n==1,1, Pwith(n-1) + Pdom(n-1) + Pundom(n-1));
       Pdom(n)     = n>=1||error(); if(n==1,0, Pwith(n-1));
       Pundom(n)   = n>=1||error(); if(n==1,1, Pdom(n-1));
       P(n)        = n>=1||error(); Pwith(n) + Pdom(n);
       vector(12,n, Pwith(n))
       vector(12,n, Pdom(n))
       vector(12,n, Pundom(n))
       vector(12,n, P(n))
    */
    printf("path          domsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",domsets_count[n]);
    printf("\n");

    /* path  minimal_domsets_count
       0 1 2 3 4 5 6 7  8  9 10
       1,1,2,2,4,4,7,9,13,18,25,36,49,70,97,137,191,268,376,526,738,
       A253412 maximal set of 1s

       P(n) = if(n<=1,1, n<=3,2, n<=5,4, P(n-2) + P(n-3) + P(n-4) - P(n-6))
       vector(21,n,n--; P(n))
    */
    printf("path  minimal_domsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",minimal_domsets_count[n]);
    printf("\n");

    /* path  totdomsets count
       not in OEIS: 1,0,1,3,4,5,9,16,25,39,64,105,169,272,441,715,1156,1869,3025,4896,7921,
       .            0 1 2 3 4 5 6 7  8  9 10

       P(n) = if(n==0,1, n==1,0, n==2,1, n==3,3, P(n-1) + P(n-3) + P(n-4));

       vector_modulo([2/5, -4/5, -2/5, 4/5],n)
       + gf (3/5 + 1/5*x)/(1 - x - x^2)
       P(n) = (fibonacci(n) + 3*fibonacci(n+1) \
       + vector_modulo([2,-4,-2,4],n))/5;

       GP-DEFINE  Lucas(n) = fibonacci(n+1) + fibonacci(n-1);   \\ Lucas numbers
       P(n) = (Lucas(n+2) + vector_modulo([2, -4, -2, 4],n)) / 5;
       vector(21,n,n--; P(n))
    */
    printf("path  totdomsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",totdomsets_count[n]);
    printf("\n");

    /* path  indomsets count
       count 1,1,2,2,3,4,5,7,9,12,16,21,28,37,49,65,86,114,151,200,265,
       n =   0 1 2 3 4 5 6 7 8  9 10
       A000931 Padovan numbers
       compositions of n into parts == 2 mod 3
    */
    printf("path  indomsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",indomsets_count[n]);
    printf("\n");

    printf("\n");
  }

  {
    /* n-cycle */

    const int max_n = 20;
    unsigned long long  indsets_count[max_n+1];
    unsigned long long  domsets_count[max_n+1];
    unsigned long long  minimal_domsets_count[max_n+1];
    unsigned long long  totdomsets_count[max_n+1];
    unsigned long long  indomsets_count[max_n+1];

    int n;
    for (n = 0; n <= max_n; n++) {
      /* printf ("cycle n=%d\n", n); */
      nautyextra_make_cycle_sg(&sg,n);
      /* nautyextra_print_sg(&sg); */
      nautyextra_validate_sg(&sg);
      /* nautyextra_view_sg(&sg); */

      indsets_count[n]         = nautyextra_indsets_count_sg(&sg);
      domsets_count[n]         = nautyextra_domsets_count_sg(&sg);
      minimal_domsets_count[n] = nautyextra_minimal_domsets_count_sg(&sg);
      totdomsets_count[n]      = nautyextra_totdomsets_count_sg(&sg);
      indomsets_count[n]       = nautyextra_indomsets_count_sg(&sg);
    }

    /* cycle         indsets_count
       1,2,3,4,7,11,18,29,47,76,123,199,322,521,843,1364,2207,3571,5778,9349
       A169985 round phi^n
       A000032 Lucas numbers, after initial
    */
    printf("cycle         indsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",indsets_count[n]);
    printf("\n");

    /* cycle         domsets_count
       A001644 1,1,3,7,11,21,39,71,131,241,443,815,1499,
       form of tribonacci a(n) = a(n-1) + a(n-2) + a(n-3)
    */
    printf("cycle         domsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",domsets_count[n]);
    printf("\n");

    /* cycle minimal_domsets_count
       A253413 1,1,2,3,6,5,5,14,14,21,27,44,57,
       circular n-bit with maximal 1s (minimal 0s)
    */
    printf("cycle minimal_domsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",minimal_domsets_count[n]);
    printf("\n");

    /* cycle         totdomsets_count
       not in OEIS: 1,0,1,4,9,11,16,29,49,76,121,199,324,521,841,1364,2209,3571,5776,9349,15129,
       .            0 1 2 3 4  5  6

       GP-DEFINE  C(n) = if(n==0,1, n==1,0, \
       GP-DEFINE            Lucas(n) + vector_modulo([2,0,-2,0],n));
       GP-Test  my(v=[1,0,1,4,9,11,16,29,49,76,121,199,324,521,841,1364,2209,3571,5776,9349,15129]); \
       GP-Test  vector(21,n,n--; C(n)) == v
    */
    printf("cycle      totdomsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",totdomsets_count[n]);
    printf("\n");

    /* cycle indomsets_count
       1,1,2,3,2,5,5,7,10,12,17,22,29,39,51,68,90,119,158,209,277,
       A001608       7,10,12,17,22,29,39,51,68,90,119,158,209,277,
       Perrin a(n) = a(n-2) + a(n-3)  after initial values
    */
    printf("cycle       indomsets_count ");
    for (n = 0; n <= max_n; n++)
      printf("%Lu,",indomsets_count[n]);
    printf("\n");

    printf("\n");
  }

  {
    /* square grid */

    DYNALLSTAT(graph,g,g_sz);
    const int max_size = 5;
    int   num_edges_list[max_size+1];
    long  cycles_count_list[max_size+1];
    unsigned long long  indsets_count_list[max_size+1];
    unsigned long long  indomsets_count_list[max_size+1];
    unsigned long long  domsets_count_list[max_size+1];
    unsigned long long  minimal_domsets_count_list[max_size+1];
    unsigned long long  totdomsets_count_list[max_size+1];
    memset(indsets_count_list,'\0',sizeof(indsets_count_list));
    memset(indomsets_count_list,'\0',sizeof(indomsets_count_list));
    memset(domsets_count_list,'\0',sizeof(domsets_count_list));
    memset(minimal_domsets_count_list,'\0',sizeof(minimal_domsets_count_list));
    memset(totdomsets_count_list,'\0',sizeof(totdomsets_count_list));
    int s;
    for (s = 1; s <= max_size; s++) {
      int n = s*s;
      /* printf ("s=%d  n=%d\n", s, n); */
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g,g_sz,m,n,"malloc");
      nautyextra_make_grid(g,m,n, s,s);
      nauty_to_sg(g,&sg,m,n);
      /* nautyextra_print_matrix(g,m,n); */
      /* nautyextra_view(g,m,n); */

      unsigned long edges = 99999999;
      int mindeg=-1, mincount=-1, maxdeg=-1, maxcount=-1, eulerian=-1;
      degstats(g,m,n, &edges,&mindeg,&mincount,&maxdeg,&maxcount,&eulerian);
      long cycles = cyclecount(g,m,n);
      num_edges_list[s] = edges;
      cycles_count_list[s] = cycles;


      if (s*s <= 16) {   /* too slow for 25 */
        indsets_count_list[s] = nautyextra_indsets_count_sg(&sg);
        indomsets_count_list[s]= nautyextra_indomsets_count_sg(&sg);

        domsets_count_list[s]        = nautyextra_domsets_count_sg(&sg);
        minimal_domsets_count_list[s]= nautyextra_minimal_domsets_count_sg(&sg);
      }
      if (s*s <= 25) {   /* bit slow at 25 */
        totdomsets_count_list[s]     = nautyextra_totdomsets_count_sg(&sg);
      }
    }

    /* square grid edges 0,4,12,24,40
       A046092 4 times triangular
    */
    printf("square grid           edges  ");
    int num_vertices;
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%d,",num_edges_list[num_vertices]);
    }
    printf("\n");

    /* square grid cycles A140517 0,1,13,213,9349
     */
    printf("square grid          cycles  ");
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%ld,",cycles_count_list[num_vertices]);
    }
    printf("\n");

    /* square grid         indsets
       A006506 2,7,63,1234,55447,
    */
    printf("square grid         indsets  ");
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%Lu,",indsets_count_list[num_vertices]);
    }
    printf("\n");

    /* square grid domsets count
       A133515 1,11,291,28661
    */
    printf("square grid         domsets  ");
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%Lu,",domsets_count_list[num_vertices]);
    }
    printf("\n");

    /* square grid minimal domsets count
       A290382 1,6,16,306
    */
    printf("square grid minimal domsets  ");
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%Lu,",minimal_domsets_count_list[num_vertices]);
    }
    printf("\n");

    /* square grid indomsets (maximal indsets)
       A197048 1,2,10,42,358,
    */
    printf("square grid indomsets  ");
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%Lu,",indomsets_count_list[num_vertices]);
    }
    printf("\n");

    /* square grid totdomsets count
       A133793 0,9,161,11236,3270375
    */
    printf("square grid         totdomsets  ");
    for (num_vertices = 1; num_vertices <= max_size; num_vertices++) {
      printf("%Lu,",totdomsets_count_list[num_vertices]);
    }
    printf("\n");

    printf("\n");
  }


  {
    /* complete 2-ary tree  domsets_count
       not in OEIS: 1,5,49,4325,39394801

       . 2     2     1,2
       ./ \    1,3   2,3
       1   3         1,2,3

       .     1          none 2,4,5 = 16          with 2,3 = 32
       .   /   \        none 3,6,7 = 16       with 3,4,5 not 2 = 8 * 2
       . 2       3      128 - 32 = 96         with 1,4,5,6,7  = 1
       ./ \     / \     non  1,2,3 = 16
       4   5   6   7

       n==1,1,
       n==1,0,
       n==1,1,
       Twith(n)    = n>=0||error(); if(n==0,0, (Twith(n-1) + Tdom(n-1) + Tundom(n-1))^2 );
       Tdom(n)     = n>=0||error(); if(n==0,1, (Twith(n-1) + Tdom(n-1))^2 - Tdom(n-1)^2);
       Tundom(n)   = n>=0||error(); if(n==0,0, Tdom(n-1)^2);
       T(n)        = n>=0||error(); Twith(n) + Tdom(n);

       Twith(n)    = n>=0||error(); if(n==0,0, (T(n-1) + Tundom(n-1))^2 );
       Tdom(n)     = n>=0||error(); if(n==0,1, T(n-1)^2 - Tundom(n));
       Tundom(n)   = n>=0||error(); if(n==0,0, Tdom(n-1)^2);
       T(n)        = n>=0||error(); Twith(n) + Tdom(n);
       vector(6,n, Tundom(n))

       Twith(n)    = n>=0||error(); if(n==0,0, n==1,1, (T(n-1) + Tdom(n-2)^2)^2 );
       Tdom(n)     = n>=0||error(); if(n==0,1, n==1,0,  T(n-1)^2 - Tdom(n-1)^2);
       T(n)        = n>=0||error(); Twith(n) + Tdom(n);
       vector(6,n, Twith(n))

       Tdom(n)     = n>=0||error(); if(n==0,1, n==1,0,  T(n-1)^2 - Tdom(n-1)^2);
       T(n)        = n>=0||error(); if(n==0,1, n==1,1, (T(n-1) + Tdom(n-2)^2)^2 + T(n-1)^2 - Tdom(n-1)^2 )
       vector(6,n, Tdom(n))
       vector(6,n, T(n))
       T(1) == P(1)
       T(2) == P(3)
       T(n)        = n>=0||error(); if(n==0,1, n==1,1, (T(n-1) + Tdom(n-2)^2)^2 + Tdom(n) )
       Tdom(n-1)^2 = T(n-1)^2 - Tdom(n)
       T(n)        = n>=0||error(); if(n==0,1, n==1,1, (T(n-1) + T(n-2)^2 - Tdom(n-1))^2 + Tdom(n) )

       not in OEIS: 1, 4, 25, 2500, 24019801, 1825462026881476
       not in OEIS: 0, 1, 24, 1825, 15375000, 1315559720829601
       not in OEIS: 1, 0, 1, 576, 3330625, 236390625000000
       not in OEIS: 1, 5, 49, 4325, 39394801, 3141021747711077

       vector(6,n, T(n) - T(n-1))
       not in OEIS: 4, 44, 4276, 39390476, 3141021708316276

       Twith(n)    = ( Twith(n-1) + Tdom(n-1) + Tdom(n-2)^2 )^2
       Tdom(n)     = ( Twith(n-1) + Tdom(n-1) )^2  - Tdom(n-1)^2
       Tdom(n)     = Twith(n-1)^2 + 2*Twith(n-1)*Tdom(n-1)
       Tdom(n)     = Twith(n-1) * (Twith(n-1) + 2*Tdom(n-1))

       -----------------

       complete 2-ary tree  minimal_domsets_count
       not in OEIS: 1,2,4,41,1438,

       complete 2-ary tree  totdomsets_count
       not in OEIS: 0,3,25,1088,
    */

    DYNALLSTAT(int,vpar,vpar_sz);
    const int max_height = 4;
    int  num_children = 2;
    int  num_vertices[max_height+1];
    unsigned long long  domsets_count[max_height+1];
    unsigned long long  minimal_domsets_count[max_height+1];
    unsigned long long  totdomsets_count[max_height+1];

    int h;
    for (h = 1; h <= max_height; h++) {
      int n = nautyextra_make_complete_tree_size(num_children, h);
      num_vertices[h] = n;
      DYNALLOC1(int,vpar,vpar_sz,n+1,"malloc");
      nautyextra_make_complete_tree_vpar(vpar, num_children,h);
      nautyextra_vpar_to_sg (vpar,n, &sg, 0); /* undirected */

      domsets_count[h]         = nautyextra_domsets_count_vpar(vpar,n);
      minimal_domsets_count[h] = nautyextra_minimal_domsets_count_vpar(vpar,n);
      totdomsets_count[h]      = nautyextra_totdomsets_count_sg(&sg);
    }

    printf("complete %d-ary tree  num_vertices ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%d,",num_vertices[h]);
    }
    printf("\n");

    printf("complete %d-ary tree  domsets_count ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%Lu,",domsets_count[h]);
    }
    printf("\n");

    printf("complete %d-ary tree  minimal_domsets_count ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%Lu,",minimal_domsets_count[h]);
    }
    printf("\n");

    printf("complete %d-ary tree  totdomsets_count ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%Lu,",totdomsets_count[h]);
    }
    printf("\n");

    printf("\n");
  }

  {
    /* complete n-ary tree indnum, domnum */

    DYNALLSTAT(int,vpar,vpar_sz);
    const int max_height = 8;
    int  num_children = 3;
    int  indnum[max_height+1];
    int  domnum[max_height+1];

    int h;
    for (h = 1; h <= max_height; h++) {
      int n = nautyextra_make_complete_tree_size(num_children, h);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"malloc");
      nautyextra_make_complete_tree_vpar(vpar, num_children,h);
      nautyextra_validate_vpar(vpar,n);

      indnum[h] =  nautyextra_indnum_vpar(vpar,n);
      domnum[h] =  nautyextra_domnum_vpar(vpar,n);
    }

    /* complete 2-ary tree indnum
       1,2,5,10,21,42,85,170,341,682,1365,2730,5461,10922,21845,
       A000975 alterating bits 0,1
    */
    printf("complete %d-ary tree  indnum ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%d,",indnum[h]);
    }
    printf("\n");

    /* complete 2-ary tree domnum 1,1,2,5,9,18,37,73,146,293,585,1170,2341,4681
       A077947
       h=4  5 = 4 + 1
       h=5  9 = 8 + 1
       h=6  18 = 16 + 2
       h=7  37 = 32   + 4 + 1
       64  16 8   2
       h=8  73 =  64  +   8  +  1
       128  32 16  4  2
       1-bit gap 1 high, then 3 apart, gap <=1 low
    */
    printf("complete %d-ary tree  domnum ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%d,",domnum[h]);
    }
    printf("\n");


    /* complete 3-ary tree  indnum 1,3,10,30,91,273,820,2460,
       A033113
    */
    /* complete 3-ary tree  domnum
       not in OEIS: 1,1,3,10,28,84,253,757,
       ternary 1s spread as for binary above
    */

    /*
     *  @    @           .
     *      / \         / \
     *     .   .       @   @
     *                / \ / \
     */

    printf("\n");
  }

  {
    /* complete n-ary tree indsets */

    DYNALLSTAT(int,vpar,vpar_sz);
    const int max_height = 6;
    int num_children = 2;
    unsigned long long  indsets[max_height+1];
    unsigned long long  indomsets[max_height+1];

    int h;
    for (h = 1; h <= max_height; h++) {
      int n = nautyextra_make_complete_tree_size(num_children, h);
      DYNALLOC1(int,vpar,vpar_sz,n+1,"malloc");
      nautyextra_make_complete_tree_vpar(vpar, num_children,h);
      nautyextra_validate_vpar(vpar,n);
      /* printf("%d-ary h=%d ",num_children,h);nautyextra_print_vpar(vpar,n); */
      /* if (h==3) nautyextra_view_vpar(vpar,n); */

      indsets[h] = nautyextra_indsets_count_vpar(vpar,n);
      indomsets[h] = nautyextra_indomsets_count_vpar(vpar,n);
    }

    /* complete 2-ary tree  indsets 2,5,41,2306 = A076725
     */
    printf("complete %d-ary tree  indsets ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%Lu,",indsets[h]);
    }
    printf("\n");

    /* complete 2-ary tree  indsets maximal
       not in OEIS: 1,2,4,23,736,591137
    */
    printf("complete %d-ary tree  indsets maximal ",num_children);
    for (h = 1; h <= max_height; h++) {
      printf("%Lu,",indomsets[h]);
    }
    printf("\n");

    printf("\n");
  }
}

int
main (void)
{
  setbuf(stdout,NULL);
  SG_INIT(sg);
  SG_INIT(sg2);

  try_graphs();
  try_trees();
  try_ltrees();
  try_various();
  return 0;
}
